-----------------------------------------------------------------------------
-- |
-- Maintainer  :  Joost Visser
-- Stability   :  experimental
-- Portability :  portable
--
-- Test for type Inference module 
--
-----------------------------------------------------------------------------

module Language.Gnumeric.TypeInferenceHUnit where

import Language.Gnumeric.TypeInference
import Language.Gnumeric.Formula

import Data.Relation.SetOfPairs

import HUnit
import Debug.HUnitExtras

-----------------------------------------------------------------------------

main = do
 runTestsTT
   [ testSuite "inferExpression"
     [ TestCase $ do assertEqual "constant" emptyRel $ inferExpression _A1 (Var "oi")
     , TestCase $ do 
         let frm = Ref _A2 Nothing
         assertEqual "just a cell" (mkRel [(_A1,_A2)]) $ inferExpression _A1 frm
     , TestCase $ do 
         let frm = Ref _A2 (Just _A4)
         assertEqual "just a block" (mkRel [(_A1,_A2),(_A1,_A3),(_A1,_A4)]) $ inferExpression _A1 frm
     , TestCase $ do 
         let frm = Fun "sum" [Ref _A2 (Just _A4)]
         assertEqual "sum of block" (mkRel [(_A1,_A2),(_A1,_A3),(_A1,_A4)]) $ inferExpression _A1 frm
     , TestCase $ do 
         let frm = Fun "sum" [Ref _A2 Nothing, Ref _A3 Nothing, Ref _A4 Nothing]
         assertEqual "sum of cells" (mkRel [(_A1,_A2),(_A1,_A3),(_A1,_A4)]) $ inferExpression _A1 frm
     , TestCase $ do 
         let frm = Fun "if" [Ref _A2 Nothing, Ref _A3 Nothing, Ref _A4 Nothing]
         assertEqual "if with cells" (mkRel [(_A1,_A3),(_A1,_A4)]) $ inferExpression _A1 frm
     , TestCase $ do 
         let frm = Fun "if" [Ref _A2 Nothing, Fun "sum" [Ref _A3 Nothing], Ref _A4 Nothing]
         assertEqual "if with sum" (mkRel [(_A1,_A3),(_A1,_A4)]) $ inferExpression _A1 frm
     ]
    ]

-----------------------------------------------------------------------------

_A1 = mkCR "A" 1
_A2 = mkCR "A" 2
_A3 = mkCR "A" 3
_A4 = mkCR "A" 4

mkCR ls ds
  = CellRef {sheet="", colNr=toColNr ls, rowNr=toRowNr (show ds)}
  
-----------------------------------------------------------------------------
