--
-- Units.hs -- Units
-- Owner -- Robin Abraham
--

module Etc.Units where

import Util.PrintList
import Sheet.Sheet
import Sheet.Indx
import Sheet.Expr
--import SpreadSheet
-- import Sheet.Eval 
import Maybe
import List
import UCheck.HdrDef

{-

-- import Inputs

-- **Carried-over from previous version
-- Not clear what signOf is required for
-- Try to replace with definitions in new framework.
-- signOf "/"  = "/-"
-- signOf "/-" = "/-"

--signOf :: BinOp -> String
--signOf SlashM = "/-"
--signOf Slash  = "/-"

-- **

--
-- Types and Units
-- 

-}

data Type = NumT | StrT | TypeError deriving (Eq, Show)


data Unit = Base      String
          | Dependent Unit Unit    -- bottom-up 
	  | And       [Unit]
	  | Or        [Unit]
   	  | One
          | Unknown                -- we don't deal with these
          | Error                  -- for unit error
	   deriving (Show,Ord)

  
--instance Fractional Int where
--  (/) :: a ->a ->a 
 
 
instance Eq Unit where
  Base u        == Base v        = u == v
  Dependent u v == Dependent w x = u == w && v == x
  And us        == And vs        = all (`elem` vs) us && all (`elem` us) vs
  Or us         == Or vs         = all (`elem` vs) us && all (`elem` us) vs
  One          == One          = True
  Unknown      == Unknown      = True
  _             == _             = False

---- normalization of unit expressions
---- 
normU (Or [Dependent u' u,Dependent v' v]) | u==v = 
          Dependent (normU (Or [u',v'])) u
normU (Or [u,One]) = u
normU (Or (One:us)) = normU (Or us)
normU (Or [u,v]) | u==v = u
normU (Or ((And [a,b]):(And [c,d]):andU)) | a == c = normU (Or ((And [a, normU (Or (remOne [b,d]))]):(remOne andU)))
                                          | a == d = normU (Or ((And [a, normU (Or (remOne [b,c]))]):(remOne andU)))
                                          | b == c = normU (Or ((And [b, normU (Or (remOne [a,d]))]):(remOne andU)))
                                          | b == d = normU (Or ((And [b, normU (Or (remOne [a,c]))]):(remOne andU)))
                                          | otherwise =  Or (normUA ((And (remOne [c,d])):andU'))
                                                         where andU' = normUA ((And (remOne [a,b])):(remOne andU))
normU (Or [(Or (c:cs)),v])| cs == [] = normU (Or (remOne [c,v])) 
		          | c==v = normU (Or (remOne (c:cs)))
                          | otherwise= normU (Or (remOne ([c, v]++cs)))

normU (Or [v,(Or (c:cs))])| cs == [] = normU (Or (remOne [c,v])) 
		          | c==v = normU (Or (remOne (c:cs)))
                          | otherwise= normU (Or (remOne ([c, v]++cs)))
normU (And [u,One]) = u		          
-- Extra "simple" simplifications - check with Dr E
normU (Or [Or v]) = normU (Or v)
normU (And [And v]) = normU (And v)
normU (Or [And v]) = normU (And v)
normU (And [Or v]) = normU (Or v)

-- Handling of Unknown - they cannot cause unit error.
normU (Or [u,Unknown]) = u
normU (Or [Unknown,u]) = u
normU (And [u,Unknown]) = u
normU (And [Unknown,u]) = u
normU (And [a,b]) | a == b = a
                  | otherwise = And [a,b]
normU (And as) = (And as')
                  where as' = filter (\x -> x /= One) as
normU (Or as) = (Or as')
                  where as' = filter (\x -> x /= One) as
normU u = u

normUA [x] | x /= One = [x]
           | otherwise = []
normUA ((One):andU) = normUA andU 
normUA ((And []):andU) = normUA andU
normUA ((And [a,b]):(And [c,d]):andU)  | a == c = normUA ((And [a,b,d]):andU)
                                       | a == d = normUA ((And [a,b,c]):andU) 
                                       | b == c = normUA ((And [a,b,d]):andU) 
                                       | b == d = normUA ((And [a,b,c]):andU) 
                                       | otherwise =  ((And [c,d]):andU')
                                                      where andU' = normUA ((And [a,b]):andU)
normUA as = as   -- send it all back for other cases.

-- remOne removes occurances of One from list of units.
remOne :: [Unit] -> [Unit]
remOne [] = []
remOne (u:us) | u == One = remOne us
              | otherwise = u:(remOne us)


-- | ***********************************************************************
-- This doesn't work

simpfy :: SheetU -> Headers -> SheetF -> SheetU
simpfy (Sheet us) hs (Sheet cs) = Sheet (map (\(a,c)->(a,equit c hs cs)) us)
 
equit :: Unit -> Headers -> [Cell Fml] ->Unit
equit (And [Dependent us u, Dependent vs c]) hs  cs
       |us == getAllU u hs cs = And [u, Dependent vs c] 
	-- us == getAllU u hs cs && vs == getAllU c hs cs = And [u,c]
	-- us == getAllU u hs cs = And [u, Dependent vs c]
       |vs == getAllU c hs cs = And [Dependent us u, c]
      |otherwise = And [Dependent us u, Dependent vs c]
equit (Dependent us u) hs  cs = Dependent us u
equit (Base u) hs  cs = Base u
equit one hs  cs = One

getAllU :: Unit ->Headers -> [Cell Fml] -> Unit
getAllU u hs cs = normU (Or [(getSU a cs)|(a,[u'])<- hs, str==u'])
			where str = lookupAddr (show u) cs

getSU a cs = case lookup a cs of
		Just z -> Base (show z)
		_	-> One

lookupAddr :: String ->[Cell Fml]->Indx
lookupAddr u xys = head [x|(x,y) <- xys, Inp(S u) == y]
-- | ***********************************************************************

{-
-- instance Show Unit where
--  show (Base s)         = s
--  show (Dependent s u)  = show u ++ "[" ++ show s ++ "]"
--  show (And us)         = printList ["","&",""] show us
--  show (Or us)          = printList ["","|",""] show us
--  show One              = "_"
-}



type TypeA = [(Indx,Type)]


-- Unit inference
--
lookupStr :: Indx -> [Cell Fml] -> Maybe String
lookupStr a cs = case lookup a cs of 
                    Just (Inp (S s)) -> Just s
		    _                -> Nothing


inferU :: Headers -> SheetF -> SheetU
inferU hs (Sheet cs) = Sheet (map (\(a,c)->(a, inferUC cs hs (a,c))) cs)

--inferU hs (Sheet cs) = Sheet (map (\(a,c)->(a, (equit (inferUC cs hs (a,c))
--  hs cs)))  cs)                                                                     

inferUC :: [Cell Fml] -> Headers -> Cell Fml -> Unit
inferUC cs hs (a,Inp v) = getUnit cs hs a
-- first check the header info
-- no header then check Ref
-- "yes" header, then And units of header and Ref cell units
inferUC cs hs (a,Ref b) = normU (case lookup a hs of		   
			    -- Just v -> getUnit cs hs a
                            Just v -> case lookup b cs of
                                      Just k -> And (remOne [(getUnit cs hs a),(inferUC cs hs (b,k))])
				      _      -> (getUnit cs hs a)	
			    _ 	   -> case lookup b cs of  
			    		Just k -> inferUC cs hs (b,k)
			    		_      -> One)

-- Code for handling case of Binary operations.
-- we might have the cases:
-- 1) = 2 + 3   - simple mathematical operation
-- 2) = B2 + B3 - math operation involving references.
-- 3) = 2 * B3 - mixture of 1 and 2
-- we adopt same strategy as for Aggregation (the commented code adopts same strategy as for Ref).
-- inferUC cs hs (a,Bin o f1 f2) = normU (case lookup a hs of		   
--                                Just v -> And [(inferUC cs hs (a,f1)),(inferUC cs hs (a,f2))]
--			        _    -> And [(inferUC cs hs (a,f1)),(inferUC cs hs (a,f2))] )

-- inferUC cs hs (a,Bin o f1 f2)  = normU (Or [(inferUC cs hs (a,f1)),(inferUC cs hs (a,f2))])
inferUC cs hs (a,Bin o f1 f2)  = normU (Or (remOne [(inferUV cs hs f1),(inferUV cs hs f2)]))

-- Making changes to remove it's own header from affecting a aggregation's unit.
-- inferUC cs hs (a,Agg Plus xs)  = normU (And [bu,hu])
--			    where hu = getUnit cs hs a
--			    	  bu = normU (Or (map (inferUV cs hs) xs))

--inferUC cs hs (a,Agg Mult (x:xs))
--      |xs == [] = hunit
--      |unit1 /= One && all(== One)(map (inferUV cs hs)xs)
--	      = And [unit1, hunit]
--      |unit1 == One = inferUC cs hs (a,Agg Mult xs)
--      |otherwise = hunit
--				  where unit1 = inferUV cs hs x
--					hunit = getUnit cs hs a  
				   				   
inferUC cs hs (a,Agg Plus xs)  = normU (Or (remOne (map (inferUV cs hs) xs)))
inferUC cs hs (a,Agg Mult (x:xs))
      |unit1 /= One && all(== One)(map (inferUV cs hs)xs)
	      = inferUV cs hs x
      |unit1 == One = inferUC cs hs (a,Agg Mult xs)
      |otherwise = One
                   where unit1 = inferUC cs hs (a,Agg Mult xs)
{-
inferUC cs hs (a,Agg  f (x:xs))
      = case signOf f of
	 "/-" -> case x of
	          Ref b ->normU (And [(inferUC cs hs (a, Ref b)),(getUnit cs hs a)])
		  _     -> One			   		
	 _      -> One  			
-}
inferUC _ _ _ = One


inferUV :: [Cell Fml] -> Headers -> Fml -> Unit
inferUV cs hs (Inp v) = One
inferUV cs hs (Ref a) = case lookup a cs of
                            Just v -> inferUC cs hs (a,v)
			    _      -> One
-- new code for handling Bin
inferUV cs hs (Bin o f1 f2) = Or [(inferUV cs hs f1),(inferUV cs hs f2)]
inferUV _ _ _ = One

getUnit :: [Cell Fml] -> Headers -> Indx -> Unit
getUnit cs hs a = case lookup a hs of
		     Nothing  -> One
		     Just []  -> One
-- 		     Just [b] -> maybe One Base (lookupStr b cs)
		     Just bs -> traceUnits cs hs (zip bs (map (getU cs) bs))

getU cs b = maybe One Base (lookupStr b cs)

traceUnits :: [Cell Fml] -> Headers -> [(Indx,Unit)] -> Unit
traceUnits cs hs bus = case map (traceUnit cs hs) bus of
                         [u] -> u
			 us  -> And us

traceUnit :: [Cell Fml] -> Headers -> (Indx,Unit) -> Unit
traceUnit cs hs (a,u) = case getUnit cs hs a of
                          One -> u
			  v    -> Dependent u v    


-- unitChecker and the following functions check for unit errors.

unitChecker :: SheetU -> [Indx]
unitChecker (Sheet us) = map fst (filter (\(i,u) -> not (validUnit u)) us) 
                         -- map (\(i,u) -> (i,validUnit u)) us 


--validUnit :: Unit -> Bool
--validUnit One = True
--validUnit (Base u) = True
--validUnit (Dependent u1 u2) = (validUnit u1) && (validUnit u2)
--validUnit (Or os) = (map (\x -> validUnit x) os) && (validOr os)
--validUnit (And as) = (map (\x -> validUnit x) as) && (validAnd as)

--validOr :: [Unit] -> Bool


--validAnd :: [Unit] -> Bool


-- | **Auxilliary functions for creating SpreadSheets for testing purposes.
-- ** 
-- Smart constructors

ii = Inp . I
is = Inp . S
ie = Inp . E
iu = Inp Unused

agg :: BinOp -> [Indx] -> Fml
agg op ind = Agg op (map Ref ind)

lift :: SheetV -> SheetF
lift = fmap Inp

type SheetU = Sheet Unit

-- ** Stuff for testing 

--
-- Example
--

{-
    A      B      C     D       E
1        Fruit
2  Month Apple   Orange Plum    Total 
3    May     8     11    21     *  
4   June    75     5     15     /
5   July    30     @B3   30     +
6    @D2   SUM    SUM  SUM      @Z9
7          @C3    

-}

cols@[a,b,c,d,e] = map Indx [1..5]
refs@[ra,rb,rc,rd,re,rf] = map (Ref .) cols

harvest :: SheetF
harvest = Sheet [(b 1,is "Fruit"), 
                 (a 2,is "Month"),(b 2,is "Apple"),(c 2,is "Orange"),(d 2,is "Plum"),(e 2,is "Total"), 
                 (a 3,is "May"),(b 3,ii 8),(c 3,ii 11),(d 3,ii 21),(e 3,agg Plus [b 3,c 3,d 3]),
                 (a 4,is "June"),(b 4,ii 75),(c 4,ii 5),(d 4,ii 15),(e 4,agg Plus [b 4,c 4,d 4]),
                 (a 5,is "July"),(b 5,ii 30),(c 5,Ref (b 3)),(d 5,ii 30),(e 5,agg Plus [b 5,c 5,d 5]),
                 -- (a 6,Ref (d 2)),
                 (a 6, iu),(a 7,iu), (e 6, agg Plus [b 6,c 6,d 6]),
                 (b 6,agg Plus [b 3,b 4,b 5]),(c 6,agg Plus [c 3,c 4,c 5]),(d 6,agg Plus [d 3,d 4,d 5]),
                 (b 7,iu),(c 7,iu),(d 7,iu),(e 7,iu)
          ]

-- for testing ...
--Sheet h = harvest 
-- testing stuff
-- (c 5,Ref (b 3)) agg Plus [(b 3),(c 4)]

--result :: SheetV
--result = eval harvest

-- not needed yet ...
-- harvestTypeA = []

-- information about headers gathered from the user
-- 
harvestHeaders = [(b 1,[]),(b 2,[b 1]),(c 2,[b 1]),(d 2,[b 1]),
                  (a 2,[]),(a 3,[a 2]),(a 4,[a 2]),(a 5,[a 2]),
                  (b 3,[a 3,b 2]),(c 3,[a 3,c 2]),(d 3,[a 3,d 2]),
                  (b 4,[a 4,b 2]),(c 4,[a 4,c 2]),(d 4,[a 4,d 2]),
                  (b 5,[a 5,b 2]),(c 5,[a 5,c 2]),(d 5,[a 5,d 2])]

-- ** Uncomment for testing
units :: SheetU
units = inferU harvestHeaders harvest
-- sunits = simpfy units harvestHeaders harvest

-- we only have a form of weak subject reduction:
-- 
--unitsR :: SheetU
--unitsR = inferU harvestHeaders (lift result)

-- ** Header inference **
-- inferH :: SheetF -> Headers

myne = Or [(And [(Dependent (Or [Base "May",Base "June",Base "July"]) (Base "Month")) ,(Dependent (Base "Apple") (Base "Fruit"))]), (And [(Dependent (Or [Base "May",Base "June",Base "July"]) (Base "Month")) ,(Dependent (Base "Orange") (Base "Fruit"))]),(And [(Dependent (Or [Base "May",Base "June",Base "July"]) (Base "Month")) ,(Dependent (Base "Plum") (Base "Fruit"))]) ]

myne1 = Or [(And [(Dependent (Or [Base "May",Base "June",Base "July"]) (Base "Month")) ,(Dependent (Base "Apple") (Base "Fruit"))]), (And [(Dependent (Or [Base "May",Base "July",Base "June"]) (Base "Month")) ,(Dependent (Base "Orange") (Base "Fruit"))]),(And [(Dependent (Or [Base "July",Base "June",Base "July"]) (Base "Month")) ,(Dependent (Base "Plum") (Base "Fruit"))]) ]

-- Auxilliary functions

-- pairUp generates all possible pairs of elements in two lists.
pairUp :: Eq a => [a] -> [a] -> [(a,a)]
pairUp xs ys = [(x,y) | x <- xs, y <- ys, x /= y]



-- **** Old (incorrect) stuff -> Refactor this.

validUnit :: Unit -> Bool
validUnit One = True
validUnit (Base u) = True
validUnit (Dependent u1 u2) = (validUnit u1) && (validUnit u2)
validUnit (And []) = True
validUnit (And as) = aCheckPrefix ps && (foldr1 (&&) (map (\x -> validUnit x) as))
                       where ps = pairUp as as
validUnit (Or []) = True
validUnit (Or os) = oCheckPrefix ps && (foldr1 (&&) (map (\x -> validUnit x) os))
                    where ps = pairUp os os


aCheckPrefix :: [(Unit,Unit)] -> Bool
aCheckPrefix [] = True
aCheckPrefix ((Base a,Base b):ps) = aCheckPrefix ps
aCheckPrefix ((Base a, Dependent u v):ps) = (Base a) /= v && aCheckPrefix ps
aCheckPrefix ((Dependent u v,Base a):ps) = (Base a) /= v && aCheckPrefix ps
aCheckPrefix ((Base a,Or os):ps) = (oCheckPrefix (pairUp [Base a] os)) && aCheckPrefix ps
aCheckPrefix ((Or os,Base a):ps) = (oCheckPrefix (pairUp [Base a] os)) && aCheckPrefix ps
aCheckPrefix ((Dependent u' u, Dependent v' v):ps) = v /= u && (aCheckPrefix ps)
aCheckPrefix ((Dependent u' u, Or os):ps) =  (aCheckPrefix ps') && (aCheckPrefix ps)
                                             where ps' = pairUp [Dependent u' u] os
aCheckPrefix ((Or os, Dependent u' u):ps) =  (aCheckPrefix ps') && (aCheckPrefix ps)
                                             where ps' = pairUp [Dependent u' u] os
aCheckPrefix ((And as, And as'):ps) = aCheckPrefix ps' && aCheckPrefix ps
                                     where ps' = pairUp as as'
aCheckPrefix ((Or as, Or as'):ps) = oCheckPrefix ps' || aCheckPrefix ps  -- \|\| to &&
                                     where ps' = pairUp as as'
aCheckPrefix ((a,b):ps) = a == b && aCheckPrefix ps

oCheckPrefix :: [(Unit,Unit)] -> Bool
oCheckPrefix [] = True
oCheckPrefix ((Base a,Base b):ps) = oCheckPrefix ps
-- oCheckPrefix ((Base a, Dependent u v):ps) = (Base a) == v && oCheckPrefix ps
-- oCheckPrefix ((Dependent u v, Base a):ps) = (Base a) == v && oCheckPrefix ps
oCheckPrefix ((Base a, Dependent u v):ps) = oCheckPrefix ps
oCheckPrefix ((Dependent u v, Base a):ps) = oCheckPrefix ps
oCheckPrefix ((Dependent u' u, Dependent v' v):ps) = v == u && (oCheckPrefix [(u',v')]) && (oCheckPrefix ps)
oCheckPrefix ((Dependent u' u, And os):ps) =  (oCheckPrefix ps') && (oCheckPrefix ps)
                                             where ps' = pairUp [Dependent u' u] os
oCheckPrefix ((And os, Dependent u' u):ps) =  (oCheckPrefix ps') && (oCheckPrefix ps)
                                             where ps' = pairUp [Dependent u' u] os
oCheckPrefix ((And as, And as'):ps) = aCheckPrefix ps' && oCheckPrefix ps --- || to &&
                                     where ps' = pairUp as as'
oCheckPrefix ((Or as, Or as'):ps) = oCheckPrefix ps' || oCheckPrefix ps  --- || to &&
                                     where ps' = pairUp as as'
oCheckPrefix ((a,b):ps) = a == b && oCheckPrefix ps

-- **** End of old stuff
--This is a comment232 to see if the CVS repository is working
