--
-- UnitsFinal.hs -- Units info (should change file and module names...)
-- Owner -- Scott Andrew
--

module Etc.UnitsFinal where

import Util.PrintList
import Sheet.Sheet
import Sheet.Indx
import Sheet.Expr
-- import SpreadSheet
-- import Sheet.Eval 
import Maybe
import List
import UCheck.Header
import UCheck.HdrAux

type Base = String

newtype Unit = U {unU :: [[[Base]]]}
   deriving(Ord,Eq)

type Headers = [(Indx,[Indx])]

type ErrorReport = [ErrSit]
type ErrSit =  (Indx, ErrDesc)

stNoGen = (Sheet $ [],[],False)
	
data ErrDesc = RefErr Indx
		| RngSml Indx Indx
		| RngOfst Indx Indx
		| UnExRef
		| Omission Indx  
		| RngLrg Indx Indx
		| SuspectRange Indx Indx
		| UnkErr
		| Temp
		| NoErr
		  deriving(Show,Eq)

-- signOf :: BinOp -> String
-- signOf SlashM = "/-"
-- signOf Slash  = "/-"

-- Unit inference Begin --
lookupStr :: Indx -> [Cell Fml] -> Maybe String
lookupStr a cs = case lookup a cs of 
                    Just (Inp (S s)) -> Just s
		    _                -> Nothing

inferU :: Headers -> SheetF -> SheetU
inferU hs (Sheet cs) = Sheet (map (\(a,c)->(a, inferUC cs hs (a,c))) cs)
                                                                
inferUC :: [Cell Fml] -> Headers -> Cell Fml -> Unit
inferUC cs hs (a,Inp v) = getUnit cs hs a
-- first check the header info
-- no header then check Ref
-- "yes" header, then And units of header and Ref cell units
inferUC cs hs (a,Ref b) =  (case lookup a hs of		   			  
                            Just v -> case lookup b cs of
                                      Just k -> mA ((Sheet cs),hs) (getUnit cs hs a) (inferUC cs hs (b,k))
				      _      -> (getUnit cs hs a)	
			    _ 	   -> case lookup b cs of  
			    		Just k -> inferUC cs hs (b,k)
			    		_      -> U [[[]]])

-- Code for handling case of Binary operations.
-- we might have the cases:
-- 1) = 2 + 3   - simple mathematical operation
-- 2) = B2 + B3 - math operation involving references.
-- 3) = 2 * B3 - mixture of 1 and 2
-- we adopt same strategy as for Aggregation 


inferUC cs hs (a,Bin o f1 f2)  = mU ((Sheet cs),hs,True) (inferUV cs hs f1) (inferUV cs hs f2)			   	   
inferUC cs hs (a,Agg Plus xs)  =  foldr1 (mU ((Sheet cs),hs,True))  (map (inferUV cs hs) xs)
inferUC cs hs (a,Agg Mult (x:xs))
      |unit1 /= U [[[]]] && all(== U [[[]]])(map (inferUV cs hs)xs)
	      = inferUV cs hs x
      |unit1 == U [[[]]] = inferUC cs hs (a,Agg Mult xs)
      |otherwise = U [[[]]]
                   where unit1 = inferUC cs hs (a,Agg Mult xs)				   				  
inferUC cs hs (a,Agg  f (x:xs))
--      = case signOf f of
--	 "/-" -> 
	= case x of
	          Ref b -> mA ((Sheet cs),hs) (inferUC cs hs (a, Ref b)) (getUnit cs hs a)	   				
		  _     -> U [[[]]]			   		
--	 _      -> U [[[]]]  		
inferUC _ _ _ = U [[[]]]


inferUV :: [Cell Fml] -> Headers -> Fml -> Unit
inferUV cs hs (Inp v) = U [[[]]]
inferUV cs hs (Ref a) = case lookup a cs of
                            Just v -> inferUC cs hs (a,v)
			    _      -> U [[[]]]
-- new code for handling Bin
inferUV cs hs (Bin o f1 f2) = 	mU ((Sheet cs),hs,True) (inferUV cs hs f1) (inferUV cs hs f2)			
inferUV _ _ _ = U [[[]]]

getUnit :: [Cell Fml] -> Headers -> Indx -> Unit
getUnit cs hs a = case lookup a hs of
		     Nothing  -> U [[[]]]
		     Just []  -> U [[[]]]
		     Just bs -> traceUnits cs hs (zip bs (map (getU cs) bs))

getU cs b = maybe (U [[["One"]]]) liftB (lookupStr b cs)

traceUnits :: [Cell Fml] -> Headers -> [(Indx,Unit)] -> Unit
traceUnits cs hs bus = case map (traceUnit cs hs) bus of
                         [u] ->  u 
		 	 us  -> foldr1 (mU ((Sheet cs),hs,True)) us 

traceUnit :: [Cell Fml] -> Headers -> (Indx,Unit) -> Unit
traceUnit cs hs (a,U [[u:us]]) = case getUnit cs hs a of
                        	   U [[[]]] -> U [[u:us]] 
			           v   ->   mU ((Sheet cs),hs,True) v (U [[u:us]]) 

traceUnit cs hs (a, u) = case getUnit cs hs a of
			 U [[[]]] -> u
			 v  -> mU ((Sheet cs),hs,True) v u 

-- **Auxilliary functions for creating SpreadSheets for testing purposes -- 

-- Smart constructors --

ii = Inp . I
is = Inp . S
ie = Inp . E
iu = Inp Unused

agg :: BinOp -> [Indx] -> Fml
agg op ind = Agg op (map Ref ind)

lift :: SheetV -> SheetF
lift = fmap Inp

type SheetU = Sheet Unit

--
-- Example
--

{-
    A      B      C     D       E
1        Fruit
2  Month Apple   Orange Plum    Total 
3    May     8     11    21     *  
4   June    75     5     15     /
5   July    30     4   30     +
6          SUM    SUM  SUM      SUM
7              

-}

cols@[a,b,c,d,e,f] = map Indx [1..6]
refs@[ra,rb,rc,rd,re,rf] = map (Ref .) cols

-- ** Uncomment for testing
units :: SheetU
units = inferU harvestHeaders harvest


harvest :: SheetF
harvest = Sheet [(b 1,is "Fruit"),
                 (a 2,is "Month"),(b 2,is "Apple"),(c 2,is "Orange"),(d 2,is "Plum"),(e 2,is "Total"), 
                 (a 3,is "May"), (b 3, ii 5), (c 3, ii 9),(d 3,ii 21),(e 3,agg Plus [b 3,c 3, d 3]),
                 (a 4,is "June"), (b 4, ii 2), (c 4, ii 4),(d 4,ii 15),(e 4,agg Plus [b 4,c 4, d 4]),
                 (a 5,is "July"),(b 5,ii 5),(c 5, ii 2),(d 5, ii 10), (e 5, Agg Plus [Ref(b 5), Ref(c 5), Ref(d 5)]),           
                 (a 6, iu),(a 7,iu), (e 6, agg Plus [b 6,c 6, d 6]), (b 6,agg Plus [ b 3 ,b 4, b 5])
 		 ,(c 6,agg Plus [ c 3,c 4, c 5]),(d 6,agg Plus [d 3, d 4, d 5])]
          
-- information about headers gathered from the user -- 

harvestHeaders = [(b 1,[]),(b 2,[b 1]),(c 2,[b 1]),(d 2,[b 1]),
                  (a 2,[]),(a 3,[a 2]),(a 4,[a 2]),(a 5,[a 2]),
                  (b 3,[a 3,b 2]),(c 3,[a 3,c 2]),(d 3,[a 3,d 2]),
                  (b 4,[a 4,b 2]),(c 4,[a 4,c 2]),(d 4,[a 4,d 2]),
                  (b 5,[a 5,b 2]),(c 5,[a 5,c 2]),(d 5,[a 5,d 2])]


-- unitChecker and the following functions check for unit errors. --

unitChecker :: SheetU -> [Indx]
unitChecker (Sheet us) = map fst (filter (\(i,u) -> not (wfu u)) us) 
                         -- map (\(i,u) -> (i,validUnit u)) us 
		
-- Determines if a unit is well-formed
wfu:: Unit -> Bool
wfu (U[[[]]]) = True
wfu (U[[d:ds]]) = True
wfu u@(U[a'@(a:as)]) =  validAnd u
wfu u@(U(o'@(o:os))) =  p && p' && validAnd u
			where 
			   p = not $ elem [[]] o'
			   p'= and $ map(validOrPair rules) unitPairs
			   unitPairs = map(\(x,y) -> (liftO x, liftO y)) $ pairUp o' o'
			   rules = orRules $ orExcep u
		
validAnd:: Unit -> Bool
validAnd (U [a'@(a:as)]) =  isUnique $ map(myInit) a'
validAnd (U a'@(a:as)) = and $ map(validAnd.liftO) a'
		
-- WFU "analyzing Or Unit" functions Begin--

-- Takes a unit to generate a list of dependent exceptions that are allowed.
-- For example if we have: F[A] & M[M] || F[A] & M[J] => [[M],[J]]
orExcep:: Unit -> [[[String]]]
orExcep (U(o'@(o:os))) = nub $ map(snd) $ nubBy(\x y ->( length (elemIndices x result) < 2) == True ) $ result
			where
			result =  snd $ break(\x -> fst x == True) $ sort $ map(\(x,y) -> isEquiv (liftO x) (liftO y)) $(pairUp o' o')
orExcep _ = []


-- Takes a list of exceptions and generates a list of allowable well formed units.

orRules:: [[[String]]] -> [Unit]
orRules xs@(x:y:xs') =  nub $ filter(validAnd) $ distribute (mA (Sheet [],[])) (xss) (xss)
		     where xss = map(liftA) $ concat xs
orRules xs = []

-- An OrPair is valid if: 1.  It differs in at most one dependent leaf
--                      : 2.  Its structure allows an exception
				-- For example: F[A] & M[M] || F[A] & M[J] || F[O] & M[M] || F[O] & M[J]
			        -- Then the pair (F[A] & M[M],F[O] & M[J]) is allowed. 

validOrPair:: [Unit] -> (Unit,Unit) -> Bool
validOrPair rules (u,u') =  p || p'
			 where p =  (elem u rules) && (elem u' rules)
			       p' = fst $ isEquiv u u'

-- WFU analyzing OR functions End--

----------------------------------------------
-- Smart Constructors. --
----------------------------------------------


mA:: (SheetF,Headers) -> Unit -> Unit -> Unit
mA  (fs,hs) (U [a'@(a:as)]) (U[b'@(b:bs)]) = normA fs hs $ U $ nub $ [a'++b']	-- && on two Ands

mA  (fs,hs) (U [[a'@(a:as)]]) (U [[[b]]]) = normA fs hs $ U $ nub $  [[a',[b]]] -- And on base/dep
mA st u@(U [[[b]]]) u'@ (U [[a'@(a:as)]]) = mA st u' u

mA  (fs,hs) u@(U a'@(a:as))  u'@(U [b'@(b:bs)]) =  normA fs hs  $ U $ distribute (\x y -> (++) x y) [b'] a' 	-- && on And/Or
mA  st u@(U [b'@(b:bs)])  u'@(U a'@(a:as)) =  mA st u' u



mU:: (SheetF,Headers,Bool) -> Unit -> Unit -> Unit
mU st (U [[[a]]]) (U [[[b]]]) =  U [[[a,b]]] -- Dep' on two bases
mU (fs,hs,g) (U [[a'@(a:as)]]) (U [[b'@(b:bs)]]) = normO fs hs g $ U [[a',b']] -- && on two dependents
mU (fs,hs,g) (U [a'@(a:as)]) (U[b'@(b:bs)]) =  normO fs hs g  $ U $ [a',b'] -- \|\| on two ands 
mU (fs,hs,g) a b = normO fs hs g $ U (unU a ++ unU b)

normO:: SheetF -> Headers -> Bool -> Unit ->  Unit
normO fs hs g u | g = generalize u' fs hs
		| otherwise = u' 
		  where u' = U $ sort $ nub $ map( sort.nub) $ unU u


normA:: SheetF -> Headers -> Unit ->  Unit
normA fs hs u = remOne $ U $  sort $ nub $ map( (filter(/= ["One"])).sort.nub)$ unU u		

--End of SmartConstructors


-- Show functions for Units --

instance Show Unit where
	show( U [[[]]]) = "One"
        show(U u) = printOr u


printOr  =  printList[""," || ",""] printAnd
printAnd =  printList[""," && ",""] printDep 
printDep =  printList["","<",""] id

 

-- Unit utility Functions --

-- Returns the leaves(Dependent Units) of a unitTree
unitLeaf:: Unit -> [[String]]
unitLeaf (U [a'@[a:as]]) = a'
unitLeaf (U [a'@(a:as)]) =   foldr1(++) $ map(unitLeaf.liftA ) (a:as)
unitLeaf (U a'@(a:as)) =  foldr1(++) $  map(unitLeaf.liftO)  a'			
unitLeaf  _ = [[]]

unitTrav:: Unit -> ([String] -> [String]) -> Unit
unitTrav (U [a'@[a:as]]) f = 	liftO $ map(f) a'
unitTrav (U [a'@(a:as)]) f =   foldr1(mA (Sheet $[],[]) ) $ map((flip unitTrav f ).liftA ) a'
unitTrav (U a'@(a:as)) f =  foldr1(mU stNoGen) $  map((flip unitTrav f ).liftO)  a'			
unitTrav (U [[[]]])  _ = U[[[]]]


-- Experimental implementation of partial ordering/EQ stuff----------------------------------

--Returns true if the Units differ in only one leaf. Also returns those leaves that they differ.
isEquiv:: Unit -> Unit -> (Bool,[[String]])
isEquiv u1 u2 = if length ls <=1 && length ls' <=1 then (True, [head (ls \\ ls'),head (ls' \\ ls)]) else (False,[])
		where 
			lvs  = unitLeaf u1
			lvs' = unitLeaf u2
			ls   = lvs \\ lvs'
			ls'  = lvs' \\ lvs
			


pO:: Unit -> Unit -> Bool
pO u u' = (fst $ addDep) &&( fst $ isEquiv uDep u')
	where
	  uDep = unitTrav u (f addDep) 
	  addDep = addDeps u u'
	  f = (\(x,[a,b]) z -> if x then if a == z then b else z else z) 

addDeps:: Unit -> Unit -> (Bool,[[String]])
addDeps u u' = if result == [] then (False,[]) else result !! 0
	where 
	    l  = unitLeaf u
	    l' = unitLeaf u'
	    result = filter(fst) $ map(\(x,y) -> (subset x y,[x,y])) $ pairUp l l' 
----------------------------------------------------------------------------------------
	   
liftB:: String -> Unit
liftB s = U [[[s]]]

liftA:: [String] -> Unit
liftA s = U [[s]]

liftO:: [[String]] -> Unit
liftO s = U [s]

liftU:: [[[String]]] -> Unit
liftU s = U s

remOne:: Unit -> Unit
remOne (U [us]) = U $ [filter(/= []) us]
remOne u = u

remUnit:: Unit -> Unit -> Unit
remUnit (U x) (U y) = U( y \\ x)

--End of Unit Utility Functions -- 

-- Generalization Code (Start) -- 

generalize:: Unit -> SheetF -> Headers -> Unit
generalize u@(U u') sFml@(Sheet fs) hs | suggs == [] = u
				       | otherwise = generalize (applyGen (sFml,hs,False) u (head suggs)) sFml hs 
				where 
				f = filter(\(x,ys) -> and $ map(\x -> subset (unU x) u') ys)
				suggs = f $ genSugg (getHeaderUnit sFml hs (flipHeader hs))
			              

applyGen:: (SheetF,Headers,Bool) -> Unit -> (Unit,[Unit]) -> Unit
applyGen st u (u', us) = mU st   u'  result
			where result = foldr remUnit u us
			

-- Derrives the units associated with the headers
-- Ex: (U "Fruit",[ U Fruit[Apple], U Fruit[Orange], U Fruit[Plum] ])
-- Second argument is flipped Headers --
getHeaderUnit:: SheetF -> Headers -> Headers -> [(Unit,[Unit])]
getHeaderUnit sFml@ (Sheet fs)  hs ((i,(i':is)):iss) 
			|  isString f = (gen,reqs): getHeaderUnit sFml hs iss
		        |  otherwise =  getHeaderUnit sFml hs iss  
				where gen = traceUnit fs hs (i,liftB $ fromJust $ lookupStr i fs)
				      u' = head $ filter(/= i) $ fromJust $ lookup i' hs
				      reqs =  map(\x -> traceUnit fs hs (x,liftB $ fromJust $ lookupStr x fs)) (i':is)	
				      f = getFml i' sFml
				      hs' = flipHeader hs
getHeaderUnit sFml hs iss = [] 

-- Takes a list of headerUnits and creates a generalization suggestion
-- (Unit to generalize to, if [Unit] is present)
genSugg :: [(Unit,[Unit])] -> [(Unit,[Unit])]
genSugg hds = z ++ z'
	   where 
		z =  concat $ zipWith(\(a,b) (c,d) -> map (\x -> (mA st x c, map(\y -> mA st x y) d ) ) b )  hds hds'
		z' = zipWith(\(a,b) (c,d) ->(mA st a c,map(mA st a ) d )) hds hds'
		hds' = reverse hds
		st = (Sheet $ [], [])

-- Generalization Code (End) -- 

-- Utility Functions (Start) --

subset:: Eq a => [a] -> [a] -> Bool
subset as bs = if as \\ bs == [] then True else False

distribute :: Eq a => (a -> a -> a) -> [a] -> [a] -> [a]
distribute f xs ys =  [(f x y) | x <- xs, y <- ys, x /= y]

isUnique:: Eq a => [a] -> Bool
isUnique xs = (length xs) == (length $ nub xs)

pairUp :: Ord a => [a] -> [a] -> [(a,a)]
pairUp xs ys = [(x,y) | x <- xs, y <- ys, x < y]

-- Needed to compare two and bases
myInit:: Eq a => [a] -> [a]
myInit (x:xs) = if init (x:xs) == [] then [x] else init (x:xs) 
myInit xs = []

-- Utility Functions (End) --

-- Testing Stuff. Useful functions and shortcuts for testing/debugging on a console.

getUnit':: Indx -> SheetF -> Headers -> Unit
getUnit' x f h = snd $ fromMaybe (a 0,U [[["Error"]]]) (find (\(idx,u) -> idx == x) s)
		where s = drop' $ inferU h f

getIndx:: Unit -> SheetU -> Indx
getIndx u s = fst $ fromMaybe (a 0,U [[["Error"]]]) (find (\(idx,u') -> u == u') (drop' s)) 

drop':: SheetU -> [(Indx,Unit)]
drop' (Sheet u) = u

dropF:: SheetF -> [Cell Fml]
dropF (Sheet f) = f

--Returns the formula of a cell at a given Indx.
getFml:: Indx -> SheetF -> Fml
getFml i (Sheet l) = snd $ fromMaybe (a 0,UnkFml) (find (\(idx,fml) -> i == idx) l)



d7 = foldr1(mU (Sheet $ [],[],False)) [d3,d4,b4,c3,c4]
b6 = getUnit'(b 6) harvest harvestHeaders
b3 = getUnit' (b 3) harvest harvestHeaders
b4 = getUnit' (b 4) harvest harvestHeaders
b5 = getUnit' (b 5) harvest harvestHeaders
c3 = getUnit' (c 3) harvest harvestHeaders
d3 = getUnit' (d 3) harvest harvestHeaders
c4 = getUnit' (c 4) harvest harvestHeaders
c5 = getUnit' (c 5) harvest harvestHeaders
c6 = getUnit' (c 6) harvest harvestHeaders
d4 = getUnit' (d 4) harvest harvestHeaders
d5 = getUnit' (d 5) harvest harvestHeaders
d6 = getUnit' (d 6) harvest harvestHeaders
e6 = getUnit' (e 6) harvest harvestHeaders
e4 = getUnit' (e 4) harvest harvestHeaders
e5 = getUnit' (e 5) harvest harvestHeaders

hdr = getHeaderUnit harvest harvestHeaders (flipHeader harvestHeaders)
suggestions = genSugg hdr
