--
-- Indx.hs -- Spreadsheet indices, rows, columns, and vectors
--

module Sheet.Indx where

import Char (chr,toLower)


---------------------------------------------------------------------------
-- SPREADSHEET ROWS AND COLUMNS
---------------------------------------------------------------------------

type Col = Int
type Row = Int

type RowRng = (Row,Row)
type ColRng = (Col,Col)

data Indx = Indx Col Row
            deriving (Eq,Ord)

origin :: Indx
origin = Indx 1 1

instance Show Indx where
  show (Indx c r) = shwCol c++shwRow r
  
shwCol :: Int -> String
shwCol n | n<1       = map toLower (shwCol (abs n+1))
shwCol n | q==0      = [chr (r+64)]
         | otherwise = chr (q+64):shwCol (n-26*q)
           where (q,r) = n `divMod` 26

shwRow :: Int -> String
shwRow n = show (skipZero n)

skipZero :: Int -> Int
skipZero i = if i<1 then i-1 else i

col :: Indx -> Col
col (Indx c _) = c

row :: Indx -> Row
row (Indx _ r) = r


{-

-- vectors are offsets to indexes used to shift sheets
--

type Vec = (Col,Row)

hor :: Int -> Vec
hor x = (x,0)

ver :: Int -> Vec
ver y = (0,y)

[vecL, vecR, vecU, vecD,
 vecUL,vecUR,vecDL,vecDR] = [(-1,0), (1,0), (0,-1),(0,1),
                             (-1,-1),(1,-1),(-1,1),(1,1)]
toVec :: Indx -> Vec
toVec ix = (row ix,col ix)


-- The Shift class contains functions to shift/adjust columns and rows
-- that are part of a type
-- 

class Shift a where
  shift  :: Vec -> a -> a
  shift _ = id  -- default for types not containing indexes

colShift :: Shift a => Int -> a -> a
colShift c = shift (c,0)

rowShift :: Shift a => Int -> a -> a
rowShift r = shift (0,r)


instance Shift Indx where
  shift (i,j) (Indx c r) = Indx (c+i) (r+j)

-}
