--
-- FaultLoc.hs -- Fault info
-- Owner -- Scott Andrew
--

module UCheck.FaultLoc where

import UCheck.Units
import UCheck.Header
import UCheck.HdrAux
import UCheck.UnitErrors
import Sheet.Sheet
import Sheet.Expr
import Sheet.Indx
import Util.PrintList
import List
import Maybe

type DependentCells = (Int, [Indx])

-- faultLoc takes output of unitChecker and then assigns level of fault
-- this is used by shading function in VBA

faultLoc :: Sheet Fml -> ErrRpt -> Sheet ErrorInfo
faultLoc s is = faultColor is'' is
                where errIndxs = map fst is
                      is'  = fault s errIndxs errIndxs
                      is'' = Sheet (map (\x -> mergeFault x is') errIndxs)

fault :: Sheet Fml -> [Indx] -> [Indx] -> Sheet DependentCells
fault _ [] _ = Sheet []
fault s (i:is) isr = Sheet (iss' ++ unSheet (fault s is isr))
                     where Sheet iss' = expander i (s # i) isr

mergeFault :: Indx -> Sheet DependentCells -> (Indx,DependentCells)
mergeFault i is = (i,(is'',rs'))
                  where is_unsheet = unSheet is
                        is' = filter (\(ix,(n,rs))->ix == i) is_unsheet
                        is'' = sum (map (\(ix,(n,rs))->n) is')
                        rs = map (\(ix,(n,rs))->rs) is'
                        rs' = if rs /= [] then (foldr1 union rs) else []

-- faultColor takes the fault localization information as input
-- and assigns colors to the cells for display

faultColor :: Sheet DependentCells -> ErrRpt -> Sheet ErrorInfo
faultColor fs errRpt | fr == [] =  Sheet (map (\(x,(y,z)) -> (x,(6,propErr z))) fo)
                     | otherwise = Sheet ((map (\(x,(y,z)) -> (x,(6,propErr z))) fo)++(map (\(x,(y,z))-> (x, (44,localErr x errRpt))) fr)) -- [scott] i changed from z to x...what was z for?
                       where fs_unsheet = unSheet fs
                             fr = filter (\(ix,(n,rs)) -> n >= 100) fs_unsheet
                             fo = fs_unsheet \\ fr

expander :: Indx -> Maybe Fml -> [Indx] -> Sheet DependentCells
expander i (Nothing) is = Sheet [(i,(0,[]))]
expander i (Just f) is | fs' == [] = Sheet [(i,(100,fs))]
                       | otherwise = Sheet ((i,(1,fs')):(map (\x -> (x,(2,[]))) fs'))
                         where fs = aggEx f
                               fs' = intersect fs is

-- colorMap is hardcoded color mapping for Excel 
-- since RGB format in Excel produces "strange" results.

colorMap :: Int -> Color
colorMap x = case x of 
                  0 ->  36
                  1 ->  6
                  2 ->  44
                  3 ->  45
                  4 ->  46
                  5 ->  3
                  _ ->  16

assignColor :: Sheet Int -> Int -> Int -> Sheet Color
assignColor fo mx c = Sheet (map (\(ix,n) -> (ix,colorMap n) ) fo')
                      where fo' = map (\(ix,n) -> (ix, (myDiv n mx)*c)) (unSheet fo)
                               
myDiv :: Int -> Int -> Int
myDiv x y | r * 10 >= (5 * y) = d+1
          | otherwise = d 
            where d = x `div` y
                  r = x `rem` y

------------------------------------
-- Compose error messages
------------------------------------

propagate :: String
propagate = "Error in this cell possibly because of error in: "

propErr :: [Indx] -> String
propErr is = propagate ++ (concat is')
             where is' = intersperse "," (map show is)

localErr :: Indx -> ErrRpt -> String
localErr i errRpt = printErrs (fromMaybe [UnkErr] (lookup i errRpt))

printErrs :: [ErrDesc] -> String
printErrs [] = ""
printErrs (e:es) = case e of
                      (AddRng is)      -> "Suggested Change: Include " ++ printList ["",",",""] show is ++ " in range" ++ (printErrs es)
                      (UnkFill)        -> "Unable to Suggest Range Extension: Multiple Possibilities" ++ (printErrs es)
                      (ShrnkRng is)    -> "Suggested Change: Remove " ++ printList ["",",",""] show is ++ " from range" ++ (printErrs es)
                      (Omission o)     -> "Omission Error" ++ (printErrs es)
                      _                -> "Unknown Problem" ++ (printErrs es)
