--
-- MParser.hs -- Parser
-- Owner -- Robin Abraham
--

module UCheck.MParser where

import Char

import UCheck.Header
import UCheck.HdrDef
import UCheck.HdrAux
import Sheet.Sheet
import List 
import Sheet.Expr
import Sheet.Indx
import Sheet.ExcelParser
import UCheck.Units
import UCheck.UnitErrors

-- Keyword determines the message class
type Keyword = String

-- to remove RefRng from aggregations.
remRng :: Fml -> Fml
remRng (Bin b f1 f2) = Bin b (remRng f1) (remRng f2)
remRng (Agg b r) = Agg b (opener r)
remRng x = x

opener :: [Fml] -> [Fml]
opener fml = nope ++ yep
            where yep = concatMap openRng (filter isRng fml)
                  nope = filter (\x -> not (isRng x)) fml 
                  
isRng :: Fml -> Bool
isRng (RefRng _) = True
isRng (_) = False

openRng :: Fml -> [Fml]
openRng (RefRng rs) = rs

-- genRefCols builds the list of cells referenced by a range.
genRefs :: Indx -> Indx -> [Fml]
genRefs (Indx c1 r1) (Indx c2 r2) = map (Ref ) [(Indx c r)| c <- [c1..c2], r <- [r1..r2]] 

-- msgProcess calls the appropriate function depending
-- on the keyword (for Excel to Haskell)
-- recebe uma string que representa uma celula: 
-- ex
-- *Main> msgProcess "c 1 1 4 1 2 3" (Sheet [])
--      |  A  
-- -----------
--   1 | 4123
-- *Main> msgProcess "c 1 1 4 1 2 3" (Sheet [(Sheet.Indx.Indx 2 2 ,Sheet.Expr.Inp (Sheet.Expr.S "2"))]) 
--     |  A   |  B  
-- ------------------
--   1 | 4123 |     
--   2 |      | "2" 

msgProcess :: String -> Sheet Fml -> Sheet Fml
msgProcess msg s =  case k of { 
                         "c" -> cell newCell s }
                where args = strTkn msg
                      k = fst args   
                      parm = snd args
                      newCell = readMsg parm
 
-- pouco eficiente: percorre toda a ss para substituir o valor que la estava
-- pelo novo. devia haver asseco directo a celula (map?, array?, ?)
cell :: Cell a -> Sheet a -> Sheet a
cell (i, fml) (Sheet s) = Sheet ((i,fml):s')
            where s' = filter (\(i',fml') -> i' /= i) s 


-- hdrMsg generates the messages to display headers in Excel
hdrMsg :: Headers -> String
hdrMsg [] = ""
hdrMsg (l:ls) =  ms' ++ " & " ++ hdrMsg ls
               where c  = fst l
                     hs = snd l
                     ms = if hs /= [] then map (\x -> show x ++ ";" ++ show c) hs else []
                     ms' = if ms /= [] then foldr1 (++) (intersperse " & " ms) else ""

-- unitMsg generates messages for cells with unit errors
unitMsg :: [(Indx,ErrorInfo)] -> String
unitMsg bs = "UNITS: " ++ ms
            where -- is = filter (\(i,b) -> not b) bs
                  is' = map (\(ix,(c,s)) -> (show ix)++";"++(show c)++";"++s) bs
                  ms = if is' /= [] then foldr1 (++) (intersperse " & " is') else ""


-- errMsg generates messages for cells whose formulas system has not been able to parse
errMsg :: Sheet Fml -> String
errMsg (Sheet s) | s1 /= [] = foldr1 (++) (intersperse " & " s3) 
                 | otherwise = ""  
                   where s1 = filter (\(i,f) -> isUnk f) s 
                         s2 = map fst s1
                         s3 = map show s2


-- old stuff ** rewrite when time allows

strTkn :: String -> (Keyword,[String])
strTkn input = (f,inTkn s)
          where [(f,s)] = lex input
          -- = head.lex

inTkn :: String -> [String]
inTkn [] = []
inTkn str = f:(inTkn s)
         where [(f,s)] = lex str 

readMsg :: [String] -> Cell Fml
readMsg s = (Indx c r, fml)
           where c = read (head s)
                 r = read (head (tail s))
                 fstr = foldr1 (++) (intersperse "" (tail (tail s)))
		 fml = remRng (runfml formula fstr) 	

