--
-- test_server.hs -- Server code (change name to something else)
-- Owner -- Robin Abraham
--

module UCheck.UCheckServer where

import Network.Socket
import Control.Concurrent
import System (getArgs,exitFailure)
import Control.Exception(finally)
import Control.Concurrent.MVar
import IO
import UCheck.MParser
import UCheck.Header
import Sheet.Sheet
import UCheck.UnitErrors
import UCheck.Units
import UCheck.FaultLoc (faultLoc)

  {-
cols@[a,b,c,d,e] = map Indx [1..5]
refs@[ra,rb,rc,rd,re,rf] = map (Ref .) cols

harvest :: SheetF
harvest = Sheet [(b 1,is "Fruit"), 
                 (a 2,is "Month"),(b 2,is "Apple"),(c 2,is "Orange"),
                   (d 2,is "Plum"),(e 2,is "Total"), 
                 (a 3,is "May"),(b 3,ii 8),(c 3,ii 11),(d 3,ii 21),
                 (e 3,agg Mult [b 3,c 3,d 3]),
                 (a 4,is "June"),(b 4,ii 75),(c 4,ii 5),(d 4,ii 15),
                 (e 4,agg Plus [b 4,c 4,d 4]),
                 (a 5,is "July"),(b 5,ii 30),(c 5,Ref (b 3)),
                 (d 5,ii 30),(e 5,agg Plus [b 5,c 5,d 5]),
                 (a 6,Ref (d 2)),(b 6,agg Plus [b 3,b 4,b 5]),
                 (c 6,agg Plus [c 3,c 4,c 5]),(d 6,agg Plus [d 3,d 4,d 5]),
                 (b 7,Ref (c 3))]
                 -}

server_sock :: IO (Socket)
server_sock = do
    s <- socket AF_INET Stream 6
    setSocketOption s ReuseAddr 1
    --bindSocket s (SockAddrInet (PortNum portnum) iNADDR_ANY)
    bindSocket s (SockAddrInet portnum iNADDR_ANY)
    listen s 2
    return s

echo_server :: Socket -> IO ()
echo_server s = do
    (s', clientAddr) <- accept s
    proc <- read_data s' 0 (Sheet [])
    -- putStrLn ("server processed "++(show proc)++" bytes")
    sClose s'
    where
        read_data sock totalbytes sheet = do
            len <- recv sock 3
            str <- recv sock (read len) 
            case str of {

                  "kill server" -> do {
                     return totalbytes;
                     };

                  "Clear" -> do {
                     (read_data sock 0 (Sheet []));
                     };

                  "Testing" -> do { 
                     --putStrLn ("Server read: " ++ str ++ show totalbytes);
                     writ <- send sock str;
                     (read_data sock $! (totalbytes+(length $! str))) $! sheet;
                     };

                  "Headers" -> do {
                     --putStrLn ("Regions \n" ++ show (preHdrInf sheet));
                     --putStrLn ("headers: \n" ++ show (allRun sheet));
                     --putStrLn ("\nHeader \n" ++ (hdrMsg (allRun sheet))++"\n"++(errMsg sheet));
                     writ <- send sock ("HEADERS: "++(hdrMsg (allRun sheet))++"&&"++(errMsg sheet));
                     (read_data sock $! (totalbytes+(length $! str))) sheet;
                     };

                  "Units" -> do {
                     putStrLn ("Forms1 \n" ++ show (sheet));
                     putStrLn ("Units1 \n" ++ show unitSheet);
                     putStrLn ("UnitChecker1 \n" ++ show (unitChecker unitSheet unitSheet (allRun sheet) sheet) ++ "\n");
                     putStrLn ("After Refine Errors \n" ++ show (refineErrs (unitChecker unitSheet unitSheet (allRun sheet) sheet)) ++ "\n");
                     --putStrLn ("After UnitMsg \n" ++ show ((unitMsg (faultLoc sheet (refineErrs (unitChecker (inferU allRunSheet sheet) allRunSheet sheet))))));
                     writ <- send sock ((unitMsg (unSheet $ faultLoc sheet (refineErrs (unitChecker unitSheet unitSheet allRunSheet sheet)))) ++"&&"++(errMsg sheet));
                     (read_data sock $! (totalbytes+(length $! str))) sheet;
                     };

                  "Tables" -> do {
                     --putStrLn ("Tables \n" ++ show (fences sheet));
                     writ <- send sock (hdrMsg (allRun sheet));                   
                     (read_data sock $! (totalbytes+(length $! str))) sheet;
                     };

                  otherwise -> do { 
                     --putStrLn ("Server read: " ++ str ++ show totalbytes);
                     --writ <- send sock str;
                     --putStrLn ("Server wrote: " ++ str)
 	             --putStrLn (show sheet);
               -- acho k isto faz parse a uma ss
                     (read_data sock $! (totalbytes+(length $! str))) $!  (msgProcess str sheet);
                     };

            }

            where allRunSheet = allRun sheet
                  unitSheet = inferU (allRun sheet) sheet

portnum :: PortNumber
portnum = 7000


{-
main = withSocketsDo $ do {
	print("UFix version 1.0");
    ssock <- server_sock;
    -- fork off the server
    s <- myForkIO (echo_server ssock);
    join s;
    putStrLn "join s";
	};
    
 -}   
-- these are used to make the main thread wait until
-- the child threads have exited
myForkIO :: IO () -> IO (MVar ())
myForkIO io = do
    mvar <- newEmptyMVar
    forkIO (io `finally` putMVar mvar ())
    return mvar

join :: MVar () -> IO ()
join mvar = readMVar mvar
