module Data.Transform.IMDBExample where


-- problems to test/resove:
--
--  a key has 2 different foreign keys


import Data.Transform.Coupled
import Data.Transform.TwoLevelPF
import Data.Transform.RulesPF
import Data.Transform.Rewriting (many, once, top)

import Data.Map as Map
import Data.Set as Set


type Year = String
type Title = String
type Director = String
type Review = String
type Role = String
type Award = String
type ActorName = String
type Country = String
type Value = Int
type RevId = Int
type BoxId = Int

type Shows' = Map (Year, Title) ()
type Movies = Map (Year, Title) Director
type BoxOffices = Map ((Year, Title), BoxId) (Country, Value)
type Reviews =  Map ((Year, Title), RevId) Review
type Awards = Map (((Year, Title), ActorName), Role) Award
type Actors = Map ActorName ()

year = Tag "Year" String
title = Tag "Title" String
review = Tag "Review" String
director = Tag "Director" String
value = Tag "Value" Int
country = Tag "Country" String
actorName = Tag "ActorName" String
role = Tag "Role" String
award = Tag "Award" String
boxId = Tag "BoxId" Int
revId = Tag "RevId" Int

shows' = Map (Prod year title) One
movies = Map (Prod year title) director
boxOffices = Map (Prod (Prod year title) boxId) (Prod country value)
reviews = Map (Prod (Prod year title) revId) review
awards = Map (Prod (Prod (Prod year title) actorName) role) award
actors = Map actorName One


--fstdom to = COMP to (SET FST) DOM

imdb0 :: Type (Movies, Shows')
imdb0 = TInv (Prod movies shows') inv0
  where 
    inv0 = SetInc typeof (COMP typeof DOM FST) (COMP typeof DOM SND)


imdb1 :: Type (BoxOffices, (Movies, Shows'))
imdb1 = TInv (Prod boxOffices imdb0) inv1
  where
    inv1 = SetInc typeof (COMP typeof (SET FST) (COMP typeof DOM FST)) 
                         (COMP typeof DOM (COMP typeof FST SND))


imdb2 :: Type (Reviews, (BoxOffices, (Movies, Shows')))
imdb2 = TInv (Prod reviews imdb1) inv2
  where
    inv2 = SetInc typeof
                  (COMP typeof (SET FST) (COMP typeof DOM FST))
                  (COMP typeof DOM (COMP typeof FST (COMP typeof SND SND)))

imdb2' :: Type (Reviews, (BoxOffices, (Movies, Shows')))
imdb2' = TInv (Prod reviews ( (Prod boxOffices imdb0) )) inv
  where
    inv = SetInc typeof   
                 (COMP typeof (SET FST) (COMP typeof DOM FST))
                 (COMP typeof DOM (COMP typeof FST (COMP typeof SND SND)))
   
    --inv1 = SetInc typeof (COMP typeof (SET FST) (COMP typeof DOM FST)) 
      --                   (COMP typeof DOM (COMP typeof FST SND))
 

imdb3 :: Type (Awards, (Reviews, (BoxOffices, (Movies, Shows'))))                  
imdb3 = TInv (Prod awards imdb2) inv3
  where
    inv3 = SetInc typeof
                  (COMP typeof (SET (COMP typeof FST FST)) (COMP typeof DOM FST))
                  (COMP typeof DOM (COMP typeof FST (COMP typeof SND (COMP typeof SND SND))))

imdb4 :: Type (Actors, (Awards, (Reviews, (BoxOffices, (Movies, Shows')))))
imdb4 = TInv (Prod actors imdb3) inv4
  where
    inv4 = SetInc typeof
                  (COMP typeof (SET (COMP typeof SND (COMP typeof FST FST))) (COMP typeof DOM (COMP typeof FST SND)))
                  (COMP typeof DOM FST)


imdbActor :: Type (Actors, Awards)
imdbActor = TInv (Prod actors awards) inv
  where
    inv = SetInc typeof
                  (COMP typeof (SET (COMP typeof SND (COMP typeof FST FST))) (COMP typeof DOM SND))
                  (COMP typeof ID (COMP typeof DOM FST))



iex1 = (mex, bex)


mex :: Movies 
mex = Map.fromList [    (("2008", "Sweeny Tod"), "John"),
                        (("2007", "zoodiac"), "john"), 
                        (("2007", "predator"), "tarantino") ]

bex :: BoxOffices
bex = Map.fromList [ ((("2007", "zoodiac"), 1), ("PT", 712)), 
                     ((("2007", "predator"), 2), ("PT", 12))
                    ]

-- example that doesn't respect the FD
--
rdb1 = (mex1, bex1) 

mex1 = [ ((2008, "Sweeny Tod"), "John"),
         ((2007, "zoodiac"), "john"), 
         ((2007, "predator"), "tarantino") ]

bex1 = [ (((2007, "predator"), 2), ("PL", 123)), 
         (((2007, "predator"), 2), ("PT", 12)) ]
                    

{-
migration = 
    do let evolve = addTracks
	   tordb = removetags >>> flatten
           (Just vw1) = tordb albums
	   (Just vw2) = (evolve >>> tordb) albums
       putStrLn $ "\nrdb1 = " ++ showType vw1
       putStrLn $ "\nrdb2 = " ++ showType vw2   
       let (Just (to1,fro1)) = unView vw1 rdb
	   (Just (to2,fro2)) = unView vw2 rdb'
	   to = COMP typeof to2 fro1
	   iso1 = COMP typeof fro1 to1
	   iso2 = COMP typeof fro2 to2
	   inv1 = COMP typeof to1 fro1
	   inv2 = COMP typeof to2 fro2
       putStrLn $ "\nto  = "++show to
       simplify False optimize to
       putStrLn $ "\niso1  = "++show iso1
       simplify False optimize iso1
       putStrLn $ "\niso2  = "++show iso2
       simplify False optimize iso2
       putStrLn $ "\ninv1  = "++show inv1
       simplify False optimize inv1
       putStrLn $ "\ninv2  = "++show inv2
       simplify False optimize inv2
-}


{-
imdbTest = do
  putStrLn ">>> START TEST <<<"

  let Just vw = rdb2ss imdb1
  let b = director
      b' = Prod country value
  let vwt = (Prod (TInv (List (Prod (Prod (Tag "Year" Int) (Tag "Title" String)) (Tag "Director" String))) (ListInc (Prod b b) CompList ListId)) (TInv (List (Prod (Prod (Prod (Tag "Year" Int) (Tag "Title" String)) (Tag "Id'" Int)) (Prod (Tag "Country" String) (Tag "Value" Int)))) (ListInc (Prod b' b') CompList ListId)))
  let Just (to1, from1) = unView vw vwt
  let tod = eval to1 iex1
  let fromd = eval from1 rdb1 -- tod
  putStrLn "from rdb to ss"
  putStrLn $ show tod
  putStrLn "\nfrom ss to rdb"
  putStrLn $ show fromd

  -- Running the type-level transformation to obtain target format as String.
  let (Just vw) = rdb2ss imdb1
  let targetFormatAsString = showType vw
  putStrLn $ {-snip $ -} "Target format: " ++ targetFormatAsString
  if targetFormatAsString == show 'a' -- intendedTargetFormat
    then putStrLn "Type-level evolution OK"
    else putStrLn "fail" -- fail "Type-level evolution FAILED"
  {-
  -- Then running the forward value-level transformation.
  let cv = Prod country value
  let targetFormat = (Prod (TInv (List (Prod (Prod (Tag "Year" Int) (Tag "Title" String)) (Tag "Director" String))) (ListInc ((Prod director director)) CompList ListId)) (TInv (List (Prod (Prod (Prod (Tag "Year" Int) (Tag "Title" String)) (Tag "Id'" Int)) (Prod (Tag "Country" String) (Tag "Value" Int)))) (ListInc ((Prod cv cv)) CompList ListId)))
  let (Just targetAlbum) = forth vw targetFormat iex1
  putStrLn $ gshow targetFormat targetAlbum
 -- if targetAlbum == "i" -- targetLP
   -- then putStrLn "Forward migration OK"
    --else fail "Forward migration FAILED"

 -- And the backward value-level transformation.
  let (Just originalAlbum) = back vw targetFormat targetAlbum
  putStrLn $ snip $ gshow albumFormat originalAlbum
  if lp==originalAlbum
    then putStrLn $ "Backward migration OK"
    else fail "Backward migration FAILED"
    
  -- Running the type-level data mapping on both hierarchical formats.
  let toRDB = once enum2int >>> removetags >>> flatten
  let (Just vw0) = toRDB (List albumFormat)
  putStrLn $ snip $ showType vw0
  let (Just vw1) = toRDB (List targetFormat)
  putStrLn $ snip $ showType vw1
  if (showType vw0)==(show intendedTargetDBS0) && (showType vw1)==(show intendedTargetDBS1)
    then putStrLn "Type-level data mapping OK"
    else fail "Type-level data mapping FAILED"
-}
  -- Running the forward value-level data mapping on two albums.
  --let dbs0 = Map Int (Prod String (Prod String (Prod String Int)))
  let cv = Prod country value
  let targetFormat = (Prod (TInv (List (Prod (Prod (Tag "Year" Int) (Tag "Title" String)) (Tag "Director" String))) (ListInc ((Prod director director)) CompList ListId)) (TInv (List (Prod (Prod (Prod (Tag "Year" Int) (Tag "Title" String)) (Tag "Id'" Int)) (Prod (Tag "Country" String) (Tag "Value" Int)))) (ListInc ((Prod cv cv)) CompList ListId)))
  let (Just db) = forth vw targetFormat iex1
  putStrLn $ snip $ show db
  putStrLn "Forward value-level data mapping OK (it seems)"
  {-
  -- Migrate from and old database to an update one.\begin{code}
  let (Just lvw) = (addTracks >>> addDvd) (List albumFormat)
  let dbs1 = Prod (Map Int (Prod String (Prod String (Prod String Int)))) (Map (Prod Int Int) String)
  let (Just x) = back vw0 dbs0 db 
  let (Just y) = forth lvw (List targetFormat) x
  let (Just z) = forth vw1 dbs1 y
  putStrLn $ snip $ show z
  putStrLn "Value-level data base migration OK (it seems)"

  putStrLn ">>> TEST SUCCESSFUL <<<"
 -}

snip :: String -> String
snip x = "  "++(take 45 x)++" ... "++(takeLast 25 x)
  where
    takeLast n = reverse . (take n) . reverse

-}


-- | addock tests


i1 :: PF (Pred (([(Char, (Char, Char))], ([(Char, Char)], [(Char, (Char, Maybe Char))]))))
i1 = SetInc typeof (COMP (List Char) LIST2SET (COMP typeof (LIST FST) FST))
                   (COMP typeof DOM (COMP typeof FST (COMP (Prod (Map Char Char) (Prod (Map Char Char) (Map Char Char))) SND (COMP typeof (PROD ID Map2maps) (COMP typeof (PROD Table2map Table2map) SND)))))

-- ((list2set . ((list fst) . fst))) <= (dom . (fst . (snd . ((id >< map2maps) . ((table2map >< table2map) . snd)))))



{-
t0 :: PF (([((Char, Char),Char)], [(Char, (Char, Maybe Char))]) -> Map Char Char)
t0 = COMP typeof FST (COMP typeof SND (COMP typeof (PROD ID Map2maps) Tables2maps')) 
-- (fst . (snd . ((f >< g) . x)))

t0' :: PF ((Char, ([((Char, Char), Char)], [(Char, (Char, Maybe Char))])) -> (Map Char Char, Map Char Char))
t0' = COMP typeof SND (COMP typeof (PROD ID Map2maps) (COMP typeof Tables2maps' SND)) 
--snd . ((id >< maptomaps) . (tablestomaps' . snd))


t1 :: PF (Pred (Map (Char, Int) String, Map Char (Bool, Maybe Bool)))
t1 = SetInc typeof 
            (COMP typeof (COMP typeof (SET FST) DOM) FST)
            (COMP typeof (COMP typeof (COMP typeof DOM FST) SND) (SPLIT FST (COMP typeof Map2maps SND)))
-- (((((set fst) . dom) . fst) =< ((dom . fst) . snd)) . (fst /\ (maptomaps . snd)))


t1' :: PF ((String, Map Char (Bool, Maybe Bool)) -> Set Char )
t1' = (COMP typeof (COMP typeof (COMP typeof DOM FST) SND) (SPLIT FST (COMP typeof Map2maps SND)))
-- ((dom . fst) . snd)) . (fst /\ (maptomaps . snd)))))


t2 :: PF ((String, (Map Char (Bool, Maybe Bool))) -> Set Char)
t2 = COMP typeof (COMP typeof (COMP typeof DOM FST) SND) (PROD ID Map2maps)
-- ((dom . fst) . snd)) . (id >< maptomaps)))


t3' :: PF ((Int, Map Char (Int, Maybe Int)) -> Set Char)
t3' = (COMP typeof (COMP typeof DOM FST) (COMP typeof Map2maps SND))
-- ((dom . fst) . (maptomaps . snd)))


t3 :: PF (Pred (Map (Char, Int) Int, Map Char (Int, Maybe Int)))
t3 = SetInc typeof 
            (COMP typeof (SET FST) (COMP typeof DOM FST))
            (COMP typeof (COMP typeof DOM FST) (COMP typeof Map2maps SND))
-- (((set fst) . (dom . fst)) =< ((dom . fst) . (maptomaps . snd)))


t5 :: PF (Pred (Map (Char, Int) Int, Map Char (Int, Maybe Int)))
t5 = SetInc typeof 
            (COMP typeof (COMP typeof (SET FST) DOM) FST)
            (COMP typeof (COMP typeof (COMP typeof DOM FST) SND) (PROD ID (COMP typeof (COMP typeof ID Map2maps) ID)))
-- (((((set fst) . dom) . fst) =< ((dom . fst) . snd)) . (id >< ((id . maptomaps) . id)))

-}

-- to test map2table
ac = TInv (Map String String) (SetInc typeof DOM RNG)

-- to test maps2map
ss1 = TInv (Prod shows' shows') (SetInc typeof (COMP typeof DOM SND) (COMP typeof DOM FST))

--ss2 = TInv (Prod shows' movies) (SetInc typeof (COMP typeof DOM FST)
  --                                             (COMP typeof (COMP typeof (SET FST) DOM) SND))

--ss3 = TInv (Prod movies shows') (SetInc typeof (COMP typeof DOM SND)
  --                                             (COMP typeof (COMP typeof (SET FST) DOM) FST)) 
                                                

-- to test maps2tables
--mt1 = TInv (Prod shows' movies) (SetInc typeof (COMP typeof (COMP typeof (SET FST) DOM) SND) 
  --                                             (COMP typeof DOM FST))

--mt2 = TInv (Prod movies movies) (SetInc typeof (COMP typeof (COMP typeof (SET FST) DOM) SND) 
  --                                             (COMP typeof (COMP typeof (SET FST) DOM) FST))

