module CFDSCare where


import Types

import Data.List as List
import Data.Map as Map





-- ======================================== REDUNDANT

-- ok
-- computes nonredundant covers
-- alg 5.4 da 73, changed to work with CFDS
-- this doesn't work if the CFD dosn't have right side
--
nonRedunCFD :: Eq a => CFDS a -> CFDS a
nonRedunCFD fds = shiftSet $ nonRedun' fds fds
   -- where
nonRedun' [] f = f
nonRedun' (xy:gs) f = if memberCFD (List.delete xy f) xy
                                    then nonRedun' gs (List.delete xy f)
                                    else nonRedun' gs f

-- tests nonRedunCFD
--
-- pag 89, ex 5.17
g' = [([Atts [Att "a"], Atts [Att "a", Att "b"], Atts [Att "b"]],Atts [Att "c", Att "d"]), 
      ([Atts [Att "a", Att "e"]], Atts [Att "i", Att "j"])]

-- pag 89, ex 5.18
g'' = [([Atts [Att "a"], Atts [Att "b"]],Atts [Att "a", Att "b", Att "c", Att "d"]), 
        ([Atts [Att "a", Att "e"]], Atts [Att "i", Att "j"])]

g3 = [([Atts [Att "a", Att "b"], Atts [Att "b"]],Atts [Att "a", Att "c", Att "d"]), 
       ([Atts [Att "a", Att "e"]], Atts [Att "i", Att "j"])]



-- ============================================= CLOSURE

closureCFD :: Eq a => Attributes a -> CFDS a -> Attributes a
closureCFD ats fds = closure' fds (Atts []) ats
  where
    closure' fds olddep newdep = if newdep /= olddep
                                            then let newdep' = aux fds newdep
                                                 in closure' fds newdep newdep'
                                            else newdep
    aux [] newdep = newdep
    aux ((a, c):fds) newdep = 
        if  or (List.map (<| newdep) a) 
          then aux fds (Atts $ (foldr List.union (atts c) (List.map atts a)) `List.union` 
                                                     (atts newdep)) 
          else aux fds newdep



-- to CFDS
-- need tests
--
linClosure :: Ord a => Attributes a -> CFDS a -> Attributes a 
linClosure x fs
   =
    let (list, count) = inicialization fs
        aux = fromList fs
    in computation aux x list x x count

inicialization :: Ord a => CFDS a -> (Map (Attribute a) (LeftSets a), Map (LeftSets a) Int)
inicialization (f:fs) 
  =
    let count = createMap (f:fs)
        list = createList (f:fs)
    in (list, count)
  where
    createMap [] = empty
    createMap (f:fs) = Map.insert (fst f) (length $ fst f) (createMap fs)

    createList [] = empty
    createList (f:fs) = Map.unionWith (++) (createListaux (fst f) (fst f)) (createList fs)

    createListaux _ [] = empty
    createListaux f (h:t) = unionWith (++) (createListaux2 f (atts h)) (createListaux f t)
   
    createListaux2 _ [] = empty
    createListaux2 f (h:t) = Map.insertWith (++) h f (createListaux2 f t)
    

computation :: Ord a => Map (LeftSets a) (Attributes a) ->
                        Attributes a -> Map (Attribute a) [Attributes a] -> 
                        Attributes a -> Attributes a -> Map [Attributes a] Int ->
                        Attributes a 

computation _ (Atts []) _ newdep _ _ = newdep
computation aux (Atts (up:ups)) list newdep updat count
  =  -- choose A in update
        let fd = list Map.! up -- up :: Attributes a, fd :: [Attributes a]
            (newdep', updat', count') = compaux2 aux fd (newdep, Atts $ List.delete up (atts updat), count)
        in computation aux updat' list newdep' updat' count'

  -- for each FD in list A
  
compaux2 :: Ord a => Map (LeftSets a) (Attributes a) -> LeftSets a -> 
                     (Attributes a, Attributes a, Map (LeftSets a) Int) ->
                     (Attributes a, Attributes a, Map (LeftSets a) Int)
                     
compaux2 _ [] t = t 
compaux2 aux (fd:fds) (newdep, updat, count)
      = 
        let c = count Map.! [fd]
        in if (c-1) == 0
              then
                let add = List.union (atts fd) (atts $ aux Map.! [fd])
                    addup = (atts $ aux Map.! [fd]) List.\\ (atts newdep)
              -- (newdep, update, count)
                in compaux2 aux fds
                   ( Atts $ List.union (atts newdep) add, 
                     Atts $ List.union (atts updat) addup,
                     update (\x -> Just (x-1)) [fd] count)
              else
                compaux2 aux fds (newdep, updat, Map.update (\x -> Just (x-1)) [fd] count)



-- =========================================== SHIFT

-- shift
--
shift :: Eq a => CFDS a -> CFDS a
shift cfd = aux cfd cfd
  where
    aux cfds [] = cfds
    aux alcf (cfd:cfds) = let cf = aux2 alcf cfd (fst cfd)
                          in aux cf cfds

    aux2 cfds _ [] = cfds
    aux2 cfds (a, c) (at:ats) = let (cf, fd) = aux3 cfds (a, c) at at
                                in aux2 cf fd ats 

    aux3 cfds (a, c) _ (Atts []) = (cfds, (a, c))
    aux3 cfds (a, c) (Atts at) (Atts (x:xs))
      = 
        let na = Atts $ List.delete x at
            nf1 = if na /= (Atts []) then subs a (Atts at) na else List.delete (Atts at) a
            nf = (nf1, Atts $ (atts c) `List.union` [x])
            nfs = subs cfds (a, c) nf
        in 
          if equivCFD cfds nfs                                 
             then aux3 nfs nf na na
             else aux3 cfds (a, c) (Atts at) (Atts xs)

  -- substitui b por c em a
    subs a b c = c:(List.delete b a)



-- shift tests
--
sh1 = [([Atts [Att "a"], Atts [Att "a", Att "b"], Atts [Att "b"]],Atts [Att "c", Att "d"]),
       ([Atts [Att "a", Att "e"]], Atts [Att "i", Att "j"])]

sh1' = [([Atts [Att "a"], Atts [Att "b"]],Atts [Att "c", Att "d", Att "a", Att "b"]),
       ([Atts [Att "a", Att "e"]], Atts [Att "i", Att "j"])]

sh1'' = [([Atts [Att "a", Att "b"], Atts [Att "b"]],Atts [Att "c", Att "d", Att "a"]),
         ([Atts [Att "a", Att "e"]], Atts [Att "i", Att "j"])]


-- seems not ok g'' 89
-- shift sets to nonredun
--
shiftSet :: Eq a => CFDS a -> CFDS a
shiftSet cfd = aux cfd cfd
  where
    aux cfds [] = cfds
    aux alcf (cfd:cfds) = let cf = aux2 alcf cfd (fst cfd)
                          in aux cf cfds

    aux2 cfds _ [] = cfds
    aux2 cfds (a, c) (at:ats) 
      = 
        let nf = (List.delete at a, Atts $ (atts c) `List.union` (atts at))
            nfs = List.delete (a, c) cfds
        in if equivCFD cfds (nf:nfs)                                  
             then aux2 (nf:nfs) nf (fst nf)
             else aux2 cfds (a, c) ats



-- ========================================= REDUCE

reduceCFD :: Eq a => CFDS a -> CFDS a
reduceCFD cfd = rightRedCFD $ shift cfd


rightRedCFD :: Eq a => CFDS a -> CFDS a
rightRedCFD fds = aux1 fds fds
  where
    aux1 [] fds = fds
    aux1 ((a, c):fds) fds' 
      =     
        if (foldr1 List.union (List.map atts a)) `intersect` (atts c) == [] -- if it's already reduced
          then (a,c):(aux1 fds (List.delete (a, c) fds')) -- doesn't reduce again
          else
            let fds'' = aux2 fds' (a, c) c
            in aux1 fds fds''

    aux2 f _ (Atts []) = f
    aux2 f (x, Atts alys) (Atts (y:ys)) = 
        if (memberCFD ((List.delete (x, Atts alys) f) `List.union` 
                                    [(x, Atts $ List.delete y alys)]) (x, Atts [y]))
                                              then 
                                                let alys' = Atts $ List.delete y alys
                                                in aux2 (List.delete (x, Atts alys) f `List.union` 
                                                          [(x, alys')]) (x, alys') (Atts ys)
                                              else
                                                aux2 f (x, Atts alys) (Atts ys)



-- =========================================== derives, equiv, ...

-- a -> b e c -> d => ac -> db
-- I'll restrict to the fds having the same consequent or the same antecedent
--
composecfd :: (Eq a, Ord a) => CFDS a -> CFDS a
composecfd fds 
  =
    let fds' = concatMap join $ groupBy group1 $ sortBy sort2 fds
    in fds' -- map join fds'
  where
    join [f] = [f]
    join ((a, c):r) = if c /= (Atts [])
                          then
                            let [(a', c')] = join r
                            in [(a `List.union` a', c)]
                          else 
                            ((a, c):r)



equivCFD :: Eq a => CFDS a -> CFDS a -> Bool
equivCFD f1 f2 
  = 
    derivesCFD f1 f2 && derivesCFD f2 f1
    

--pag 72, alg 5.1
--
derivesCFD :: Eq a => CFDS a -> CFDS a -> Bool
derivesCFD f1 f2 
  = 
    and $ List.map (memberCFD f1) f2


memberCFD :: Eq a => CFDS a -> CFD a -> Bool
memberCFD fds (a, c) = (Atts $ (concatMap atts a) `List.union` (atts c)) <| 
                         (closureCFD (Atts $ foldr List.union [] (List.map atts a)) fds)


-- verifies if x <-> y (x and y are equivalent) in a set of fd
-- simes ok
--equivalentCFD :: Eq a => Attributes a -> Attributes a -> CFDS a -> Bool
--equivalentCFD x y fds = memberCFD fds (x, y) && memberCFD fds (y,x)




