module FDCare where


import HaExcel.FDS2Type
import Types
import Combinators
import FDInference
import Synthesize hiding (memberFD, closureS)

import Data.List as List
import Data.Map (fromList, Map)
import Data.Maybe
import System.IO.Unsafe
import Debug.Trace





--cfds2rdb :: (Show a, Eq a) => [FDS a] -> CFDS a -> [FDS a]
cfds2rdb [(a, Atts c)] = tordb a a (Atts c) -- (Atts $ c `union` (foldr1 union (map atts a)))
cfds2rdb ((a, Atts c):t)
  =
    let res = cfds2rdb t
        tab = tordb a a (Atts c) -- (Atts $ c `union` (foldr1 union (map atts a)))
    in [s ++ v | v <- tab, s <- res]

tordb :: Eq a => [Attributes a] -> [Attributes a] -> Attributes a -> [FDS a] 
tordb _ [] _ = []
tordb alas (a:as) alatts 
  = 
    [(a, -- Atts (foldr union (atts alatts) (map atts alas)))] --Atts $ filter (`notElem` (atts a)) (foldr1 union (map atts alas)))]
       Atts $ filter (`notElem` (atts a)) $ foldr union (atts alatts) (map atts $ delete a alas))] 
                        :(tordb alas as alatts)


-- | receives the result of the synthesize and returns a single set of FDs
--   chooses as keys the smalest candidate
--
--cfds2rdb' :: CFDS a -> FDS a
cfds2rdb' cfds = -- map (\(a, c) -> head $ sortBy smal $ concat $ tordb a a c) cfds
                  map (\(a, c) -> concat $ tordb a a c) cfds
  where
    smal (Atts l1, Atts _) (Atts l2, Atts _) | length l1 <= length l2 = LT
    smal _ _   | otherwise = GT


-- | removes the FDs that have attributes already represented
--   starts from the end because is where are bad FDs
--
remBad :: Eq a => FDS a -> FDS a
remBad l = aux2 [] l l
  where
    aux2 ac _ [] = ac
    aux2 ac al (h:t) = if (auxx h (delete h al))
                         then aux2 ac (delete h al) t
                         else aux2 (h:ac) al t

auxx (Atts a, Atts c) l = (a ++ c)  <|| givall l

givall l = concatMap ((uncurry (++)).(atts >< atts)) l


rel2rdb r rl = let fs = fun r rl >>= lks2fds
               in fds2type $ rel2rdb1 fs r rl


-- |  seems a good version!
--
--rel2rdb :: (Ord a, ToInt a, Eq b, Show a) => R a -> Relation b -> [FDS a]
rel2rdb1 l imdbSch imdbRel
  = 
           let -- l = fun imdbSch imdbRel >>= lks2fds
              -- need to remove those FDs that have columns with formulas as keys
              -- because PKs can't change
               l' = remForms l
           -- this 3 lines down create a FD with all the non-used attr
           -- the key will be a integer auto incremented and the values are
           -- all the atts not used
               allAttsInFDS = collectAllAtts l'
               attsNotInFDS = (atts imdbSch) List.\\ allAttsInFDS
               newFD = if attsNotInFDS /= [] then [(Atts [Att "reserved"], Atts attsNotInFDS)]
                                             else []
               f = synthesize l'
               g = cfds2rdb' f
           in  newFD ++ (compose $ remBad . reverse $ (map (last.((List.sortBy rat).reverse)) g))
  where
rat (Atts a, Atts c) (Atts a', Atts c') | (dic a c) < (dic a' c') = GT
                                            | (dic a c) > (dic a' c') = LT
                                            | otherwise = EQ

dic a b = (realToFrac (length a)) / (realToFrac (length b))                   

collectAllAtts = foldr (\(Atts a, Atts c) res -> a ++ c ++ res) []

    -- filters FDs with any att in the antecedent being prefixed with "FORM"
remForms = filter (\(Atts a, c) -> not $ or $ map ((isPrefixOf "FORM") . att) a)


-- | final FDs clean up. if a -> b, c and d -> c, then we should have a -> b, d and d -> c 
--    with a foreign key inv.
--    write this function


{-
*Tests> showfds $ remBad $ rel2rdb imdbSch imdbRel
([Att "award",Att "role"],[Att "actor"])
([Att "value",Att "role"],[Att "actor"])
([Att "country",Att "role"],[Att "actor"])
([Att "title",Att "year"],[Att "review",Att "director"])

*Tests> showfds $ compose $ remBad $ rel2rdb imdbSch imdbRel
([Att "title",Att "year"],[Att "review",Att "director"])
([Att "award",Att "role",Att "value",Att "country"],[Att "actor"])

*Tests> showfds $ remBad $ compose $ rel2rdb imdbSch imdbRel
([Att "title",Att "award"],[Att "director"])
([Att "award",Att "role",Att "director",Att "value",Att "country",Att "year",Att "review"],[Att "actor"])

*Tests> showfds $ compose $ rel2rdb imdbSch imdbRel
([Att "title",Att "award"],[Att "director"])
([Att "title",Att "year"],[Att "review",Att "director"])
([Att "year",Att "director"],[Att "title",Att "review"])
([Att "title",Att "review"],[Att "year",Att "director"])
([Att "title",Att "director"],[Att "year",Att "review"])
([Att "director",Att "actor",Att "year",Att "title"],[Att "award"])
([Att "award",Att "role",Att "director",Att "value",Att "country",Att "year",Att "review"],[Att "actor"])
([Att "year",Att "country",Att "value",Att "actor",Att "role",Att "award",Att "director"],[Att "review"])
([Att "country",Att "director",Att "review",Att "actor",Att "award",Att "year",Att "title"],[Att "value"])
-}

-- for a RDB finds the foreign keys on it
--
--findFK :: (Eq b, Ord a) => RDB a b -> [Map (Attribute a) [b]] -- [((FD a, Attribute a), (FD a, Attribute a))]
findFK r 
  =
    fk1 r r
  where
    fk1 _ [] = []
    fk1 r (((a, c), rl):t)  
      = 
        let ig = filter ((isIn a).fst.fst) (delete ((a, c), rl) r)  
        in if ig /= []
             then (fk2 ig ((a, c), rl)):(fk1 r t)
             else fk1 r t

    fk2 r1 ((a, c), rl) = map (fk3 ((a, c), rl)) r1

    fk3 ((Atts a, c), rl) ((Atts a', c'), rl') 
      = 
        let int = intersect a a'
            inds = map (fromJust.(flip elemIndex a)) int
            inds' = map (fromJust.(flip elemIndex a')) int
            rls = map (projection rl) inds
            rls' = map (projection rl') inds'
            cont = zipWith (<||) rls rls'
            cont' = zipWith (<||) rls' rls
        in (cont, cont')
              
    isIn (Atts a) (Atts a') = or $ map (flip elem a') a




-- given a set of fds returns the set of all possible db
--
allDBs :: Eq a => R a -> FDS a -> [FDS a]
allDBs (Atts r) = (filter ((== (length r)).length.(foldr union []).(map ((uncurry union).(atts >< atts))))).powerset' 



showl :: Show t => [[t]] -> IO [()]
showl =  mapM (\c -> showfds c >> (putStr "\n")) 







-- new idea

-- receives the relation schema, the relation and returns a set of FDs representing the new database
-- the idea is to remove all the FDs that doesn't agree with the 2NF and the 3BF
-- can't do this because I can end up with a set that is not equivalent to the start one
-- in one example I finished with less att than the original relation schema had!!
newId :: (Monad m, Eq b, Eq a, Show a) => R a -> Relation b -> m (FDS a)
newId r rl =
  do
    lks <- fun r rl 
    let fds' = (lks2fds lks)         
    -- remove the super keys and order by the biggest first
        fds = fds' -- sortBy sort1 $ fds' \\ (giveSupKeys r fds')
        fds2nf = filternf nf2 fds fds
        fds3nf = filternf nf3 fds2nf fds2nf
    return fds3nf
   
filternf :: Eq a => (FDS a -> FD a -> Bool) -> FDS a -> FDS a -> FDS a
filternf _ fs [] = fs
filternf nf fs (fd:fds) = if nf fs fd then filternf nf fs fds else filternf nf (delete fd fs) fds


nf23 :: Eq a => FDS a -> FDS a 
nf23 fds = let fds2nf = filternf nf2 fds fds
           in  filternf nf3 fds2nf fds2nf



-- tests if a FD respects the 2NF
--
nf2 :: Eq a => FDS a -> FD a -> Bool
nf2 fds (a, c) 
  =
    if length (atts a) > 1
      then
        let pa = (powerset a) \\ [Atts [], a]
            pc = (powerset c) \\ [Atts [], c]
            papc = [(a, c) | a <- pa, c <- pc]
        in not $ or $ map (memberFD fds) papc
      else 
        True


-- tests if a FD respects the 3NF
--
nf3 :: Eq a => FDS a -> FD a -> Bool
nf3 fds (a, c) 
  =
    if length (atts c) > 1
      then
        let pc = (powerset c) \\ [Atts [], c]
            pcfd = [(a, c) | a <- pc, c <- pc, a /= c]
        in not $ or $ map (memberFD fds) pcfd
      else
        True

-- returns teh FDs that are superkeys
--
giveSupKeys :: R a -> FDS a -> FDS a
giveSupKeys (Atts r) = filter (\(Atts a, Atts c) -> length a + length c == length r)



--
nfds r rl =
  do
    x <- mapM ((flip (fun r)) 100) (fuu 20 rl)
    return x

    
-- splits a list into lists with length 10
--
fu :: [a] -> [[a]]    
fu l | length l >= 20 = ((uncurry (:)).(id >< fu).(splitAt 10)) l
     | otherwise = [l] 


fuu :: Int -> [a] -> [[a]]
fuu n l | length l >= 20 = let (h, t) = splitAt n l
                           in h:(fuu (n+20) t)
        | otherwise = [l]

------------ trying to eliminate FDs which the consequent is less then the antecedent
--

-- gives the FDs where |consequent| <= |antecedent|
takeless :: FDS a -> FDS a
takeless = filter (\(Atts x, Atts y) -> length y <= length x)

discardless :: Eq a => FDS a -> FDS a
discardless fds = 
    let less = takeless fds
    -- comeamos por descartar aqueles que menos nos interessam
        rest = rem fds less 
    -- depois descartamos todos os outros
    in rem rest rest
    where
        rem fds [] = fds
        rem fds (fd:fd') = if memberFD (delete fd fds) fd 
                                then rem (delete fd fds) fd'  
                                else rem fds fd'

-- | ************************************* *******************
--
-- verifies if a given fd holds in a set of fds
memberFD :: Eq a => FDS a -> FD a -> Bool
memberFD fds (a, c) = a <| (closureS c fds)


-- a weak version, but working
closureS :: Eq b => Attributes b -> FDS b -> Attributes b
closureS ats fds = closure' fds (Atts []) ats
    where
        closure' fds olddep newdep = if newdep /= olddep
                                            then let newdep' = aux fds newdep
                                                 in closure' fds newdep newdep'
                                            else newdep

        aux [] newdep = newdep
        aux (fd:fds) newdep = if fst fd <| newdep
                                then aux fds (Atts $ (atts newdep) `List.union` ((atts.snd) fd))
                                else aux fds newdep


-- sera k o algoritmo ja tem isto em conat???
-- verificamos se as novas FDs l podem ser inferidas a 
-- partir das ja existente em fd, se sim descartam-se
-- recebemos as FDS e os LKS e retornamos o LKS correctos
--
ismember [] _ _ fds = fds
ismember (l:ls) fd (l_:ls_) fd_ = if memberFD fd l 
                                then ismember ls fd ls_ fd_ 
                                else ismember ls fd ls_ (l_:fd_)  





-- remove redundat FDs
--
elimRed :: Eq a => FDS a -> FDS a
elimRed fds = filter (\fd -> not $ memberFD (delete fd fds) fd) fds

-- remove ab -> a  
--
elimIn :: Eq a => FDS a -> FDS a
elimIn = filter (\(a, c) -> not $ c <| a )

-- test
in1 = [ ([1], [1, 2]), ([1], [4]), ([2, 3], [4]), ([3, 4], [4]), ([3], [5]), ([4], [2]) ]

-- a -> ab => a -> b
--
elimMin :: Eq a => FDS a -> FDS a
elimMin = map rem
  where
    rem (Atts a, Atts c) = (Atts a, Atts $ filter (`notElem` a) c)



-- a -> b e c -> d => ac -> db
-- I'll restrict to the fds having the same consequent or the same antecedent
--
compose :: (Eq a, Ord a) => FDS a -> FDS a
compose fds 
  =
    let fds' = sortBy sort3 $ map join $ groupBy group1 $ sortBy sort2 fds
    in fds' -- map join fds'
  where
    join [f] = f
    join ((Atts a, c):r) = let (Atts a', c') = join r
                           in (Atts $ a `union` a', c)



--
-- a -> c and b -> c can be eliminated creating ab -> c
-- a -> c && b -> c => ab ->c 
--
fdunion :: Eq a => Int -> FDS a -> FDS a 
fdunion 0 fds = fds 
fdunion n t = let t' = aux t
              in fdunion (n-1) t'
  where
    aux [x] = [x]
    aux ((a, c):(a1, c1):t) | c == c1 = (Atts $ (atts a) `union` (atts a1), c1):t 
                            | otherwise = (a1, c1):(aux ((a, c):t)) 

-- test
un1 = [ ([1], [1, 2]), ([1], [4]), ([2, 3], [4]), ([3, 4], [4]), ([3], [5]), ([4], [2]) ]


-- a -> c && b -> c => ab -> c
-- we remove ab -> c because has a bigger key
--
desunion :: Eq a => FDS a -> FDS a
desunion fds = 
              let x = groupBy (\(a, c) (a', c') -> c == c') fds
                  y = concatMap (\l -> aux2 l l) x
              in y
  where
    aux2 [] x = x
    aux2 (h:t) x = 
                  let xh = delete h x
                      z = aux1 xh
                  in if h `elem` z then aux2 t xh else aux2 t x
--
-- combines the elems in the list
-- aux1 [(1,7), (9, 7), (5, 7)] = [([1, 9], 7), ([1, 5], 7), ([9, 5], 7)]
--     
    aux1 [] = []
    aux1 ((Atts a, c):t) = (map (\(Atts a', _) -> (Atts $ a `union` a', c)) t)++(aux1 t)


-- test
--
des2 = (map (Atts >< Atts)) [([Att 1],[Att 4]),([Att 2,Att 3],[Att 4]),([Att 3,Att 4],[Att 4])]
                     
des1 = map (\(a, b) -> (Atts $ map Att a, Atts $ map Att b))  
          [ ([1], [1, 2]), ([4], [1, 2]), ([4, 1], [1, 2]) , ([4], [4]), ([3], [4]), ([3, 4], [4]), ([3], [5])]


-- ******************************** Armstrong theorems *****************
--
-- versao 1
{-
armstrong1 :: Ord b => Int -> Int -> FDS b -> FDS b
armstrong1 n1 n2 fds = 
    let fds' = t6 n1 $ t5 n1 fds
    in if fds' /= fds && n2 > 0 
            then armstrong1 n1 (n2 -1) fds'
            else fds' 

-- versao 2
--
armstrong2 :: Ord b => Int -> Int -> FDS b -> FDS b
armstrong2 n1 n2 fds = 
    let fds' = t5 n1 $ t6 n1 fds
    in if fds' /= fds && n2 > 0
        then armstrong2 n1 (n2 - 1) fds'
        else fds'
-}

-- seems that it is not important the order that the rules are apllied - NOT TRUE
-- in the IMDB case trans (ptrans (addi x)) 134; ptrans (trans (addi x)) 131 
-- in the IMDB case trans (ptrans x) 133; ptrans (trans x) 133 
-- the problem here is the order: when addi is applied, it changes the order and
-- so the others give different results - BAD, not so bad: the closure 
-- is not unique, so it depends on the algorithm
--
--armstrong :: Eq a => FDS a -> FDS a
armstrong fds =
  let n = length fds
--  in desunion $ (transitivity n) $ {- (ptransitivity n) $ -} (additivity n) $ elimRed $ elimIn fds 
  in elimIn $ elimRed $ desunion $ transitivity (n+1) fds
-----------------------


-- ok
--
additivity :: Eq a => Int -> FDS a -> FDS a
additivity 0 fds = fds 
additivity n t = let t' = aux t
                   in additivity (n-1) t'
  where
    aux [x] = [x]
    aux ((a, c):(a1, c1):t) | a == a1 = (a, Atts $ (atts c) `union` (atts c1)):t 
                            | otherwise = (a1, c1):(aux ((a, c):t)) 

-- tests
-- ok
--
addi = map (id >< ((:[]))) [(1,2), (1, 3), (2, 4), (3, 5), (2, 5)]


-- ok
transitivity :: Eq a => Int -> FDS a -> FDS a
transitivity 0 fds = fds 
transitivity n t = let t' = aux t
                   in transitivity (n-1) t'
  where
    aux [x] = [x]
    aux ((a, c):(a1, c1):t) | c == a1 = (a, c1):t 
                            | otherwise = (a1, c1):(aux ((a, c):t)) 

-- tests
-- ok
--
trans1 = [ (1, 2), (1, 4), (2, 4), (3, 4), (3, 5), (4, 2) ]

trans fds = let fds' = transitivity2 fds
            in if fds == fds' then fds else transitivity2 fds'

transitivity2 [a, b] = aux a b
transitivity2 (a:b:t) = (aux a b) ++ (transitivity2 t) 
 
aux (x, y1) (y2, z) | y1 == y2 = [(x, z)]
                        | otherwise = [(x, y1), (y2, z)]

-- ok
ptransitivity :: Eq a => Int -> FDS a -> FDS a
ptransitivity 0 fds = fds 
ptransitivity n t = let t' = aux t
                    in ptransitivity (n-1) t'
  where
    aux [x] = [x]
    aux ((a, c):(a1, c1):t) | c <| a1 = (Atts $ (atts a) `union` ((atts a1) \\ (atts c)), c1):t 
                            | otherwise = (a1, c1):(aux ((a, c):t)) 

-- tests
-- ok
--
ptrans = [ ([1], [2]), ([1], [4]), ([2, 3], [4]), ([3], [4]), ([3], [5]), ([4], [2]) ]




-- old stuff


-- transitivity
-- tornar isto mais eficiente
-- aplicar a lei ate nao haver alteracoes
-- posso parar quando o tamanho nao se alterar mais...
-- parece funcionar
t5 :: Eq b => Int -> FDS b -> FDS b
t5 n fds = 
    let ref = [((x, z), (x, y1), (y2, z)) | (x, y1) <- fds, (y2, z) <- fds, y1 == y2]
        toRem = concatMap (\(x, y, z) -> y:[z]) ref    
        toAdd = map (\(x, y, z) -> x) ref
        res = List.union (fds List.\\ toRem) toAdd 
    in if res /= fds && n > 0
        then t5 (n-1) res
        else res




-- pseudo-transitivity
-- parece k esta a criar mais FDs...
{-
t6 :: Ord b => Int -> FDS b -> FDS b
t6 n fds = 
  let fds' = sort $ map (sort >< sort) fds
      ref = [((sort $ List.union x (yz List.\\ y), w), (x, y), (yz, w)) 
                | (x, y) <- fds, (yz, w) <- fds, y <| yz, x /= yz, y /= w] 
      toRem = concatMap (\(x, y, z) -> y:[z]) ref    
      toAdd = map (\(x, y, z) -> x) ref
      res = sort $ L.union (fds L.\\ toRem) toAdd
  in if res /= fds && n > 0
        then t6 (n-1) res
        else res 
-}
-- tornar isto mais eficiente
-- aplicar a lei ate nao haver alteracoes
-- pode haver aqui um problema com a ordem dos elementos que pode fazer com que o /= nao esteja
-- a ser aplicado correctamente RESOLVIDO
-- talves isto nao seja o problema, talves seja possivel isto nao parar... em certos casos
-- isto nao para...
-- parece funcionar
{-
t6' :: (Ord b, Show b) => FDS b -> IO (FDS b)
t6' fds = 
 do
  let fds' = sort $ map (sort >< sort) fds
      ref = [((sort $ L.union x (yz L.\\ y), w), (x, y), (yz, w)) | (x, y) <- fds, (yz, w) <- fds, y <| yz] 
      toRem = concatMap (\(x, y, z) -> y:[z]) ref    
      toAdd = map (\(x, y, z) -> x) ref
      res = sort $ L.union (fds L.\\ toRem) toAdd
  putStrLn $ show res
  putStrLn ""
  if res /= fds
        then t6' res
        else return res 
-}

-- "thinked test"
--

-- teste em que ha mt pseudo-transitividade
-- quando ha mt pseudo-transitividade, a quantodade de FDs cresce bastante
-- em vez de diminuir. parece que a pseudo-transitividade nao e la mt util...
--
-- nestes casos a transitividade funciona bastante bem
h0 = [  (["a","b","c"], ["d"]), 
        (["a","b","d"], ["c"]), 
        (["b","c"],     ["a","d"]), 
        (["a","d"],     ["a"]), 
        (["a"],         ["a","b","d"]) ] 


-- generic tests
-- 

g0 =  [ (["a"], ["d"]),
        (["a","b"], ["e"]), (["b","i"], ["e"]), (["c","d"], ["i"]), (["e"], ["c"])]

g1 = [(["a","b"],["c"]), (["c"],["b"])]

g2 = [  (["b","e"], ["a","b","c","d","e","i","j"]),
        (["j","i","k"], ["a"]),
        (["b"], ["b","c","a","d"]) ]





