module Synthesize where


import Types
import Combinators
import CFDSCare 
import FDInference (fun, lks2fds, powerset')

import Data.List as List
import Data.Map as Map
import Data.Char
import Debug.Trace


--
-- module with functions to trnasform a schema into a 3NF schema w.r.t. a set of FDs
--




--
--synthesize3 :: Eq a => FDS a -> [FDS a]
synthesize3 fds
  = 
    let h1 = nonRedun $ leftRed fds
        h2 = part h1
        fds' = groupBy (\((a, _)) ((a', _)) -> equivalent a a' h1) h1
        (j, h3) = mergefds fds' h2 -- newmerge fds h2 h2
    in elimT h1 h3 j

--
part :: Eq a => FDS a -> [FDS a]
part fds =  groupBy (\a b -> (fst a) == (fst b)) fds


-- receives the grouped fdss by left side equivalence and all the fds h
-- returns J and the new H
-- 
mergefds :: Eq a => [FDS a] -> [FDS a] -> (FDS a, [FDS a])
mergefds [] al = ([], al)--  groupBy (\((a, _)) ((a', _)) -> equivalent a a' (concat al)) (concat al))
mergefds (h:t) al 
  = 
    let (fds, as) = formfd h
        (res1, res2) = mergefds t (List.map (List.\\ as) al)
    in (fds++res1, res2)
  where
    formfd [fd] = ([],[])
    formfd fds = let as = List.map fst fds
                     newfds = [(a, c) | a <- as, c <- as, a /= c]
                     newh = [(a, Atts [c]) | (a, bs) <- newfds, c <- (atts bs)] ++
                            [(a, Atts [c]) | (bs, a) <- newfds, c <- (atts bs)]
                 in (newfds, newh)

{-
newmerge :: Eq a => FDS a -> [FDS a] -> [FDS a] -> (FDS a, [FDS a])
newmerge hs ht alfs = let ht' = map concat $ groupBy () ht
aux [] ht alfs
  where
    aux j [] fs = (j, fs)
    aux j (h:t) alfd = let (j', fs') = merge2 j h (List.delete h alfs) 
                       in aux j' t fs'

    merge2 j _ [] = (j, [])
  --  merge2 j [] _ = (j, [])
    merge2 j h1@((a, c):r) h@(((a', c'):r'):fs)  
      = 
        if equivalent a a' hs
          then
            let (j', h') = merge2 j h1 (List.map (clear a a') fs)
            in ((a, a'):((a', a):j'), (h1 ++ ((a', c'):r')):h')
          else 
            let (j', h') = merge2 j h1 fs
            in (j', (((a', c'):r')):h')

    clear a a' fds = fds List.\\ ([(a, Atts [c]) | c <- atts a'] ++ [(a', Atts [c]) | c <- atts a])
   -}         


-- receives the H and the J and do the trick
--
elimT :: Eq a => FDS a -> [FDS a] -> FDS a -> [FDS a]
elimT alfs h j = aux $ powerset' $ concat h
  where
    aux (h':h's) 
      = 
        let ch'j = h' `List.union` j
            chj = (concat h) `List.union` j
        in if and (List.map (memberFD ch'j) chj) && and (List.map (memberFD chj) ch'j)
                  then zipf (groupBy (\((a, _)) ((a', _)) -> equivalent a a' alfs) h') j
                --  then (groupBy (\((a, _)) ((a', _)) -> equivalent a a' alfs) h')
                  else aux h's

    zipf h' [] = h'
    zipf h' ((a, c):t) = zipf (zipf2 h' (a, c)) t

    zipf2 [] (a, c) = [[(a, c)]]
    zipf2 (h1:h2) (a, c) = if or (List.map (\x -> equivalent x a alfs) (List.map fst h1))
                                        then ((a, c):h1):h2
                                        else h1:(zipf2 h2 (a, c)) 




ff :: Eq a => [(Attributes a, Attributes a)] -> [Attribute a]
ff = \z -> nub $ (concatMap (atts.fst) z) ++ (concatMap (atts.snd) z)

-- 
mergefds' :: Eq a => [FDS a] -> ([FDS a], [FDS a])
mergefds' (fd:fds) = aux fds fds
  where
    aux :: Eq a => [FDS a] -> [FDS a] -> ([FDS a], [FDS a])
    aux [] h = ([], h)
    aux (fd:fds) h =
      let (fd', h') = merge1 (head fd) fds h
          (x, y) = aux fds h'
      in (fd'++x, y)

    merge1 :: Eq a => FD a -> [FDS a] -> [FDS a] -> ([FDS a], [FDS a])
    merge1 (a, c) [] h = ([[(a, c)]], h)
    merge1 (a, c) (((a', c'):r):fds) h = if equivalent a a' (concat h) 
                                      then let (x, y) = merge1 (a, c) fds $ clear a a' h
                                           in (((a', c'):r):x, y)
                                      else merge1 (a, c) fds h

    clear :: Eq a => Attributes a -> Attributes a -> [FDS a] -> [FDS a]
    clear a a' h =
      let aa' = cons a a'
          a'a = cons a' a
      in List.map (List.\\ (aa' ++ a'a)) h

    cons :: Attributes a -> Attributes a -> FDS a 
    cons _ (Atts []) = []
    cons a (Atts (a':a's)) = (a, Atts [a']):(cons a (Atts a's))


mergefds2 h fds alfs =
    let fds' = groupBy (\((a, _)) ((a', _)) -> equivalent a a' alfs) fds
        fds'' = aux fds'
    in (aux fds', clear h (todel fds''))
  where
    aux [] = []
    aux (fds:fdss) = 
      let fsts = List.map fst fds
          fds' = [(x, y) | x <- fsts, y <- fsts, x /= y]
      in fds':(aux fdss)
        
    todel [] = []    
    todel (fds:fdss) = (cons (List.map fst fds) (List.map snd fds)) ++ (todel fdss)

    cons as cs = [(a, Atts [c]) | a <- as, c <- concatMap atts cs] ++ 
                 [(a, Atts [c]) | a <- cs, c <- concatMap atts as]
        
    clear h fds = List.map (List.\\ fds) h   







-- transforming FDs into CFDs and vice-versa
--
fds2cfd :: FDS a -> CFD a
fds2cfd fds = (List.map fst fds, snd $ head fds)

cfd2fds :: CFD a -> FDS a
cfd2fds ([], _) = []
cfd2fds ((h:t), c) = (h,c):(cfd2fds (t, c))

cfds2fds :: CFDS a -> FDS a 
cfds2fds = concatMap cfd2fds



-- |  based on the end of the page 90, in the end it's necessary to perform a reduce step
--    this was comproved by tests
{-
*Main Types Synthesize FDInference FDCare Tests> minimize [(Atts [Att "clientNo",Att "propertyNo",Att "cName"],Atts [])]
[]
*Main Types Synthesize FDInference FDCare Tests> reduce [(Atts [Att "clientNo",Att "propertyNo",Att "cName"],Atts [])]
[([],[])]
*Main Types Synthesize FDInference FDCare Tests> annular [(Atts [Att "clientNo",Att "propertyNo",Att "cName"],Atts [])]
[([[Att "clientNo",Att "propertyNo",Att "cName"]],[])]
-}
-- | testar com  [(Atts [Att "clientNo",Att "propertyNo",Att "cName"],Atts [])]
--
synthesize :: (ToInt a, Eq a) => FDS a -> CFDS a
synthesize = annular.reduce.minimize

-- ok
{-
rightRedCFD.annular.reduce.minimize

([[Att "rentStart"],[Att "rentFini"],[Att "cname",Att "rent"]],[])
([[Att "prop"],[Att "addr"],[Att "rent"]],[Att "oname"])
([[Att "owner"],[Att "oname"]],[])
([[Att "client"],[Att "cname"]],[])

rightRedCFD.shift.nonRedunCFD.annular.reduce.minimize

([[Att "cname"]],[Att "client"])
([[Att "oname"]],[Att "owner"])
([[Att "rent"]],[Att "oname",Att "prop",Att "addr"])
([[Att "cname",Att "rent"]],[Att "rentStart",Att "rentFini"])

rightRedCFD.shift.nonRedunCFD.annular.reduce.minimize

([[Att "rent"]],[Att "oname",Att "prop",Att "addr"])
-}


synthesize' :: (ToInt a, Show a,  Ord a) => FDS a -> CFDS a
synthesize' = reduceCFD.nonRedunCFD.annular.minimize

--
synthesize'' :: (ToInt a, Show a,  Ord a) => FDS a -> CFDS a
synthesize'' = rightRedCFD.annular.minimize

-- tests of synthesize
--

-- pag 115, ex 6.24
-- ok
syn1 = [([1], [2]), ([2], [1]), ([1,3], [4,5]), ([2,4], [3])]



-- ===================================== MINIMIZE

-- alg 5.8 da 85
--
minimize :: (Eq a, ToInt a) => FDS a -> FDS a
minimize fds 
  = 
        let fds' = nonRedun fds
            efsbar = generatesEFXBar fds' -- fds'
        in (minimize' efsbar efsbar fds')
    where
        minimize' [] _ fds = fds
        minimize' (ef:efs) efsbar fds = minimize' efs efsbar (minimize'' ef ef fds)

        minimize'' [] _ fds = fds
        minimize'' (ef:efs) efs' fds = 
            minimize'' efs efs' (replace ef (List.delete ef efs') fds)

        replace _ [] fds = fds
        replace fd (fd':fds') fds =
            if direct fds (fst fd, fst fd') 
                then
                    (fst fd', Atts $ (atts $ snd fd) ++ (atts $ snd fd')) : 
                      (replace fd fds' $ List.delete fd $ List.delete fd' fds)
                else
                    replace fd fds' fds

-- tests to minimize
-- need more tests
--
 
-- ok
min1 = [(Atts [Att 1], Atts [Att 2]), (Atts [Att 1], Atts [Att 3])]

-- ok
min2 = [(Atts [Att 1, Att 2, Att 3], Atts [Att 4]), (Atts [Att 1], Atts [Att 2])]




-- ======================================== redundant

-- ok
-- computes nonredundat covers
-- alg 5.4 da 73
--
nonRedun :: Eq a => FDS a -> FDS a
nonRedun fds = nonRedun' fds fds
    where
        nonRedun' [] f = f
        nonRedun' (xy:gs) f = if memberFD (List.delete xy f) xy
                                    then nonRedun' gs (List.delete xy f)
                                    else nonRedun' gs f


-- =================================== Efx

-- much simpler version than the one bellow
generatesEFXBar :: Eq a => FDS a -> [FDS a]
generatesEFXBar fds = groupBy (\a b -> equivalent (fst a) (fst b) fds) $ sortBy (sortAnn fds) fds


{-
-- ok
-- pag 78
generatesEFXBar :: Eq a => FDS a -> FDS a -> [FDS a]
generatesEFXBar [] _ = []
generatesEFXBar (fd:fds) aux = 
                        let (fds', fd') = generatesEFX fd aux (fd:fds) ([], [])
                        in fd' : (generatesEFXBar fds' aux)
-}


-- a version equivalent to the one below but much simpler
--
generatesEFX :: Eq a => FD a -> FDS a -> FDS a
generatesEFX (a, _) fds = List.filter (\(a', _) -> equivalent a a' fds) fds

{-
-- given a set X of attributes and a set of fds, this function calculates
-- the E_F(X) in the given set of fds and the set of fds without the used fds
-- (FDS, FDS) the first one is the new set of fds and the second one the E_F(X)
-- ok
generatesEFX :: Eq a => FD a -> FDS a -> FDS a -> (FDS a, FDS a) -> (FDS a, FDS a)
generatesEFX _ _ [] p = p
generatesEFX fd fds' (fd':fds) (new, efx) =
                                    if equivalent (fst fd) (fst fd') fds'
                                        then generatesEFX fd fds' fds (new, fd':efx)
                                        else generatesEFX fd fds' fds (fd':new, efx)
-}


-- ====================================== CLOSURE
-- computes the closure of a set of attributes


-- to FDS
-- need tests
--
linClosure' :: Ord a => Attributes a -> FDS a -> Attributes a
linClosure' x fs
   =
    let (count, list) = inicialization' fs
        aux = fromList fs
    in computation' aux x list x x count

inicialization' :: Ord a => FDS a -> (Map (FD a) Int, Map (Attribute a) (FDS a))
inicialization' fs
  =
    let count = createMap fs
        list = createList empty fs
    in (count, list)
  where
    createMap [] = empty
    createMap (f:fs) = Map.insert f (length $ atts $ fst f) (createMap fs)

    createList m [] = m 
    createList m (f:fs) 
      =
        let m1 = createListaux m f (atts $ fst f)
        in createList m1 fs
  
    createListaux m _ [] = m 
    createListaux m f (h:t) 
      = 
        let m1 = Map.insertWith (++) h [f] m
        in createListaux m1 f t
    

computation' :: Ord a => Map (Attributes a) (Attributes a) ->
                         Attributes a -> Map (Attribute a) (FDS a) -> 
                        Attributes a -> Attributes a -> Map (FD a) Int ->
                        Attributes a
computation' _ (Atts []) _ newdep _ _ = newdep
computation' aux (Atts (up:ups)) list newdep updat count
  =  -- choose A in update
        let fd = list Map.! up -- up :: Attributes a, fd :: [Attributes a]
            (newdep', updat', count') = compaux2' aux fd (newdep, Atts $ List.delete up (atts updat), count)
        in computation' aux updat' list newdep' updat' count'

  -- for each FD in list A
compaux2' :: Ord a => Map (Attributes a) (Attributes a) -> FDS a -> 
                     (Attributes a, Attributes a, Map (FD a) Int) ->
                     (Attributes a, Attributes a, Map (FD a) Int)
compaux2' _ [] t = t 
compaux2' aux (fd:fds) (newdep, updat, count)
      = 
        let c = count Map.! fd
        in if (c-1) == 0
              then
                let add = (atts $ snd fd) List.\\ (atts newdep)   
              -- (newdep, update, count)
                in compaux2' aux fds
                   ( Atts $ List.union (atts newdep) add, 
                     Atts $ List.union (atts updat) add,
                     update (\x -> Just (x-1)) fd count)
              else
                compaux2' aux fds (newdep, updat, Map.update (\x -> Just (x-1)) fd count)


-- a weak version, but working
closureS :: Eq a => Attributes a -> FDS a -> Attributes a
closureS ats fds = closure' fds (Atts []) ats
  where
    closure' fds olddep newdep = if newdep /= olddep
                                            then let newdep' = aux fds newdep
                                                 in closure' fds newdep newdep'
                                            else newdep

    aux [] newdep = newdep
    aux ((a, c):fds) newdep = if a <| newdep
                                then aux fds (Atts $ (atts newdep) `List.union` (atts c))
                                else aux fds newdep

-- tests to closure
--

-- ok
clo1 = tofds [([1], [4]), ([1, 2], [5]), ([2, 9], [5]), ([3, 4], [9]), ([5], [3])]




-- ========================================== REDUCE

-- reduce a set of fds
--

-- simes ok
-- generates a reduced cover
reduce :: Eq a => FDS a -> FDS a
reduce = rightRed . leftRed 
              -- filter ((/=[]).atts.snd) fds'


-- simes ok
-- o 1º set de fds é o que sera reduzido, o 2º e onde se verifica o member
leftRed :: Eq a => FDS a -> FDS a
leftRed fds = aux1 fds fds
  where
    aux1 [] fds = fds
    aux1 ((a, c):fds) fds' 
      =      
        let fds'' = aux2 fds' (a, c) a
        in aux1 fds fds''

    aux2 f _ (Atts []) = f
    aux2 f (Atts alxs, y) (Atts (x:xs)) = if (memberFD f (Atts $ List.delete x alxs, y))
                                              then 
                                                let alxs' = Atts $ List.delete x alxs
                                                in aux2 (List.delete (Atts alxs, y) f `List.union` 
                                                          [(alxs', y)]) (alxs', y) (Atts xs)
                                              else
                                                aux2 f (Atts alxs, y) (Atts xs)


-- simes not ok 
rightRed :: Eq a => FDS a -> FDS a
rightRed fds = aux1 fds fds
  where
    aux1 [] fds = fds
    aux1 ((a, c):fds) fds' 
      =      
        let fds'' = aux2 fds' (a, c) c
        in aux1 fds fds''

    aux2 f _ (Atts []) = f
    aux2 f (x, Atts alys) (Atts (y:ys)) 
      =
        let nalys = Atts $ List.delete y alys
            nf = List.delete (x, Atts alys) f
        in 
          if memberFD (nf `List.union` [(x,nalys )]) (x, Atts [y])
                      then 
                        aux2 (nf `List.union` [(x, nalys)]) (x, nalys) (Atts ys)
                      else
                        aux2 f (x, Atts alys) (Atts ys)
    
-- tests
-- need more tests

-- ok
-- pag 77, ex 5.6
red1 = [([1], [3]), ([1, 2], [4, 5]), ([1, 2], [3, 4, 9]), ([1, 3], [10])]

-- ok
-- pag 74, ex 5.5
red4 = [([1], [2, 3]), ([2], [3]), ([1, 2] ,[4])]



-- tests if red is left reduced
testLe red = not $ or $ List.map (aux red) red
  where
    aux _ (Atts [a], _) = False
    aux fds (a, c) = or $ List.map (\e -> equiv fds ((List.delete (a, c) fds) ++ 
                                                 [(Atts $ List.delete e (atts a), c)])) (atts a)

-- tests if red is right reduced
testRi red = not $ or $ List.map (aux red) red
  where
    aux _ (_, Atts [c]) = False
    aux fds (a, c) = or $ List.map (\e -> equiv fds ((List.delete (a, c) fds) ++ 
                                                 [(a, Atts $ List.delete e (atts c))])) (atts c)

-- tests if red is reduced
testRed red = testRi red && testLe red

-- ok to minimize and reduce
red2 = [([1, 2, 3], [4]), ([1, 2], [5]), ([5], [1, 2])]





-- ======================================= ANNULAR

annular :: Eq a => FDS a -> CFDS a
annular fds 
  = 
    let fs = groupBy (\a b -> equivalent (fst a) (fst b) fds) $ sortBy (sortAnn fds) fds
    in List.map aux fs
  where
    aux [] = ([], Atts [])
    aux ((a, c):r) 
        = let (a', c') = aux r
          in (a:a', Atts $ ((atts c') `List.union` (atts c)) List.\\ (concatMap atts (a:a')))



-- tests to annular
--

-- pag 88, ex 5.16
-- ok
ann1 :: [([Int], [Int])]
ann1 = [([1], [2, 3]), ([2], [1, 4]), ([1, 5], [9]), ([2, 5], [10, 9])]

rma1 = [([[1], [2]], [3, 4]), ([[1, 5]], [9, 10])]

-- pag 110, ex 6.19
-- ok
ann2 = [(["a"], ["b1", "b2", "c1", "c2", "d", "e"]), 
        (["b1", "b2", "c1"], ["a"]),
        (["b1", "b2", "c2"], ["a"]),
        (["e"], ["i1", "i2"]),
        (["c1", "d"], ["j"]),
        (["c2", "d"], ["j"]),
        (["i1", "i2"], ["i3"]),
        (["i2", "i3"], ["i1"]),
        (["i1", "i3"], ["i2"])]



-- ========================================= derives, equiv, ...

sortAnn fds (a, c) (a', c') | equivalent a a' fds = EQ
                            | otherwise = GT



fdint :: [([Int], [Int])] -> [(Attributes Char, Attributes Char)]
fdint red = List.map ((Atts . (List.map (Att. chr.(+64)))) >< (Atts . (List.map (Att. chr.(+64))))) red


-- verifies if x <-> y (x and y are equivalent) in a set of fd
-- simes ok
equivalent :: Eq a => Attributes a -> Attributes a -> FDS a -> Bool
equivalent x y fds = memberFD fds (x,y) && memberFD fds (y,x)


-- pag 72, alg, 5.2
--
equiv :: Eq a => FDS a -> FDS a -> Bool
equiv f1 f2 
  = 
    derives f1 f2 && derives f2 f1
    

--pag 72, alg 5.1
--
derives :: Eq a => FDS a -> FDS a -> Bool
derives f1 f2 
  = 
    and $ List.map (memberFD f1) f2


-- verifies if a given fd holds in a set of fds
memberFD :: Eq a => FDS a -> FD a -> Bool
memberFD [] _ = False -- needed to theorical confirmation ...
memberFD fds (a, c) = c <| (closureS a fds)


direct :: Eq a => FDS a -> FD a -> Bool
direct fds fd = let fds' = nonRedun fds
                    efx = generatesEFX fd fds'
                in memberFD ((List.\\) fds' efx) fd 

