module Tests where


import Types
import FDInference
import CFDSCare
import Synthesize
import FDCare
import TransfData
import TFile

{-
import Gnumeric2Haskell
import Language.Gnumeric.Formula

import Sheet.Sheet
import Sheet.Indx
import Sheet.Expr
import UCheck.UCheckServer
import UCheck.MParser
import UCheck.Header
import UCheck.UnitErrors
import UCheck.Units
import UCheck.FaultLoc (faultLoc)
-}

import Data.List

{-
gnu2fds f = do
  ss <- readGnu f
  let (b:h:r) = tables ss
  return (Atts $ map Att (head $ sheet2list h), sheet2list b)
-}


--
test1 sch rel1
  = 
        let rels = rel1:((drop 1 rel1):((drop 2 rel1):[(drop 3 rel1)]))
            Just xs = mapM (\rel -> fun sch rel) rels
            ys = map (synthesize''.lks2fds) xs
        in foldr1 union ys


xc :: IO (FDS String)
xc = 
  do
    let x = fun imdbSch imdbRel >>= lks2fds
    return $ reduce $ minimize x


-- =============================== FDs e CFDs

-- FDS

-- pag 88 ex 5.16
fd1 = [(Atts [Att "a"], Atts [Att "b", Att "c"]), 
       (Atts [Att "b"], Atts [Att "a", Att "d"]), 
       (Atts [Att "a", Att "e"], Atts [Att "i"]), 
       (Atts [Att "b", Att "e"], Atts [Att "j", Att "i"])]


-- pag 78, ex 5.9

fd2 = [(Atts [Att "a"], Atts [Att "b", Att "c"]),
       (Atts [Att "b"], Atts [Att "a"]),
       (Atts [Att "a", Att "d"], Atts [Att "e"])]

fd3 = [(Atts [Att "a"], Atts [Att "a", Att "b", Att "c"]),
       (Atts [Att "b"], Atts [Att "a"]),
       (Atts [Att "b", Att "d"], Atts [Att "e"])]     


-- pag 87, ex 5.15

fd4 = [(Atts [Att "a"], Atts [Att "b"]), 
       (Atts [Att "b"], Atts [Att "a"]), 
       (Atts [Att "a", Att "c"], Atts [Att "d"]), 
       (Atts [Att "b", Att "c"], Atts [Att "e"])]


-- pag 89 ex 5.17

fd5 = [(Atts [Att "a"], Atts [Att "b", Att "a"]), 
       (Atts [Att "b", Att "a"], Atts [Att "b"]), 
       (Atts [Att "b"], Atts [Att "a", Att "c", Att "d"]), 
       (Atts [Att "a", Att "e"], Atts [Att "j", Att "i"])]
 
-- pag 110 ex 6.19
--
fd6 = tofds [
        (["b1", "b2", "c2"], ["a", "c1", "d", "e", "i1", "i2", "i3", "j"]),
        (["b1", "b2", "c1"], ["a","c2", "d", "e", "i1", "i2", "i3", "j"]),
        (["a"], ["b1", "b2", "c1", "c2", "d", "e", "i1", "i2", "i3", "j"]), 
        (["e"], ["i1", "i2", "i3"]),
        (["c1", "d"], ["j"]),
        (["i1", "i2"], ["i3"]),
        (["c2", "d"], ["j"]),
        (["i2", "i3"], ["i1"]),
        (["i1", "i3"], ["i2"])]

fd6' = tofds [
        (["b1", "b2", "c1"], ["a","c2", "d", "e", "i1", "i2", "i3", "j"]),
        (["a"], ["b1", "b2",  "c2", "d", "e", "i1", "i2", "i3", "j"]), 
        (["e"], ["i1", "i2", "i3"]),
        (["b1", "b2", "c1"], ["a", "c1", "d", "e", "i1", "i2", "i3", "j"]),
        (["c1", "d"], ["j"]),
        (["i1", "i2"], ["i3"]),
        (["c2", "d"], ["j"]),
        (["i2", "i3"], ["i1"]),
        (["i1", "i3"], ["i2"])]

fd7 = tofds [(["a"], ["b1", "b2", "c1", "c2", "d", "e"]), 
        (["b1", "b2", "c1"], ["a"]),
        (["b1", "b2", "c2"], ["a"]),
        (["e"], ["i1", "i2"]),
        (["c1", "d"], ["j"]),
        (["c2", "d"], ["j"]),
        (["i1", "i2"], ["i3"]),
        (["i2", "i3"], ["i1"]),
        (["i1", "i3"], ["i2"])]

--pag 90 ex 5.21
--
fd8 = tofds [
        (["b1", "b2"], ["a"]),
        (["d1", "d2"], ["b1", "b2"]),
        (["b1"], ["c1"]),
        (["b2"], ["c2"]),
        (["d1"], ["a"]),
        (["d2"], ["a"]),
        (["a", "b1", "c2"], ["d2"]),
        (["a", "b2", "c1"], ["d1"])]


-- bernstein paper
fd9 = tofds [(["c"],["d"]), (["d"], ["c"]), (["c", "e"], ["f"])]

fd10 = tofds [(["c"],["d"]), (["d"], ["c"]), (["d", "e"], ["f"])]

fd11 = tofds [(["x1", "x2"], ["a"]), (["c"], ["x1", "x2"]), (["a", "x1"], ["b"]), 
              (["b", "x2"], ["c"])]


-- CFDs

-- pag 90 ex 5.20

cfd1 = [([Atts [Att "a"], Atts [Att "b"]], Atts [Att "c", Att "d"]), 
        ([Atts [Att "a", Att "e"]], Atts [Att "i", Att "j"])]




-- ================= schemas and relations


-- book renting

bookSch = Atts $ map Att ["ISBN","Title","Authors","PidB","NameB","AddressB","rentStart", "FORMtotalDays"]

bookRel = [ ["1","casa quieta","jose","1","carina","braga","10/07/08","12"],
            ["2","formula","rodrigo","2","jacome","braga","12/06/08","34"],
            ["3","accoes","borba","3","manuel","joane","20/03/08","111"],
            ["3","accoes","borba junior","4","antonio","taipas","20/03/08","111"],
            ["4","casa agitada","jose","3","jacome","guimaraes","20/03/08","111"],
            ["5","mais accoes","zeze","2","jacome","braga","23/4/08","222"] ]

bookfd = [(Atts $ map Att ["rentStart"], Atts $ map Att ["FORMtotalDays"])] 


bookSch1 = Atts $ init $ atts bookSch

bookRel1 = map init bookRel



-- factura

factSch = Atts $ map Att ["ProdCode", "ProdName", "FactCode", "Desc", "Quant"]

factRel = [ ["c1", "bananas",  "f1", "desc1", "2"],
            ["c2", "laranjas",  "f1", "desc1", "1.5"],
            ["c3", "pessegos", "f1", "desc1", "1"],

            ["c3", "pessegos",  "f2", "desc2", "4"],
            ["c4", "pessegos",  "f2", "desc2", "8"],
            ["c2", "laranjas",  "f2", "desc2", "3"],
            
            ["c4", "pessegos",  "f3", "desc2", "5"],
            ["c1", "bananas",  "f3", "desc2", "6"]
      ]




-- property renting

propSch = Atts $ map Att ["clientNo", "propertyNo", "cName", "pAddress", "rentStart", "rentFinish", "rent", "ownerNo", "oName"]

propSch2 = Atts $ map Att ["clientNo", "propertyNo", "cName", "pAddress", "rentStart", "rentFinish", "FORMtotalDays", "rentPerDay", "FORMtotal rent", "ownerNo", "oName"]

propRel2' = [
 ["cr76","pg4","john","6 Lawrence St.","1-Jul-00","31-Aug-01","602","50","30100","co40","tina"],
 ["cr76","pg16","john","5 Novar Dr.","1-Sep-01","1-Sep-02","365","70","25550","co93","tony"],
 ["cr56","pg4","aline","6 Lawrence St.","1-Sep-99","10-Jun-00","282","50","14100","co40","tina"],
 ["cr56","pg36","aline","2 Manor Rd","10-Oct-00","1-Dec-01","417","50","25550","co93","tony"],
 ["cr56","pg16","aline","5 Novar Dr.","1-Nov-02","10-Aug-03","282","70","19740","co93","tony"],
 ["cr56","pg4","aline","5 Novar Dr.","1-Nov-02","10-Aug-03","282","70","19740","co40","tina"]
 ]


propRel2 = [
  ["cr76","pg4","john","6 Lawrence St.","1-Jul-00","31-Aug-01","602","50","30100","co40","tina"],
  ["cr76","pg16","john","5 Novar Dr.","1-Sep-01","1-Sep-02","365","70","25550","co93","tony"],
  ["cr56","pg4","aline","6 Lawrence St.","1-Sep-99","10-Jun-00","282","50","14100","co40","tina"],
  ["cr56","pg36","aline","2 Manor Rd","10-Oct-00","1-Dec-01","417","60","25550","co93","tony"],
  ["cr56","pg16","aline","5 Novar Dr.","1-Nov-02","10-Aug-03","282","70","19740","co93","tony"]]

fdds = [(Atts $ map Att ["totalDays", "rentPerDay"], Atts $ map Att ["FORMtotal rent"]),
        (Atts $ map Att ["rentStart", "rentFinish"], Atts $ map Att ["FORMtotalDays"])]
     

propRel = [
 -- [1,40, 3,41, 5, 6, 7,42, 9],
  [1, 2, 3, 4, 5, 6, 7, 8, 9],
  [1,10, 3,11,12,13,14,15,16],
  [17,2,18, 4,19,20, 7, 8, 9],
  [17,21,18,22,23,24,25,15,16],
  [17,10,18,11,26,27,14,15,16]
--  [30,35,18,32,33,34,36,37,38]
--  [17,40,18,41,42,43,14,44,16]
 ]

propRel3 = [
  ["cr76","pg4","john","6 Lawrence St.","1-Jul-00","31-Aug-01","350","co40","tina"],
  ["cr76","pg16","john","5 Novar Dr.","1-Sep-01","1-Sep-02","450","co93","tony"],
  ["cr56","pg4","aline","6 Lawrence St.","1-Sep-99","10-Jun-00","350","co40","tina"],
  ["cr56","pg36","aline","2 Manor Rd","10-Oct-00","1-Dec-01","375","co93","tony"],
  ["cr56","pg16","aline","5 Novar Dr.","1-Nov-02","10-Aug-03","450","co93","tony"]]


propRel5 = [
 ["cr76","pg4","john","6 Lawrence St.","1-Jul-00","31-Aug-01","50","co40","tina"],
 ["cr76","pg16","john","5 Novar Dr.","1-Sep-01","1-Sep-02","70","co93","tony"],
 ["cr56","pg4","aline","6 Lawrence St.","1-Sep-99","10-Jun-00","50","co40","tina"],
 ["cr56","pg36","aline","2 Manor Rd","10-Oct-00","1-Dec-01","70","co93","tony"],
 ["cr56","pg16","aline","5 Novar Dr.","1-Nov-02","10-Aug-03","70","co93","tony"],
 ["cr56","pg4","aline","6 Lawrence St.","1-Nov-02","10-Aug-03","50","co40","tina"]]



--["client", "prop", "cname", "addr", "rentStart", "rentFini", "rent", "owner", "oname"]
propRel12 = [
  [1, 2, 3, 4, 5, 6, 7, 8, 9],
  [1,10,30,11,12,13,14,15,16],
  [17,2,18, 4,19,6, 7, 8, 9],
  [17,21,18,22,5,24,25,15,16],
  [17,10,18,11,26,27,14,15,16]
 ]


propSch1 = Atts $ map Att ["clientNo", "propertyNo", "cName"]

propRel1 = [
  ["cr76","pg4","john"],
  ["cr76","pg16","john"],
  ["cr56","pg4","aline"],
  ["cr56","pg36","aline"],
  ["cr56","pg16","aline"]]


propSch4 = Atts $ map Att ["clientNo", "propertyNo", "cName", "pAddress"]

propRel4 = 
 [["cr76","pg4","john","6 Lawrence St."],
  ["cr76","pg16","john","5 Novar Dr."],
  ["cr56","pg4","aline","6 Lawrence St."],
  ["cr56","pg36","aline","2 Manor Rd"],
  ["cr56","pg16","aline","5 Novar Dr."]] 




estSch = Atts $ map Att ["empresa","contacto","local","ordenado","estado","descricao","tempo","nr aluno","nome aluno","curso"]

estRel = [
  ["emp1","253555555","lisboa","400","cancelado","tester","6 meses","30000","antonio","lesi"]
 ,["emp1","253555555","braga","400","ok","programador","meio ano","30001","joaquim","mcc"]
 ,["emp1","253555555","braga","250","ok","tirar fotocopias","3 meses","30002","manuel","cc"]
 ,["emp2","253666666","porto","300","ok","lavar wc","5 meses","30000","antonio","lesi"]
 ,["emp2","253666666","porto","500","ok","ajudante ","meio ano","30003","jorge","lei"]
 ,["emp3","323232323","guimaraes","500","cancelado","pm","3 meses","30004","paulo","mcc"]
 ,["emp4","344444444","taipas","400","ok","tester","1 ano","30004","paulo","mcc"]
 ,["emp5","333333333","amares","300","ok","programador","12 meses","30005","joao","lesi"]
 ,["emp5","333333333","amares","200","ok","pm","6 meses","30006","miguel","mcc"]
 ,["emp7","444444444","amares","500","ok","programador","6 meses","30007","carlos","cc"]
 ,["emp8","666444444","lisboa","700","ok","programador","6 meses","30008","joao","lei"]
 ,["emp9","666884444","braga","400","ok","tester","4 meses","30009","jorge","lei"]
 ]


hotelR = Atts $  tail $ map Att ["IdRow", "A", "B", "C", "D", "E"] 

hotel = map tail [   ["1","1","100","1","2","50"],
                     ["2","4","101","1","2","50"],           
                     ["4","1","200","1","2","50"],
                     ["6","2","200","1","3","70"],
                     ["3","1","102","2","2","70"],            
                     ["5","2","101","3","3","100"],
                     ["7","3","100","3","2","50"]
                 ]


testEmp = Atts [Att "a", Att "b", Att "c"]

testEmpd = [ ["", "b1", "c1"],
             ["", "b1", "c2"],
             ["a2", "b2", "c3"]
           ]
-- maybe this FDs between a and b shouldn't exist
{-
([Att "a"],[Att "b"])
([Att "b"],[Att "a"])
([Att "c"],[Att "a",Att "b"])
-}



imdbRel1 = [
  [1,  2,  3,  4,  5,  6,  7,  8,  9],
  [10,11, 12, 12, 14, 15, 16, 17, 18],
  [10,19, 20, 12, 21, 22, 7,  23, 18],
  [24,19, 12, 25, 21, 22, 26, 17, 27]
  ]


imdbSch = Atts $  map Att ["title", "year", "review", "country", "value", "director", "award", "actor", "role"]

imdbRel = 
  [
    ["Little Man", "2006", "2.6", "USA", "58", "Keenen Wayans", "3/6", "Marlon Wayans","a"],   
    ["Little Man", "2006", "2.6", "USA", "58", "Keenen Wayans", "3/6", "Marlon Wayans","c"],  
    ["Little Man", "2006", "2.6", "USA", "58", "Keenen Wayans", "3/6", "Shawn Wayans","b"], 
    ["Little Man", "2006", "2.6", "UK", "4", "Keenen Wayans", "3/6","Marlon Wayans","a"], 
    ["Little Man", "2006", "2.6", "UK", "4", "Keenen Wayans", "3/6","Shawn Wayans","b"], 
    ["Little Man", "2006", "2.6", "NL", "1", "Keenen Wayans", "3/6","Marlon Wayans","a"],  
    ["Little Man", "2006", "2.6", "NL", "1", "Keenen Wayans", "3/6","Shawn Wayans","b"], 
  
    ["Little Man","1997","6.9","HU","33","Amyn Kaderali","2/0","Jeremiah Alexis","c"],
    ["Little Man","1997","6.9","HU","33","Amyn Kaderali","2/0","Tony Walker Jr.","d"],  
    
    ["You, Me and D. 2", "2007", "5.6", "USA", "75", "Ken", "0/1","Owen Wilson","e"], 
    ["You, Me and D. 2", "2007", "5.6", "USA", "75", "Ken", "0/1","Owen Wilson","e"],
    
    ["You, Me and D.", "2006", "5.6", "USA", "75", "Anthony Russo", "0/2","Owen Wilson","e"], 
    ["You, Me and D.", "2006", "5.6", "USA", "75", "Anthony Russo", "0/1","Kate Hudson","f"],
 
    ["Edmond", "2005", "6.2", "USA", "0", "Stuart Gordon", "3/2", "William H. Macy","g"],
    ["Edmond", "2005", "6.2", "USA", "0", "Stuart Gordon", "3/2", "Frances Bay","h"],

    ["The OH in Ohio", "2006", "6.2", "USA", "1", "Billy Kent", "0/1","Parker Posey","i"],
    ["The OH in Ohio", "2006", "6.2", "USA", "1", "Billy Kent", "0/1","Danny DeVito","j"],

    ["Captivity", "2007", "6.2", "UK", "2", "Roland Joff", "3/2","Elisha Cuthbert","k"],
    ["Captivity", "2007", "6.2", "UK", "2", "Roland Joff", "3/2","Daniel Gillies","l"],

    ["Rescue Dawn", "2006", "8.6", "DY", "2", "Werner Herzog", "0/0","Christian Bale","m"],
    ["Rescue Dawn", "2006", "8.6", "DY", "2", "Werner Herzog", "0/0","Marshall Bell","n"],

    ["Talk to Me", "2007", "8.7", "PL", "34", "Kasi Lemmons", "1/2","Jim Annan","o"],
    ["Talk to Me", "2007", "8.7", "PL", "34", "Kasi Lemmons", "1/2","Benz Antoine","p"],

    ["Talk to Me","1996","7.0","RU","23","George Esguerra","2/0","Cheryl Clifford","q"],
    ["Talk to Me","1996","7.0","RU","23","George Esguerra","2/0","Peter Welch","r"],

 ["Wedding Crashers", "2005", "7.2", "USA", "209", "David Dobkin", "9/8","Owen Wilson","s"],
 ["Wedding Crashers", "2005", "7.2", "USA", "209", "David Dobkin", "9/8","Vince Vaughn","t"],
 ["Wedding Crashers", "2005", "7.2", "UK", "13", "David Dobkin", "9/8","Owen Wilson","s"],
 ["Wedding Crashers", "2005", "7.2", "UK", "13", "David Dobkin", "9/8","Vince Vaughn","t"],
 ["Wedding Crashers", "2005", "7.2", "CZK", "2", "David Dobkin", "9/8","Owen Wilson","s"],
 ["Wedding Crashers", "2005", "7.2", "CZK", "2", "David Dobkin", "9/8","Vince Vaughn","t"],
    
    ["Happy Endings", "2005", "6.7", "USA", "1", "Don Roos", "0/5","Lisa Kudrow","u"],
    ["Happy Endings", "2005", "6.7", "USA", "1", "Don Roos", "0/5","Steve Coogan","v"],
    
   ["Happy Endings", "1969", "6.6", "USA", "1", "Richard Brooks", "0/3", "Jean Simmons","w"],
   ["Happy Endings", "1969", "6.6", "USA", "1", "Richard Brooks", "0/3", "John Forsythe","x"],

    ["I, Robot", "2004", "7.0", "USA", "52", "Alex Proyas", "1/9","Will Smith","y"],
    ["I, Robot", "2004", "7.0", "USA", "52", "Alex Proyas", "1/9","Bridget Moynahan","z"],
    ["I, Robot", "2004", "7.0", "IT", "1", "Alex Proyas", "1/9","Will Smith","y"],
    ["I, Robot", "2004", "7.0", "IT", "1", "Alex Proyas", "1/9","Bridget Moynahan","z"],
    ["I, Robot", "2004", "7.0", "SP", "3", "Alex Proyas", "1/9","Will Smith","y"],
    ["I, Robot", "2004", "7.0", "SP", "3", "Alex Proyas", "1/9","Bridget Moynahan","z"],
    
 ["A Cinderella Story", "2004", "5.3", "USA", "13", "Mark Rosman", "4/8","Hilary Duff","aa"],
 ["A Cinderella Story", "2004", "5.3", "USA", "13", "Mark Rosman", "4/8","Jennifer Coolidge","bb"],
 ["A Cinderella Story", "2004", "5.3", "NL", "0", "Mark Rosman", "4/8","Hilary Duff","aa"],
 ["A Cinderella Story", "2004", "5.3", "NL", "0", "Mark Rosman", "4/8","Jennifer Coolidge","bb"],
 ["A Cinderella Story", "2004", "5.3", "IT", "0", "Mark Rosman", "4/8","Hilary Duff","aa"],
 ["A Cinderella Story", "2004", "5.3", "IT", "0", "Mark Rosman", "4/8","Jennifer Coolidge","bb"],

 ["Swimming Pool","2003","6.8","USA","0","Franois Ozon","2/11","Charlotte Rampling","cc"],
 ["Swimming Pool","2003","6.8","USA","0","Franois Ozon","2/11","Ludivine Sagnier","dd"],
 ["Swimming Pool","2003","6.8","IT","0","Franois Ozon","2/11","Charlotte Rampling","cc"],
 ["Swimming Pool","2003","6.8","IT","0","Franois Ozon","2/11","Ludivine Sagnier","dd"],
 
 ["Swimming Pool","2001","3.7","IT","0","Boris von Sychowski","2/2","Kristen Miller","ee"],
 ["Swimming Pool","2001","3.7","IT","0","Boris von Sychowski","2/3","Elena Uhlig","ff"],
 ["Swimming Pool","2001","3.7","MX","0","Boris von Sychowski","2/2","Kristen Miller","ee"],
 ["Swimming Pool","2001","3.7","MX","0","Boris von Sychowski","2/3","Elena Uhlig","ff"],

 ["28 Days Later","2002","7.4","UK","1","Danny Boyle","7/20","Alex Palmer","gg"],
 ["28 Days Later","2002","7.4","UK","1","Danny Boyle","7/20","Bindu De Stoppani","hh"],
 ["28 Days Later","2002","7.4","SP","1","Danny Boyle","7/20","Alex Palmer","gg"],
 ["28 Days Later","2002","7.4","SP","1","Danny Boyle","7/20","Bindu De Stoppani","hh"],

 ["28 Days","2001","7.4","SP","1","Danny Boyle","7/20","Bindu De Stoppani","hh"],
 
 ["28 Days 2","2001","7.4","SP","1","Ken","7/20","Bindu De Stoppani","hh"] 
  ] 


