module Types where


import Char
import Combinators
import Data.List as List  
import Data.Map as Map

import Data.Transform.Type
import Data.Transform.Rule



type CFD a = (LeftSets a, RightSets a)

type CFDS a = [CFD a]

type LeftSets a = [Attributes a]

type RightSets a = Attributes a


class ToInt a
	where
		toInt :: a -> Int
		
instance ToInt Char
	where
		toInt = digitToInt
		
instance ToInt a => ToInt [a]
	where
		toInt = sum . (List.map toInt)

instance ToInt a => ToInt (Attribute a)
	where
		toInt (Att a) = toInt a

instance ToInt Integer -- where toInt = id


displayCFD :: (Show b, Eq b) => CFDS b -> [String]
displayCFD = List.map (\(a, c) -> show (concat (List.map atts a)) ++ "->" ++ show c)





-- data base schema
-- uma tabela e um map entre chaves e valores
--
data Table a b = T (Map a b)

-- uma DB e um tupulo de tabelas
data DB a b = DB ([Table a b])


type RDB a b = [(FD a, Relation b)]


data LK a = LK { candidate :: Candidate a, cardinality :: Int, 
               quasiClosure :: Attributes a, closure :: Attributes a}
    deriving (Eq)

instance (Show a) => Show (LK a)
    where
        show (LK cand card qclsr clsr) = 
            "LK cand: " ++ show cand ++ ", card: " ++ show card ++ 
             ", qlcsr: " ++ show  qclsr ++ ", clsr: " ++ show clsr ++ "\n"


type LKS a = [LK a]

-- type FS a = LKS a

type Candidate a = Attributes a
    

-- a set of attributes
type R a = Attributes a

-- this is the list of filds of the relation
-- this should has an invariant that garantee that the length of each list is the length of the
-- relation shema
type Relation a = [[a]]

--type Relation = 

data Attribute a = Att {att :: a} deriving (Ord, Eq)

{-
instance Eq a => Eq (Attribute a)
  where
    (Att a) == (Att b) = a == b
    (Att a) /= (Att b) = a /= b
-}

--instance Ord a => Ord (Attribute a)
  --where
    --compare (Att a) (Att b) = compare a b

instance Show a => Show (Attribute a)
    where   
        show (Att a) = "Att " ++ show a




data Attributes a = Atts {atts :: [Attribute a]}  deriving Ord

instance Eq a => Eq (Attributes a)
  where
    a == b = a <| b && b <| a
    a /= b = (not $ a <| b) || (not $ b <| a)

instance Show a => Show (Attributes a)
  where
    show = show.atts


--
type FD a = (Attributes a, Attributes a)




type FDS a = [FD a]




-- sort by consequent size, biggest first
--
sort1 (a1, c1) (a2, c2) | length (atts c1) > length (atts c2) = LT
                        | otherwise = GT

sort2 (a1, c1) (a2, c2) = compare c1 c2 

sort3 (a1, c1) (a2, c2) | length (atts a1) <= length (atts a2) = LT
                        | otherwise = GT

sort4 (a1, c1) (a2, c2) | length (atts a1) <= length (atts a2) = if length (atts c1) > length (atts c2) 
                                                                    then LT else GT
                        | otherwise = GT



sortFDS ((a, c):fds) = (List.sort a, List.sort c):(sortFDS fds)
sortFDS a = a



group1 (a, c) (a', c') | c == c' = True
                       | otherwise = False

group2 (a, c) (a', c') | length (atts a) == length (atts a') = True
                       | otherwise = False




-- cheks if the first list is in the second one
--(<|) :: Eq a => [a] -> [a] -> Bool
(<|) (Atts l1) (Atts l2) = and $ List.map (`elem` l2) l1

(<||) l1 l2 = and $ List.map (`elem` l2) l1


tofds 
  =
    let f = List.map Att
        g = Atts. f
    in List.map (g >< g)


clear = List.map (nub . (List.map (List.filter (/= '"'))) . concat >< List.map (List.filter (/= '"'))) -- (map (filter (/= '\\')) x)


showfds l = (putStr.showList') l

showList' [] = ""
showList' (h:t) = (show h) ++ "\n" ++ (showList' t)


showFDS :: FDS String -> String
-- showFDS [] = []
--showFDS ((Atts a, Atts c):t) = showAt a ++ " -> " ++ showAt c ++ "\n" ++ showFDS t
showFDS = foldr (\(Atts a, Atts c) b -> showAt a ++ " -/ " ++ showAt c ++ "\n" ++ b) []
  where
--    showAt [] = []
    showAt = foldr (\(Att a) b -> a ++ ", " ++ b) []



data Box where 
  Box :: Type a -> a -> Box 

instance Show Box where
  show (Box t r) = "\n" ++ show t ++ "\n\n" ++ show' t r 

show' :: Type a -> a -> String
show' Int i = show i
show' Double i = show i
show' One _ = "()"
show' String "" = "\"\""
show' String s = show s
show' (List String) l = showList l ""
show' (List x) l = "[" ++ showl' (List x) l ++ "]" 
show' (Prod a b) (a', b') = "(" ++ show' a a' ++ "," ++ show' b b' ++ ")"
show' (Map a b) m = let l = Map.toList m
                    in "{" ++ showl' (List (Prod a b)) l ++ "}"
show' (Tag s String) s' = s'
show' (Tag s t) s' = s ++ show' t s'
show' (TInv t i) v = show' t v
    
showl' :: Type a -> a -> String
showl' (List _) [] = ""
showl' (List a) (h:t) = show' a h ++ "," ++ showl' (List a) t
--showl' (List (Tag s' String)) (h:t) = h ++ "," ++ showl' (List (Tag s' String)) t

--showType :: View a -> String
--showType (View _ b) = show b


showtp :: View a -> String
showtp (View _ b) = showTP b

showTP :: Type a -> String
--showT Int i = show i
--showT Double i = show i
--showT One _ = "()"
--showT String "" = "\"\""
--showT String s = show s
showTP (TInv t r) = "(TInv "++ (showTP t)++" "++(gshow (PF (Func t Bool)) r)++")"
showTP (List a) = "(List " ++ showTP a ++ ")"
showTP (Map a b) = "(Map " ++ showTP a ++ showTP b ++ ")"
showTP (Prod a b) = "(Prod " ++ showTP a ++ showTP b ++ ")"
showTP (Tag s String) = s ++ " " 
    

{-
show Int = "Int"
    show Double = "Double"
    show Bool = "Bool"
    show Char = "Char"
    show String = "String"
    show One = "One"
    show (Maybe x) = "(Maybe " ++ (show x) ++ ")"
    show (List x) = "(List "++(show x)++")"
    show (Set x) = "(Set "++(show x)++")"
    show (Map x y) = "(Map "++(show x)++" "++(show y)++")"
    show (Either x y) = "(Either "++(show x)++" "++(show y)++")"
    show (Prod x y) = "(Prod "++(show x)++" "++(show y)++")"
    show (Func x y) = "("++show x++" -> "++show y++")"
    show (Tag n t) = "(Tag "++(show n)++" "++(show t)++")"
    
    show (TInv t r) = "(TInv "++ (show t)++" "++(gshow (PF (Func t Bool)) r)++")"
    show (TVar s)   = "(TVar "++ (show s)++" "++")"

    show (PF a) = "(PF "++show a++")"
    show (Dynamic) = "Dynamic"
    show (Mu f) = "(Mu "++show f++")"
-}
    
