module HaExcel.FDS2Type where -- (
                          --fds2type, 
                          --getData
                       -- )  
                       -- where


import qualified Data.Map as Map
import Data.List
import Data.Maybe
import Control.Monad
import Debug.Trace

import Combinators

import Data.Transform.Type hiding (teq)
import Data.Transform.Rule hiding (teq)
import Types




teq :: MonadPlus m => Type a -> Type b -> m (Equal a b)
teq Int Int = return Eq
teq Bool Bool = return Eq
teq Char Char = return Eq
teq One One = return Eq
teq String String = return Eq
teq (Maybe a) (Maybe b) =
    do Eq <- teq a b
       return Eq
teq (List a) (List b) = 
    do Eq <- teq a b
       return Eq
teq (Set a) (Set b) = 
    do Eq <- teq a b
       return Eq
teq (Map a b) (Map c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (Prod a b) (Prod c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (Either a b) (Either c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (Func a b) (Func c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (PF a) (PF b) = 
    do Eq <- teq a b
       return Eq
teq (Tag _ a) b =
    do Eq <- teq a b
       return Eq
teq a (Tag _ b) =
    do Eq <- teq a b
       return Eq
teq (TVar x) (TVar y) = do
  guard (x==y)
  return Eq
teq (TInv t i) t' = teq t t'
teq t (TInv t' i) = teq t t'
teq _ _ = mzero



fds2type :: FDS String -> Box
fds2type fds = 
  let fdsc = map ((map att).atts >< (map att).atts) fds -- remove Atts and Att
      fdsbox = fds2box fdsc -- transform [FD] into (FD, (FD, ...))
      fdmap = fd2map fdsbox -- transformeach FD into map
  in fdmap 



-- [(a, b), (c, d), ...] -> ((a, b), ((c, d), ...)
fds2box :: [([String], [String])] -> Box
fds2box [a] = Box (Prod (List String) (List String)) a
fds2box (h:t) = aux h (fds2box t) 
  where
    aux h (Box t r) = Box (Prod (Prod (List String) (List String)) t) (h, r)



-- ((a, b), ((c, d), ...) -> (Map a b, (Map c d, ...))
fd2map :: Box -> Box
fd2map (Box (Prod (List String) (List String)) (a, b)) = pair2map (a, b)
fd2map (Box (Prod (Prod (List String) (List String)) (Prod (List String) (List String))) ((a, b), (c, d))) 
  = 
    boxes2prod (pair2map (a, b)) (pair2map (c, d))

fd2map (Box (Prod (Prod (List String) (List String)) x) ((a, b), c)) 
  =
    boxes2prod (pair2map (a, b)) (fd2map (Box x c))



-- |  (a, b) -> Map (x, (y, ...)) (r, (s, ...))
--    NOTE: fromDistinctAscList.... very dangerous
--
pair2map :: ([String], [String]) -> Box
pair2map (ant, []) = aux (atts2pairs ant) (Box One ()) 
  where                       
    aux (Box t1 ant') (Box t2 cons') = Box (Map t1 t2) (Map.fromDistinctAscList [(ant', cons')])    
pair2map (ant, cons) = aux (atts2pairs ant) (atts2pairs cons)
  where                       
    aux (Box t1 ant') (Box t2 cons') = Box (Map t1 t2) (Map.fromDistinctAscList [(ant', cons')])    



-- [a, b, c, ...] -> (a, (b, (c, ...)))
atts2pairs :: [String] -> Box
--atts2pairs [] = Box One ()
atts2pairs [h] = Box (Tag h String) h
atts2pairs (h:t) = aux h (atts2pairs t) 
  where
    aux h (Box tp r) = Box (Prod (Tag h String) tp) (h, r)





-- need to do the same for data       

-- NOTE: this invariant verification is not working (at least for FDs verification). 
-- probably the invariant os not well written.
-- also perform tests to the other invariants!!

-- recebing the new spreadsheet model (A * B * C) * (D * E) need to 
-- trnasform the data from A * B * C * D * E to it

getData :: MonadPlus m => Box -> R String -> Relation String -> m Box
getData (Box (List a) _) r rl = 
  do
    let ats = getTableAtts (Box (List a) undefined)   
        -- tries to get the index of the att and if can't returns -1
        -- inds = map ((fromJust . (flip elemIndex $ atts r)) . Att) ats
        dats = transProjection rl $ inds ats
    --trace (show dats) return 1        
    rel2prod (Box a undefined) dats
  where
    inds [] = [-1]
    inds ats = map (\a -> maybe (-1) id (elemIndex (Att a) $ atts r)) ats 

getData (Box (TInv (List a) i) _) r rl = 
  do
    let ats = getTableAtts (Box (List a) undefined)
        -- inds = map ((fromJust . (flip elemIndex $ atts r)) . Att) ats
        -- tries to get the index of the att and if can't returns -1
        dats = nub $ transProjection rl (inds ats)
    (Box ttt rrr) <- rel2prod (Box a undefined) dats    
    x@(Box t res) <- rel2prod (Box a undefined) dats
    Eq <- teq t (List a) 
    -- trace (show x) $ return 1
    if (eval i) res then return x else error "data does not good"    
  where
    inds [] = [-1]
    inds ats = map (\a -> maybe (-1) id (elemIndex (Att a) $ atts r)) ats 

getData (Box (Prod a b) _) r rl = 
  do
    (Box t1 d1) <- getData (Box a undefined) r rl
    (Box t2 d2) <- getData (Box b undefined) r rl
    -- trace (show a) $ return 1
    return $ Box (Prod t1 t2) (d1, d2)

getData (Box (TInv (Prod a b) i) _) r rl =
  do
    (Box t1 d1) <- getData (Box a undefined) r rl
    (Box t2 d2) <- getData (Box b undefined) r rl
    Eq <- teq (Prod t1 t2) (Prod a b)
   -- trace (show a) $ return 1
    if (eval i) (d1, d2)
      then return $ Box (Prod t1 t2) (d1, d2)
      else error "data not good"



-- | returns the projection of a relation based on a list of indicies. if a -1 is found
--    it returns a reserved word
--
transProjection :: Relation String -> [Int] -> [[String]]
transProjection [] _ = []
transProjection (h:t) l = do (map (getV h) l) : (transProjection t l)
  where
    getV _ (-1) = "||RESERVED||"
    getV (x:_) 0 = x
    getV (_:xs) n = getV xs (n-1)



-- | for each table returns a list with the atts
--
getTableAtts :: Box -> [String]
getTableAtts (Box (List One) _) = []
getTableAtts (Box (List (Tag s _)) _) = [s]
getTableAtts (Box (List (Prod s r)) _) = (getTableAtts (Box (List s) undefined)) ++
                                         (getTableAtts (Box (List r) undefined))



-- |  transfrom a list of SS rows into a list of RD row acording to the type in the box
--
rel2prod :: MonadPlus m => Box -> Relation String -> m Box
rel2prod box [a] = aux $ list2prod2 box a 
  where aux (Box t v) = return $ Box (List t) [v]

rel2prod box (h:t) = (rel2prod box t) >>= aux (list2prod2 box h)
  where
    aux (Box t1 l) (Box (List t2) ls) = 
      do
        Eq <- teq t1 t2
        return $ Box (List t2) (l:ls)



-- |  it receives a list of strings and return a product of the same strings. this 
--    product is organized according to a type given in the box;
--    this is used to transform a SS row into a DB row 
--
list2prod2 :: Box -> [String] -> Box
list2prod2 (Box t _) [] = Box t undefined
list2prod2 (Box (Tag s String) _) [a] = Box String a
list2prod2 (Box t@(Prod a b) _) l = 
  let (l1, l2) = -- trace (show l) 
                splitAt (len a) l
  in boxes2prod (list2prod2 (Box a undefined) l1) (list2prod2 (Box b undefined) l2)
    where
      len :: Type a -> Int
      len (Prod a b) = len a + len b
      len a = 1
--list2prod2 (Box One _) l = Box One ()      
list2prod2 b r = error $ show b ++ "----" ++ show r





-- |  auxiliar
--

boxes2prod :: Box -> Box -> Box
boxes2prod (Box t1 m1) (Box t2 m2) = Box (Prod t1 t2) (m1, m2)






