module Formulas where


import Debug.Trace
import Data.List
import Data.Maybe

import Sheet.Expr
import Sheet.Sheet
import Sheet.Indx

{-
data Val = I Int
         | F Double
         | S String
         | B Bool
         | E String  -- Error / undefined
         | Unused    -- block out `dead parts'
           deriving Read

data Fml = Inp Val            -- input (by user)
         | Ref Indx           -- direct reference to another cell
         | RefRng Rng         -- this is hack to help parser *** remove this later
	       | Outer Fml          -- this is dummy to pretty-print the "=" for Excel
         | Bin BinOp Fml Fml  -- application of binary operation
         | Agg BinOp Rng
	       | AggL LogOp Rng
         | If Fml Fml Fml
         | Log LogOp Fml Fml
         | Func String [Fml]   -- general case of formula with name and args
         | UnkFml
           deriving (Eq,Ord)

data BinOp = Plus | Minus | Mult | Div | Avg | Exp
             deriving (Eq,Ord,Read)

data LogOp = Eq | Neq | Lt | Gt | And | Or
             deriving (Eq,Ord,Read)
-}


sheet2trigger :: Sheet Fml -> Sheet Fml
sheet2trigger s = fmap (Inp . S . (fml2trigger s)) s


fml2trigger :: Sheet Fml -> Fml -> String
fml2trigger s (Inp (I i)) = show i
fml2trigger s (Inp (F i)) = show i
fml2trigger s (Inp (B i)) = show i

-- aqui tenho the retornar alguma coisa tipo "new.col"...
fml2trigger s (Ref i) = "NEW." ++ getHeader s i
                              
fml2trigger s (Outer f) = fml2trigger s f

fml2trigger s (Bin bop a b) = fml2trigger s a ++ " " ++ show bop ++ " " ++ fml2trigger s b

fml2trigger s (Agg Avg rg) = 
      let len = length rg
      in "(" ++ (foldr (\r res -> fml2trigger s r ++ "+" ++ res) "" rg) ++ ")/" ++ show len
fml2trigger s (Agg aop rg) = foldr (\r res -> fml2trigger s r ++ show aop ++ res) "" rg
fml2trigger s (AggL aop rg) = foldr (\r res -> fml2trigger s r ++ show aop ++ res) "" rg

fml2trigger s (Log lop a b) = fml2trigger s a ++ show lop ++ fml2trigger s b

fml2trigger s (If a b c) = 
    "if(" ++ fml2trigger s a ++ "," ++ fml2trigger s b ++ "," ++ fml2trigger s c ++ ")"

fml2trigger s f = show f    


getHeader :: Sheet Fml -> Indx -> String
getHeader s (Indx c r) = 
                          let Indx c' r' = minIndx s
                          in case s # (Indx c r') of
                                Just c -> stripForm $ stripSpace $ show c
                                otherwise -> error "getHeader: cell not found"
  where
    stripForm s = maybe s id (stripPrefix "FORM" s)
    stripSpace = map (\c -> if c == ' ' then '_' else c)



calcExpr :: Sheet Fml -> Sheet Fml
calcExpr s = fmap (calculateExpr' s) s
  where
    calculateExpr' s c = 
                        let c' = calculateExpr s c
                        in case c' of
                              Just (Left a) -> Inp (F a)
                              Just (Right a) -> Inp (B a)
                              Nothing -> c


calculateExpr :: (Fractional a) => Sheet Fml -> Fml -> Maybe (Either a Bool)

calculateExpr s (Inp (I i)) = Just $ Left $ fromInteger $ toInteger i
--calculateExpr s (Inp (F i)) = Left i 
calculateExpr s (Inp (B i)) = Just $ Right i

calculateExpr s (Ref i) = case s # i of
                              Just c -> calculateExpr s c
                              otherwise -> error "no cell found"

calculateExpr s (Outer f) = calculateExpr s f

calculateExpr s (Bin Div a b) =
  do
    Left a' <- calculateExpr s a
    Left b' <- calculateExpr s b
    Just $ Left $ a' / b'
-- calculateExpr s (Bin Exp a b) = Left $ (fromLeft $ calculateExpr s a) ^ (fromLeft $ calculateExpr s b)
calculateExpr s (Bin bo a b) = do Left a' <- calculateExpr s a
                                  Left b' <- calculateExpr s b
                                  Just $ Left $ (binOp2H bo) a' b'

calculateExpr s (Agg Avg rg) = do r <- mapCalcExp s rg
                                  Just $ Left $ (sum r) / (fromInteger $ toInteger $ length r)
calculateExpr s (Agg bo rg) = do
                                r <- mapCalcExp s rg
                                return $ Left $ foldr1 (binOp2H bo) r

calculateExpr s (AggL bo rg) = do r <- mapCalcExpR s rg
                                  return $ Right $ foldr1 (logOp2H bo) r

calculateExpr s (Log lo a b) = case calculateExpr s a of
                                  Just (Left a') -> fromJL s b >>= return . Right . (lop lo) a'
                                  Just (Right a') -> fromJR s b >>= return . Right . (lop lo) a'
                                  otherwise -> Nothing
  
calculateExpr s (If a b c) = case calculateExpr s a of
                                Just (Right True) -> calculateExpr s b 
                                Just (Right False) -> calculateExpr s c
                                otherwise -> Nothing

calculateExpr _ _ = Nothing

mapCalcExp s = mapM (\e -> calculateExpr s e >>= return . fromLeft)
mapCalcExpR s = mapM (\e -> calculateExpr s e >>= return . fromRight)

fromJL s e = calculateExpr s e >>= return . fromLeft
fromJR s e = calculateExpr s e >>= return . fromRight


fromLeft :: Either a b -> a
fromLeft (Left a) = a

fromRight :: Either a b -> b
fromRight (Right b) = b


logOp2H :: LogOp -> (Bool -> Bool -> Bool)
logOp2H And = (&&)
logOp2H Or = (||)

binOp2H :: (Fractional a) => BinOp -> (a -> a -> a)
binOp2H Plus = (+)
binOp2H Minus = (-)
binOp2H Mult = (*)
binOp2H Div = (/)
-- binOp2H Exp = \a b -> a ^ b

