module HaExcel.GenDB where


import Types hiding (Attribute)
import Data.Transform.Type

import Database.HaskellDB.HSQL.MySQL
import Database.HDBC.Types hiding (SqlType)
import Database.HaskellDB
import Database.HaskellDB.PrimQuery
import Database.HaskellDB.Sql
import Database.HaskellDB.Sql.Print
import Database.HaskellDB.FieldType
import Database.HaskellDB.Database

import Data.List



genDB :: [(Maybe Box, Maybe Box)] -> MySQLOptions -> IO () 
genDB boxs opts = mdb opts (genDBAux boxs)


-- opts = MySQLOptions {server="localhost", db="test", uid="root", pwd="poiqwer"}


mdb :: MySQLOptions -> (Database -> IO a) -> IO a
mdb opts = mysqlConnect opts


genDBAux :: [(Maybe Box, Maybe Box)] -> Database -> IO () 
genDBAux bs db = 
          let zp = filter ((maybe False (const True)) . fst) bs
              sql = map (\(Just b, Just s) -> (b, s)) zp 
              bs' = map (\(b, s) -> (length' b, b, s)) sql
              bs'' = auxs 0 bs'
          in creates (head bs') db
  where
    length' :: Box -> Int
    length' (Box (Prod a b) (a', b')) = 1 + length' (Box b b')
    length' (Box a a') = 1

    auxs _ [] = []
    auxs n [(l, b, s)] = [(l + n, b, s)]
    auxs n ((l, b, s):t) = (l + n, b, s):(auxs (n + l) t)



creates :: (Int, Box, Box) -> Database -> IO () 

creates (i, Box (Map a b) m, bx) db = createTable db ("tbl"++show i) $ (toSqlCol a) ++ (toSqlCol b)

creates (i, Box (Prod a b) (tn, t1), Box (Prod a' b') (tn', t1')) db = do
                                    creates (i, Box a tn, Box a' tn') db
                                    creates (i+1, Box b t1, Box b' t1') db
creates (i, Box (TInv t _) v, Box (TInv t' _) v') db = creates (i, Box t v, Box t' v') db



toSqlCol :: Type a -> [(Attribute, FieldDesc)]
toSqlCol One = []
toSqlCol (Tag "||RESERVED||" s) = [("code", (IntT, False))] 
toSqlCol (Tag s t) = [(stripSpace $ stripForm s, (showT t, True))]
toSqlCol (Prod a b) = toSqlCol a ++ toSqlCol b


showT :: Type a -> FieldType
showT String = StringT 
showT Int    = IntT 


stripSpace = map (\c -> if c == ' ' then '_' else c)

-- | removes the FORM prefix from an att
--
stripForm s = maybe s id (stripPrefix "FORM" s)



