module HaExcel.GenSQL ( 
                        genSQL
                      )
                      where


import qualified Data.Map as Map 
import Data.List as List
import Debug.Trace

import Data.Transform.Type 
import Types


-- |  receives a lits of pair with a box having the type and the values transformed
--    together with the original sheet to allow the ceation of the triggers
--
genSQL :: [(Maybe Box, Maybe Box)] -> [[String]]
genSQL bs = 
          let zp = filter ((maybe False (const True)) . fst) bs
              sql = map (\(Just b, Just s) -> (b, s)) zp 
              bs' = map (\(b, s) -> (length' b, b, s)) sql
              bs'' = auxs 0 bs'
          in map toSQL bs''
  where
    length' :: Box -> Int
    length' (Box (Prod a b) (a', b')) = 1 + length' (Box b b')
    length' (Box a a') = 1

    auxs _ [] = []
    auxs n [(l, b, s)] = [(l + n, b, s)]
    auxs n ((l, b, s):t) = (l + n, b, s):(auxs (n + l) t)

  

-- | need to export the invariants!!?!?!? what to do with them?
--    returns tuple with list of strings with creates, inserts, triggers
--
toSQL :: (Int, Box, Box) -> [String]
-- a table
-- keys and values     
toSQL (i, Box (Map a b) m, s) = 
     ["CREATE table tbl"++show i++" (\n" ++ toVals a ++ toVals b ++ "\tPRIMARY KEY (" ++ (init $ toPK a) ++ ")\n);\n", 
     "INSERT INTO tbl"++show i++" VALUES \n" ++ (init2 $ map2table (Map a b) m) ++ ";\n",
     createTrigger "insert" i s ++ "\n",
     createTrigger "update" i s]

-- a DB
toSQL (i, Box (Prod a b) (tn, t1), Box (Prod a' b') (tn', t1')) = 
                                    let s1 = toSQL (i, Box a tn, Box a' tn') 
                                        s2 = toSQL (i+1, Box b t1, Box b' t1')
                                    in s1 ++ s2
toSQL (i, Box (TInv t _) v, s) = toSQL (i, Box t v, s)


toPK :: Type a -> String 
toPK (Tag "reserved" t) = "code,"
toPK (Tag s t) = s ++ ","
toPK (Prod a b) = toPK a ++ toPK b


showT :: Type a -> String
showT String = "VARCHAR(256)"
showT Int    = "INT"


toVals :: Type a -> String
toVals One = ""
toVals (Tag "reserved" s) = "\tcode int auto_increment,\n"
toVals (Tag s t) = "\t" ++ (stripSpace $ stripForm s) ++ " " ++ showT t ++ ",\n"
toVals (Prod a b) = toVals a ++ toVals b
    
    
stripSpace = map (\c -> if c == ' ' then '_' else c)

-- | removes the FORM prefix from an att
--
stripForm s = maybe s id (stripPrefix "FORM" s)


toValsK :: Type a -> String
toValsK One = ""
toValsK (Tag "reserved" s) = "\tcode int auto_increment,\n"
toValsK (Tag s t) = "\t" ++ s ++ " " ++ showT t ++ ",\n"
toValsK (Prod a b) = toVals a ++ toVals b


-- ver estes casos com o null...
{-
INSERT INTO tbl4 VALUES 
        (null,"36532.0","37134.0","602.0","30100.0"),
CREATE TRIGGER trggrINSERT4 BEFORE INSERT ON tbl4
        FOR EACH ROW BEGIN
                NEW.totaldays = NEW.rentFinish - NEW.rentStart where code = null;
-}

-- | receives a string with "UPDATE" or "INSERT", an int to index the trigger and the
--    table and a box with the 
--    data of the SS already in SQL. returns the trigger 
--
createTrigger :: String -> Int -> Box -> String
createTrigger iu i b =
                let tgs = toTrigger b
                in if tgs == "" then ""
                      else
                "CREATE TRIGGER trggr" ++ iu ++ show i ++ " BEFORE "++iu++" ON tbl" ++ show i ++ "\n" ++
                  "\tFOR EACH ROW BEGIN\n" ++
                  tgs ++
                  "\tEND;"


-- | creates each line of the trigger
-- for example ["NEW.totaldays = NEW.rentFinish - NEW.rentStart;"]
--
toTrigger :: Box -> String
totrigger (Box (Map _ One) _) = ""
toTrigger (Box (Map k (Prod a b)) m) = 
                                  let m1 = Map.map fst m
                                      m2 = Map.map snd m
                                  in toTrigger (Box (Map k a) m1) ++ toTrigger (Box (Map k b) m2) 
toTrigger (Box (Map k (Tag an t)) v) | isPrefixOf "FORM" an =
                                           let kv = getkv (Map k (Tag an t)) v
                                               an' = stripSpace $ stripForm an
                                               new = "\t\tSET NEW." ++ an' ++ " = "
                                           in concat $ zipWith (++) (repeat new) kv
                                     | otherwise = ""
toTrigger b = show b                                     
                    
-- | gives a list of strings with the trigger construction operation                    
--     for example, [NEW.rentFinish - NEW.rentStart;"]
--
getkv :: Type a -> a -> [String]
getkv (Map k (Tag an String)) v = 
                              let vals = Map.elems v
                              in if length (nub vals) == 1 -- if the formulas is always the same
                                        then [aux1 vals]
                                        else aux2 vals
  where
    aux1 vals = head vals ++ ";\n"

    aux2 vals = 
           let ks = init $ showKA k
               kys = map ((++";\n") . init . (plain  k)) (Map.keys v)
               f = \a b c -> a ++ b ++ c 
           in zipWith3 f vals (repeat (" where " ++ ks ++ " = ")) kys


-- | just constructs a string with all the key att separeted by a ,
--
showKA :: Type a -> String
showKA One = ""
showKA (Tag "reserved" s) = "code,"
showKA (Tag s String) = s ++ ","
showKA (Prod a b) = showKA a ++ "," ++ showKA b




map2table :: Type t -> t -> String
map2table (Map a b) = (concatMap ((++ "),\n") . init4 . row2row (Prod a b))) . Map.toList
  where
    row2row :: Type a -> a -> String
    row2row (Prod a One) (k, _) = "\t(" ++ plain a k ++ "),\n"
    row2row (Prod a b) (k, v) = "\t(" ++ plain a k ++ plain b v ++ "),\n"

plain :: Type a -> a -> String
plain (Prod a b) (a', b') = showV a a' ++ "," ++ plain b b' 
plain (Tag "reserved" String) r = "null,"
plain (Tag s String) r = showV String r ++ "," 


showV :: Type a -> a -> String
showV String s = show $ s
showV (Tag s String) r = show $ r



init2 [] = []
init2 l = init . init $ l
init4 [] = []
init4 l = init . init . init . init $ l
