module HaExcel.GenSS (
                        genSS
                     ) 
                        where

import Data.Maybe

import Language.Gnumeric.Syntax

import Types
import Data.Transform.Type


-- | wrapper to the genSSAux, the main function
--
genSS :: [Maybe Box] -> Gmr'Workbook -> Gmr'Workbook
genSS boxs wb = genSSAux boxs ([], []) (0, 0) wb


-- | uses the below genSS1 function to create the cells and regions to each box with SS tables
--
genSSAux :: [Maybe Box] -> ([Gmr'Cell], [Gmr'StyleRegion]) -> (Int, Int) -> Gmr'Workbook 
          -> Gmr'Workbook
genSSAux [] (l, stl) _ wb = subsStyle (subsCells wb (Gmr'Cells l)) (Gmr'Styles stl)
genSSAux (Nothing:t) l (sc, sr) wb = genSSAux t l (sc, sr) wb
genSSAux ((Just h):t) (l, stl) (sc, sr) wb =
    let (sc', sr', l', st) = genSS1 h 0 (sc, sr)
    in genSSAux t (l++l', st) (sc', sr'+1) wb


-- | for a given box with (SS) tables on it, a starting column to write, a pair of column, row
--    where to write each values, it returns the nest column and row to wirte the
--    next table. it also returns the list of created cells and the list of style
--    regions to make each table with borders
genSS1 :: Box -> Int -> (Int, Int) -> (Int, Int, [Gmr'Cell], [Gmr'StyleRegion])
genSS1 (Box (List a) []) col (sc, sr) = (sc, sr+1, [], [])
genSS1 (Box (List a) (x:xs)) col (sc, sr) = 
    let (sc', sr', l', st) = genSS1 (Box a x) col (col, sr) 
        (sc'', sr'', l'', st') = genSS1 (Box (List a) xs) col (col, sr+1)
    in (sc'', sr'', l' ++ l'', st++st')    
genSS1 (Box (Prod a b) (a', b')) col (sc, sr) =
    let (sc', sr', l', st) = genSS1 (Box a a') col (sc, sr) 
        (sc'', sr'', l'', st') = genSS1 (Box b b') col (sc', sr')
        stl' = createStyle (show sc, show sr) (show sc', show sr') 
    in (sc'', sr'', l' ++ l'', stl':st++st')
genSS1 (Box (Tag s String) st) col (sc, sr) = (sc+1, sr, [createCell (show sc, show sr) st], [])
genSS1 (Box String st) col (sc, sr) = (sc+1, sr, [createCell (show sc, show sr) st], [])
genSS1 b _ _ = error $ show b


-- | returns a cell without any information for a certain column c and a cetain row r 
--
--createCell :: (S -> Gmr'Cell
createCell (c, r) v =
    Gmr'Cell(Gmr'Cell_Attrs c r Nothing Nothing Nothing Nothing Nothing Nothing) v


-- | substitutes the cells from the first sheet in the workbook with the given ones
--
subsCells :: Gmr'Workbook -> Gmr'Cells -> Gmr'Workbook
subsCells (Gmr'Workbook a b c d e f g h i shts k l) cls = 
    Gmr'Workbook a b c d e f g h i (shts' shts) k l
  where
    shts' (Gmr'Sheets ((Gmr'Sheet a b c d e f g h i j k m _ n o p q):_)) = 
      Gmr'Sheets [Gmr'Sheet a b c d e f g h i j k m cls n o p q]


-- | substitutes the style from the first sheet with the one given
--
subsStyle :: Gmr'Workbook -> Gmr'Styles -> Gmr'Workbook
subsStyle (Gmr'Workbook a b c d e f g h i shts k l) stl = 
    Gmr'Workbook a b c d e f g h i (shts' shts) k l
  where
    shts' (Gmr'Sheets ((Gmr'Sheet a b c d e f g _ i j k m n o p q r):_)) = 
      Gmr'Sheets [Gmr'Sheet a b c d e f g stl i j k m n o p q r]


-- | creates a style starting at the first pair (column, row) and ending in the other pair
--    each cell with have a border
--    
createStyle :: (String, String) -> (String, String) -> Gmr'StyleRegion
createStyle (sc, sr) (ec, er) = Gmr'StyleRegion stAtts stl
  where
    stAtts = Gmr'StyleRegion_Attrs {  gmr'StyleRegionStartCol = sc,
                                  gmr'StyleRegionStartRow = sr,
                                  gmr'StyleRegionEndCol = ec,
                                  gmr'StyleRegionEndRow = er }

    stl = Gmr'Style stlAtts font (Just stlBorder)

    stlBorder = Gmr'StyleBorder top' bottom left right diag revDiag

    top' = Gmr'Top {gmr'TopStyle = "5", gmr'TopColor = Just "0:0:0"}
    bottom = Gmr'Bottom {gmr'BottomStyle = "5", gmr'BottomColor = Just "0:0:0"}
    left = Gmr'Left {gmr'LeftStyle = "5", gmr'LeftColor = Just "0:0:0"}
    right = Gmr'Right {gmr'RightStyle = "5", gmr'RightColor = Just "0:0:0"}
    diag = Gmr'Diagonal {gmr'DiagonalStyle = "0", gmr'DiagonalColor = Nothing}
    revDiag = Gmr'Rev_Diagonal {gmr'Rev_DiagonalStyle = "0", gmr'Rev_DiagonalColor = Nothing}

    font = Gmr'Font fontAttrs "Sans"

    fontAttrs = Gmr'Font_Attrs { gmr'FontUnit = "12", gmr'FontBold = "0",
                                 gmr'FontItalic = "0", gmr'FontUnderline = "0",
                                 gmr'FontStrikeThrough = "0"}

    stlAtts = Gmr'Style_Attrs { gmr'StyleHAlign = Just "1", gmr'StyleVAlign = Just "2",
                                gmr'StyleWrapText = Just "0", gmr'StyleOrient = Nothing,
                                gmr'StyleRotation = Just "0", gmr'StyleShade = Just "0",
                                gmr'StyleShrinkToFit = Just "0", gmr'StyleIndent = Just "0",
                                gmr'StyleFormat = "General", gmr'StyleFore = Just "0:0:0",
                                gmr'StyleHidden = Just "0", gmr'StyleBack = Just "FFFF:FFFF:FFFF",
                                gmr'StyleLocked = Just "1", gmr'StylePatternColor = Just "0:0:0"}
