module HaExcel.ReadSS -- (
                        -- readSS,
                    --  )
                      where


import Types
import HaExcel.FDS2Type
import Formulas
import Data.Transform.Type hiding (Func, Eq, teq)
import Data.Transform.Rule hiding (teq, Func)

import Language.Gnumeric.Syntax
import Language.Gnumeric.Formula 
import Language.Gnumeric.Parser (readXml')
import StrategyLib

import Sheet.Sheet
import Sheet.Indx
import Sheet.Expr hiding (Eq, Ref)
import UCheck.MParser
import UCheck.Header
import UCheck.HdrAux
import UCheck.UnitErrors
import UCheck.Units
import UCheck.FaultLoc (faultLoc)

import Text.XML.HaXml.Xml2Haskell hiding (Default)
import Data.Time.Calendar (showGregorian, Day(..)) 

import System
import System.IO
import Data.List as List
import Debug.Trace
import Data.Map as Map hiding (map, filter)
import Data.Char


-- TODO
-- refact all this module!!!!!!!!!!!
-- melhorar a inferencia dos attributos, agora so estou a assumir que e' a primeira linha;
-- melhorar o sistema de contagem das celulas numericas;
-- tratar as tableas mais numericas de outra forma (ja pensado em papel);
-- nao esta mt bem pensado quando a tabela nao e' um rectangulo;
-- ver o que faco com as formulas - no caso de att derivados fazer trigger para fazer o
--    seu update;


readSS :: FilePath -> IO [Sheet Fml]
readSS fp = do  -- parsing the file
     (wb ::  Gmr'Workbook) <- readGnumeric' fp
   --  hPutXml stdout (wb :: Gmr'Workbook) -- funciona :)
                  -- remove empty sheets and take just the cells
     let cs = map (cells . completeFormulas) $ 
                  filter (\sh -> (takeCells $ cells sh) /= []) $ sheets wb
         uc = map annot $ concatMap (tables . gnu2ucheck) cs                    
         -- uc = concatMap (tables . gnu2ucheck) cs                    
     return uc    


--readGnu2 :: FilePath -> IO [(R String, Relation String)]
readGnu fp = do  -- parsing the file
     (wb ::  Gmr'Workbook) <- readGnumeric' fp
                  -- remove empty sheets and take just the cells
     let cs = map (cells . completeFormulas) $ 
                  filter (\sh -> (takeCells $ cells sh) /= []) $ sheets wb
           -- transform this list of cells into Sheet (Erwig)  
   -- infers all the tables in all sheets (NOTE: do not leave empty rows or columns in a table
               -- otherwise it will think that it is 2 tables -- improve this!!!!!!!!!!!!)
         uc = map annot $ concatMap (tables . gnu2ucheck) cs                    
                    -- and calculates the formulas wheneverpossible
         ucf = map calcExpr uc
           -- decide which metod to use: if it more than half are numerical, then erwig, 
            -- otherwise ours
         (emet, mmet) = break (\sh -> ((howManyInts sh) * 2) <= (length $ unSheet sh)) ucf
         mmet' = zip (map sheet2CompleteList mmet) uc
         mmet'' = map (\(l, l') -> (Atts $ map Att (head l), tail l, l')) mmet'
     return mmet''

     
showCells (Gmr'Cells []) = return ""
showCells (Gmr'Cells (h:t)) = putStrLn (show h) >> showCells (Gmr'Cells t)



-- | this function annotates an attribute with a prefix "FORM" if its values are expressions
--    use this function on a table
--    this function can be much improved in speed...
--
annot :: Sheet Fml -> Sheet Fml
annot (Sheet s) = 
                   Sheet $ antAux (minIndx $ Sheet s) (maxIndx $ Sheet s) s
  where
antAux ind@(Indx c r) mx@(Indx c' r') s | c > c' = s
                                        | otherwise =
                   case (Sheet s) # ind of
                      Just v -> if isExpr s (getCol (Indx c (r+1)) s) 
                                  then antAux (Indx (c+1) r) mx (antHead ind s) 
                                  else antAux (Indx (c+1) r) mx s
                      otherwise -> error $ "annot: cell " ++ show ind ++ " not found!"

getCol ind@(Indx c r) = filter (\(Indx c' r', _) -> c == c' && r' > r)

isExpr s col = foldr (\(_, e) r -> isExpr' s e && r) True col

isExpr' :: [Cell Fml] -> Fml -> Bool
--    isExpr' s (Ref r) = case ((Sheet s) # r) of
  --                          Just v -> error "ok" -- isExpr' s v
    --                        otherwise -> error "isExpr': cell not found"
isExpr' s (Outer f) = isExpr' s f
isExpr' _ (Bin _ _ _) = True
isExpr' _ (Agg _ _) = True
isExpr' _ (AggL _ _) = True
isExpr' _ (If _ _ _) = True
isExpr' _ (Log _ _ _) = True
isExpr' _ (Func _ _) = True
isExpr' _ _ = False

antHead ind@(Indx c r) s = 
                              case (Sheet s) # ind of
                                  Just (Inp (S st)) -> (ind, Inp $ S $ "FORM"++st):
                                                       (filter (\(i, v) -> i /= ind) s)
                                  otherwise -> error "antHead: cell not found"



type AllForms = Map String (String, String, String)

-- |  only the formula at the most upper left corner as value. all the others, based 
--    on that one, are constructed. this function makes that construction
completeFormulas :: Gmr'Sheet -> Gmr'Sheet
completeFormulas sh = let allForms = collectFormulas sh
                          (Gmr'Sheet a b c d e f g h i j k l (Gmr'Cells cs) n o p q) = sh
                          cs' = map (expandFormula allForms) cs
                      in (Gmr'Sheet a b c d e f g h i j k l (Gmr'Cells cs') n o p q)


-- |  collects all formulas from a GmrSheet into a map with key the Expr ID and
--    values a tripple with its column, row and value (formula)
collectFormulas :: Gmr'Sheet -> AllForms
collectFormulas sh = let Gmr'Cells cels = cells sh
                     in foldr fun Map.empty cels
  where
    -- not formula, just exprid
    fun (Gmr'Cell (Gmr'Cell_Attrs c r (Just exprId) _ _ _ _ _) "") m = m
    fun (Gmr'Cell (Gmr'Cell_Attrs c r (Just exprId) _ _ _ _ _) val) m = Map.insert exprId (c, r, val) m
    fun _ m = m


-- |  increses the col and row in c and r in cells with ExprID but no value (formula)
expandFormula :: AllForms -> Gmr'Cell -> Gmr'Cell
expandFormula forms (Gmr'Cell (Gmr'Cell_Attrs c r (Just expId) valT valF cs rs ast) "") = 
    let (c', r', exp) = forms ! expId
        exp2 = parseFormula "" "" exp
        exp3 = incCR  ((read c) - (read c')) ((read r) - (read r')) exp2
    in Gmr'Cell (Gmr'Cell_Attrs c r (Just expId) valT valF cs rs ast) (showForm exp3)
expandFormula _ c = c


-- |  transform a Formula into a string just like the one in the Gnumeric
showForm :: Formula -> String
showForm (Var s) = show s
showForm (Fun s [f1, f2]) = "=" ++ showForm f1 ++ s ++ showForm f2
showForm (Fun s fs) = "=" ++ s ++ "(" ++ (init $ foldr (\f s -> f++","++s) "" $ map showForm fs) ++")"
showForm (Ref cr Nothing) = showCellRef' cr 
showForm (Ref cr (Just mcr)) = showCellRef' cr ++ ":" ++ showCellRef' mcr
    
showCellRef' cr | sheet cr == "" = (fromColNr $ colNr cr) ++ (fromRowNr $ rowNr cr)
                | otherwise = "|" ++ (sheet cr) ++ "!" ++ (fromColNr $ colNr cr) ++ (fromRowNr $ rowNr cr)

-- |  increses the col and row in c and r
incCR :: Int -> Int -> Formula -> Formula
incCR _ _ (Var s) = Var s
incCR c r (Fun s fs) = Fun s $ map (incCR c r) fs
incCR c r (Ref cr Nothing) = Ref (incRef c r cr) Nothing
incCR c r (Ref cr (Just mcr)) = Ref (incRef c r cr) (Just $ incRef c r  mcr)

-- |  increses the col and row in c and r
incRef :: Int -> Int -> CellRef -> CellRef
incRef c r (CellRef s c' r') = CellRef s (c' + c) (r' + r)




file2UCheck fp = do  -- parsing the file
     (wb ::  Gmr'Workbook) <- readGnumeric' fp
                  -- remove empty sheets and take just the cells
     let cs = map upperCells $
                map cells $ 
                  filter (\sh -> (takeCells $ cells sh) /= []) $ 
                    map completeFormulas $ sheets wb
                    -- transform this list of cells into Sheet (Erwig)
     return $ -- map (\l -> (Atts $ map Att (head l), tail l)) $ map sheet2CompleteList $ concatMap tables $ map (calcExpr . annot . gnu2ucheck) cs
              concatMap (tables) $ map ( gnu2ucheck) cs
    -- return $ map gnu2ucheck cs
     

-- |  the UCheck system wants to see functions in upper case.... but just the ones
--    it knows like AVERAGE, SUM, PROD. so it is not entirely needed
--
upperCells :: Gmr'Cells -> Gmr'Cells
upperCells (Gmr'Cells cls) = Gmr'Cells $ map upperCell cls

-- |  just upper case the cells with formulas
-- 
upperCell :: Gmr'Cell -> Gmr'Cell
upperCell (Gmr'Cell ats val@('=':v)) = -- trace ("val: " ++ show val) 
  Gmr'Cell ats $ map toUpper val
upperCell c = c    




-- | Reads a gnumeric file into a workbook
--
readGnumeric' :: (Term a, XmlContent a) => FilePath -> IO a
readGnumeric' fileName = do
  let tmpFile = "/tmp/TestGnumericTEMPFILE.Z"
  let tmpFile2 = "/tmp/TestGnumericTEMPFILE"
  -- gunzip -c -S "" file.qqextention > /tmp/xddd
  let systemCmd = "gunzip -c -S \"\" " ++ fileName ++ " > " ++ tmpFile2
  --"cp "++fileName++" "++tmpFile++" && zcat "++tmpFile++" > "++tmpFile2
 --  putStrLn systemCmd
  system systemCmd
  content <- readFile tmpFile2
  let parseResult = {- parseFormulas $ -} readXml' content
  case (parseResult) of
    (Just workbook) -> return workbook
    Nothing -> error "NO PARSE"



-- | removes the first row from a Sheet
--
dropFirstRow (Sheet sh) = let Indx minc minr = minIndx (Sheet sh)
                                  in Sheet $ filter (\(Indx c r, _) -> r /= minr) sh  
                          


-- | counts the number of fomulas and numerical cells in a Sheet
--   assuming that a formula returns a numerical value...
--   there are more cases to consider******************
--
howManyInts :: Sheet Fml -> Int
howManyInts s = length $ unSheet (rngFilter (\v -> isNum v || isFormula v) s) 

isFormula :: Fml -> Bool
isFormula (Func _ _) = True
isFormula _ = False


-- Gmr'Cell_ValueType_60 representa text, acho eu
-- Gmr'Cell_ValueType_40 float, double, ...
-- Gmr'Cell_ValueType_30 number
-- Gmr'Cell_ValueType_30, gmr'CellValueFormat = Just "m/d/yy"  datas

-- showGregorian :: Day -> String	Source
-- show in ISO 8601 format (yyyy-mm-dd) 

-- *Main> Data.Time.Calendar.showGregorian ( Data.Time.Calendar.ModifiedJulianDay (15018+36809))
-- "2000-10-10"



-- | tranforms a list of cells into a sheet (Erwig)
--    think better in other types like dates. maybe create Type constructure to them...
--
gnu2ucheck :: Gmr'Cells -> Sheet Fml
gnu2ucheck (Gmr'Cells l) = foldr fun (Sheet []) l
  where
  {-
  -- date, need to be improved...  
    fun (Gmr'Cell (Gmr'Cell_Attrs c r _ (Just Gmr'Cell_ValueType_30) (Just df) _ _ _) value) =
         msgProcess ((showcr c r) ++ 
                      "\"" ++ (showGregorian $ ModifiedJulianDay $ 15018 + (read value)) ++ "\"")
  -- date too                      
    fun (Gmr'Cell (Gmr'Cell_Attrs c r _ (Just Gmr'Cell_ValueType_40) (Just df) _ _ _) value) =
         msgProcess ((showcr c r) ++ 
                      "\"" ++ (showGregorian $ ModifiedJulianDay $ 15018 + (read value)) ++ "\"")
  -}                      
  -- all the others  
    fun (Gmr'Cell (Gmr'Cell_Attrs c r _ valType _ _ _ _) value) =
              msgProcess ((showcr c r) ++ (funAux valType value)) 
  
    funAux valType value = case valType of
                            Just Gmr'Cell_ValueType_60 -> "\"" ++ value  ++ "\""  -- text
                            Just Gmr'Cell_ValueType_30 -> value                   -- number
                            Just Gmr'Cell_ValueType_40 -> value                   -- double
                            otherwise                  -> value                  -- other, like formulas
    
    showcr c r = "c " ++ (cr c) ++ " " ++ (cr r) ++ " "
    cr c = show $ (read c) + 1



-- | receives a Sheet and returns a list of lists of strings with its content
--   NOTE: this function must be reviewed becouse the init shouldn't be necessary
sheet2CompleteList :: Sheet Fml -> [[String]]
sheet2CompleteList sh = 
    let Indx minc minr = minIndx sh 
        Indx maxc maxr = maxIndx sh 
    in init $ aux1 minc minr minc minr maxc maxr [] sh
  where
    aux1 startc startr c r nCols nRows l sh 
      | c > nCols = l:(aux1 startc startr startc (r+1) nCols nRows [] sh)
      | r > nRows = [l]
      | otherwise = case sh # (Indx c r) of
                        Nothing -> aux1 startc startr (c+1) r nCols nRows (l++[""]) sh
                        Just fml -> aux1 startc startr (c+1) r nCols nRows (l++[show fml]) sh


-- | Selector for list stored in field sheets
--
sheets :: Language.Gnumeric.Syntax.Gmr'Workbook -> [Gmr'Sheet]
sheets (Gmr'Workbook a b c d e f g h i (Gmr'Sheets sts) k l) = sts


-- | Selector for field cells
--
cells :: Language.Gnumeric.Syntax.Gmr'Sheet -> Gmr'Cells
cells (Gmr'Sheet a b c d e f g h i j k l cs n o p q) = cs


takeCells (Gmr'Cells l) = l

putCells l = Gmr'Cells l









-- NOT USED



-- | after reading the file it is tranformed into a Sheet. After discovered the
--   tables, each table is passed to this function and it will return the values
--   in products of products (in the format to the 2LT)
--   it starts in the first row but in the last column in order to create 
--   right associated products
--
sheet2box :: Sheet Fml -> Box
sheet2box sh = 
  let Indx minc minr = minIndx sh 
      Indx maxc maxr = maxIndx sh
  in aux1 maxc minr (maxc-1) minr minc maxr (aux3 sh maxc minr) sh
  where
    aux1 startc startr c r nCols nRows (Box pt p) sh 
      | c < nCols && r == nRows = Box pt p   
      | c < nCols = 
          aux2 (Box pt p) (aux1 startc startr (startc-1) (r+1) nCols nRows (aux3 sh startc (r+1)) sh)
      | r > nRows = Box pt p 
      | otherwise = 
        case sh # (Indx c r) of
          Nothing -> aux1 startc startr (c-1) r nCols nRows (Box (Prod String pt) ("", p)) sh
          Just (Inp (I i)) -> aux1 startc startr (c-1) r nCols nRows (Box (Prod Int pt) (i, p)) sh
          Just (Inp (F i)) -> aux1 startc startr (c-1) r nCols nRows (Box (Prod Double pt) (i, p)) sh
          Just (Inp (B i)) -> aux1 startc startr (c-1) r nCols nRows (Box (Prod Bool pt) (i, p)) sh
          Just (Inp (S i)) -> aux1 startc startr (c-1) r nCols nRows (Box (Prod String pt) (i, p)) sh
          Just (Inp (F i)) -> aux1 startc startr (c-1) r nCols nRows (Box (Prod Double pt) (i, p)) sh
          Just (Inp (E i)) -> aux1 startc startr (c-1) r nCols nRows (Box (Prod String pt) (i, p)) sh
          Just (Inp Unused) -> aux1 startc startr (c-1) r nCols nRows (Box (Prod String pt) ("", p)) sh
          Just fml -> aux1 startc startr (c-1) r nCols nRows (Box (Prod String pt) (show fml, p)) sh

    aux3 sh c r = case sh # (Indx c r) of
          Nothing -> Box String ""
          Just (Inp (I i)) -> Box Int i
          Just (Inp (F i)) -> Box Double i
          Just fml -> Box String (show fml)

    aux2 (Box pt p) (Box pt' p') = Box (Prod pt pt') (p, p') 


-- | the same has sheet2box but returns list of products, not products of products
--   refact this function...
--sheet2box2 :: Monad m => Sheet Fml -> m Box
sheet2box2 sh = 
  let Indx minc minr = minIndx sh 
      Indx maxc maxr = maxIndx sh
      b = aux4 (maxc-1) minr minc (aux3 sh maxc minr) sh
  in  aux1 maxc (maxc-1) (minr+1) minc maxr b (aux3 sh maxc (minr+1)) sh
  where
    aux4 c r nCols (Box pt p) sh 
      | c < nCols = Box (List pt) [p]
      | otherwise = case sh # (Indx c r) of
          Nothing -> aux4 (c-1) r nCols (Box (Prod String pt) ("", p)) sh
          Just (Inp (I i)) -> aux4 (c-1) r nCols (Box (Prod Int pt) (i, p)) sh
          Just (Inp (F i)) -> aux4 (c-1) r nCols (Box (Prod Double pt) (i, p)) sh
          Just (Inp (B i)) -> aux4 (c-1) r nCols (Box (Prod Bool pt) (i, p)) sh
          Just (Inp (S i)) -> aux4 (c-1) r nCols (Box (Prod String pt) (i, p)) sh
          Just (Inp (F i)) -> aux4 (c-1) r nCols (Box (Prod Double pt) (i, p)) sh
          Just (Inp (E i)) -> aux4 (c-1) r nCols (Box (Prod String pt) (i, p)) sh
          Just (Inp Unused) -> aux4 (c-1) r nCols (Box (Prod String pt) ("", p)) sh
          Just fml -> aux4 (c-1) r nCols (Box (Prod String pt) (show fml, p)) sh 

    aux1 startc c r nCols nRows bx@(Box pt' p') (Box pt p) sh 
      | c < nCols && r == nRows = do
             Eq <- teq (List pt) pt'
             return $ Box pt' (p' ++ [p])   
      | c < nCols = do 
             bx' <- aux2 bx (Box pt p)
             aux1 startc (startc-1) (r+1) nCols nRows bx' (aux3 sh startc (r+1)) sh
      | otherwise = case sh # (Indx c r) of
          Nothing -> aux1 startc (c-1) r nCols nRows bx (Box (Prod String pt) ("", p)) sh
          Just (Inp (I i)) -> aux1 startc (c-1) r nCols nRows bx (Box (Prod Int pt) (i, p)) sh
          Just (Inp (F i)) -> aux1 startc (c-1) r nCols nRows bx (Box (Prod Double pt) (i, p)) sh
          Just (Inp (B i)) -> aux1 startc (c-1) r nCols nRows bx (Box (Prod Bool pt) (i, p)) sh
          Just (Inp (S i)) -> aux1 startc (c-1) r nCols nRows bx (Box (Prod String pt) (i, p)) sh
          Just (Inp (F i)) -> aux1 startc (c-1) r nCols nRows bx (Box (Prod Double pt) (i, p)) sh
          Just (Inp (E i)) -> aux1 startc (c-1) r nCols nRows bx (Box (Prod String pt) (i, p)) sh
          Just (Inp Unused) -> aux1 startc (c-1) r nCols nRows bx (Box (Prod String pt) ("", p)) sh
          Just fml -> aux1 startc (c-1) r nCols nRows bx (Box (Prod String pt) (show fml, p)) sh
 
    aux3 sh c r = case sh # (Indx c r) of
          Nothing -> Box String ""
          Just (Inp (I i)) -> Box Int i
          Just (Inp (F i)) -> Box Double i
          Just fml -> Box String (show fml)

    aux2 (Box (List pt) p) (Box (List pt') p') = do 
                                                  Eq <-teq (List pt) (List pt')
                                                  return $ Box (List pt') (p ++ p')
    aux2 (Box (List pt) p) (Box pt' p') = do 
                                                  Eq <-teq (List pt) (List pt')
                                                  return $ Box (List pt') (p ++ [p'])



--gnu2fds :: Sheet Fml -> (R a, Relation b)
-- gnu2fds :: MaxCol -> Gmr'Cells -> (R a, Relation b)
gnu2fds mc (Gmr'Cells ss) = let (ats, vals) = splitAt (mc+1) ss
                            in (Atts $ toats ats, tovals vals)
  where
    toats [] = []
    toats ((Gmr'Cell (Gmr'Cell_Attrs _ _ _ _ _ _ _ _) value):ats) = (Att value):(toats ats)
    
    tovals vals = map fromCell $ groupBy lins vals

    lins (Gmr'Cell (Gmr'Cell_Attrs c r _ _ _ _ _ _) _) 
         (Gmr'Cell (Gmr'Cell_Attrs c' r' _ _ _ _ _ _) _) = r == r'

    fromCell [] = []
    fromCell ((Gmr'Cell (Gmr'Cell_Attrs _ _ _ _ _ _ _ _) v):cls) = v:(fromCell cls)

--do
  --let (b:h:r) = tables ss
  --return (Atts $ map Att (head $ sheet2list h), sheet2list b)  





-- | receives a list of cells and returns its values
--
-- values :: Gmr'Cells -> [a]
values (Gmr'Cells l) = map (\(Gmr'Cell (Gmr'Cell_Attrs col row _ _ _ _ _ _) value) -> value) l


-- | returns the max number of columns
--
maxCol :: Gmr'Sheet -> Int
maxCol (Gmr'Sheet a b (Just (Gmr'MaxCol mc)) d e f g h i j k l cs n o p q) = read mc -- :: Int
-- maxCol sh = read $ maximum $ map fst $ colrow $ cells sh 

-- | reads a Gnumeric spreadsheet and returns a Sheet (Erwig)
--   the file must have the .gnumeric.Z extension
--readGnu :: FilePath -> IO (Sheet Fml)
readGnu2 fp = do 
                (wb ::  Gmr'Workbook) <- readGnumeric' fp
                let cs = cells (head $ sheets wb ) -- the head assumes 1 sheet
                    shts = sheets wb
                    mc = maxCol $ head shts
                    cplt = completess2 mc cs
                    -- x = toUCheckSheet cplt
                return $ gnu2fds mc cplt -- x 


-- | receives a Gmr'Cells and returns a Sheet (Erwig) 
--
toUCheckSheet :: Gmr'Cells -> Sheet Fml
toUCheckSheet (Gmr'Cells l) = Sheet (map toCell l)
  where
    toCell (Gmr'Cell (Gmr'Cell_Attrs col row _ _ _ _ _ _) value) 
      = 
        (Indx ((read col)+1) ((read row)+1), Inp (S value))



-- | receives a sheet (onlu the cells) and its maximal column and returns it
--   complete, that is all the cells mist exist. the ones that doesn't exist
--   are created with an empty string on it and without any attribute
--
completess2 :: Int -> Gmr'Cells -> Gmr'Cells
completess2 maxCol (Gmr'Cells l) = let ls = groupBy cols l
                                   in Gmr'Cells $ concatMap (complet 0 maxCol) ls
  where
    cols (Gmr'Cell (Gmr'Cell_Attrs col row _ _ _ _ _ _) _) 
         (Gmr'Cell (Gmr'Cell_Attrs col' row' _ _ _ _ _ _) _) = row == row'

    complet n maxCol [cell@(Gmr'Cell (Gmr'Cell_Attrs _ row _ _ _ _ _ _) _)] = 
                                                    cell:(replicate' (maxCol-n) (n+1) row)

    complet n maxCol (cell@(Gmr'Cell (Gmr'Cell_Attrs col row _ _ _ _ _ _) _):r) =
      let colnr =  read col :: Int
      in if colnr == maxCol
           then 
             [] 
           else
             if colnr == n
              then
                cell:(complet (n+1) maxCol r)
              else 
                (replicate' (colnr-n) n row)++(cell:(complet (colnr+1) maxCol r))

    replicate' n c r | n > 0 = (emptyCell (show c) r):(replicate' (n-1) (c+1) r)
                     | otherwise = []


-- | returns a cell without any information for a certain column c and a cetain row r 
--
emptyCell :: String -> String -> Gmr'Cell
emptyCell c r = 
      Gmr'Cell (Gmr'Cell_Attrs c r Nothing Nothing Nothing Nothing Nothing Nothing) "" 


-- | returns a list of pairs of the columns and rows  of a list of celss
--
colrow :: Gmr'Cells -> [(String, String)]
colrow (Gmr'Cells s) = map cr s
  where
    cr ((Gmr'Cell (Gmr'Cell_Attrs col row _ _ _ _ _ _) _)) = (col, row)



-- | given a sheet returns a list of rows with the data
--
sheet2list :: Sheet Fml -> [[String]]
sheet2list (Sheet sh) 
  =
    let rows = toRows $ sortBy sortsh sh
    in map (map (fromFml.snd)) rows
  where
    toRows = groupBy (\(Indx c r, v) (Indx c' r', v') -> r == r')

    sortsh (Indx c r, v) (Indx c' r', v') | r == r' = EQ
                                          | r < r' = LT
                                          | otherwise = GT
--    fromFml (Inp (S s)) = s     
    fromFml = show

    {-
    fromFml (Inp (F s)) = show s     
    fromFml (Inp (I s)) = show s     
    fromFml (Inp (N s)) = show s     
    fromFml (Inp (E s)) = s     
    fromFml (Inp Unused) = ""     
    fromFml (Inp inp) = mkT (fromFmlStr)
  -}

{-
instance Show Fml where
  show (Inp v)     = show v
  show (Ref ix)    = show ix
  show (Outer f)   = '=':show f
  show (Bin b f g) = "("++show f++" "++show b++" "++show g++")"
  show (Agg b r)   = fshow b++"("++rshow r++")"
  show (AggL l r)   = show l++"("++rshow r++")"
  show (If b e1 e2) = "IF("++show b++","++show e1++","++show e2++")"
  show (Log l e1 e2) = show e1++show l++show e2 
  show (Func s rs) = s++"("++rshow rs++")"
  show (UnkFml) = "XXX"
  show (RefRng rs) = concatMap show rs

data Val = I Int
         | F Double
         | S String
         | B Bool
         | E String  -- Error / undefined
         | Unused   
-}


-- | a test based on Erwig stuff
--
readdata :: String -> Sheet Fml -> IO [Char]
readdata str sheet = do
            case str of 
                  "kill server" -> do {
                     return "kill";
                     };
                  "Clear" -> do {
                 --   putStrLn "Clear";
                    return "Clear";
                     --(readdata (Sheet []));
                     };
                  "Testing" -> do { 
                     --putStrLn ("Server read: " ++ str ++ show totalbytes);
                     --writ <- send sock str;
                   --  (readdata (totalbytes+(length $! str))) $! sheet;
                    return "Testing";
                     };
                  "Headers" -> do {
                     --putStrLn ("Regions \n" ++ show (preHdrInf sheet));
                     --putStrLn ("headers: \n" ++ show (allRun sheet));
                     --putStrLn ("\nHeader \n" ++ (hdrMsg (allRun sheet))++"\n"++(errMsg sheet));
                     writ <- return ("HEADERS: "++(hdrMsg (allRun sheet))++"&&"++(errMsg sheet));
                     putStrLn writ;
                     return writ
                --     (readdata (totalbytes+(length $! str))) sheet;
                     };
                  "Units" -> do {
                     putStrLn ("Forms1 \n" ++ show (sheet));
                     putStrLn ("Units1 \n" ++ show unitSheet);
                     putStrLn ("UnitChecker1 \n" ++ show (unitChecker unitSheet unitSheet (allRun sheet) sheet) ++ "\n");
                     putStrLn ("After Refine Errors \n" ++ show (refineErrs (unitChecker unitSheet unitSheet (allRun sheet) sheet)) ++ "\n");
                     --putStrLn ("After UnitMsg \n" ++ show ((unitMsg (faultLoc sheet (refineErrs (unitChecker (inferU allRunSheet sheet) allRunSheet sheet))))));
                     writ <- return ((unitMsg (unSheet $ faultLoc sheet (refineErrs (unitChecker unitSheet unitSheet allRunSheet sheet)))) ++"&&"++(errMsg sheet));
                     -- putStrLn writ;
                     return writ
                 --    (readdata (totalbytes+(length $! str))) sheet;
                     }; 
                  "Tables" -> do 
                     --putStrLn ("Tables \n" ++ show (fences sheet));
                               let writ = ( hdrMsg (allRun sheet))
                               putStrLn writ
                               return writ
                 --    (readdata (totalbytes+(length $! str))) sheet;
                     
                  otherwise -> do 
                    -- putStrLn ("Server read: " ++ str ++ show totalbytes);
                     --writ <- send sock str;
                     --putStrLn ("Server wrote: " ++ str)
 	             --putStrLn (show sheet);
               -- acho k isto faz parse a uma ss
                  --   (readdata (totalbytes+(length $! str))) $!  (msgProcess str sheet);
                                  return "Nada"
                     
            
            where allRunSheet = allRun sheet
                  unitSheet = inferU (allRun sheet) sheet    


